#include "dict.h"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <math.h>

int reverse;

char w[128];
int wlen;

void push(char c)
{
  if (wlen == 128) abort();
  w[wlen] = c;
  wlen++;
}

void pop(void) { wlen--; }

dictionary dict;

int utf8(unsigned char *n, int *_len)
{
  int v = 0;
  if (*n == 0xc0 || *n == 0xc1 ||
      *n >= 0xf5) abort();
  if (!(*n & 0x80)) {
    v = *n;
    *_len = 1;
  } else
  if ((*n & 0xe0) == 0xc0) { /* 110xxxxx */
    if ((*(n+1) & 0xc0) != 0x80) abort();
    v = ((*n & 0x1f) << 6) | (*(n+1) & 0x3f);
    if (v < 0x80 || v > 0x7ff) abort();
    *_len = 2;
  } else
  if ((*n & 0xf0) == 0xe0) { /* 1110xxxx */
    if ((*(n+1) & 0xc0) != 0x80 ||
        (*(n+2) & 0xc0) != 0x80) abort();
    v = ((*n & 0x0f) << (6*2)) | ((*(n+1) & 0x3f) << 6) | (*(n+2) & 0x3f);
    if (v < 0x800 || v > 0xffff) abort();
    *_len = 3;
  } else
  if ((*n & 0xf8) == 0xf0) { /* 11110xxx */
    if ((*(n+1) & 0xc0) != 0x80 ||
        (*(n+2) & 0xc0) != 0x80 ||
        (*(n+3) & 0xc0) != 0x80) abort();
    v = ((*n & 0x07) << (6*3)) | ((*(n+1) & 0x3f) << (6*2)) |
        ((*(n+2) & 0x3f) << 6) | (*(n+3) & 0x3f);
    if (v < 0x10000 || v > 0x10ffff) abort();
    *_len = 4;
  } else abort();
  return v;
}

int up(int n)
{
  if (isascii(n)) return toupper(n)-'A'+1;
  abort();
}

void read_dict(char *fn)
{
  FILE *f = fopen(fn, "r"); if (!f) { perror(fn); abort(); }
  while (1) {
    unsigned char n[128];
    int ni[128];
    int pos;
    int i;
    int len;
    int wordlen;
    if (!fgets((char*)n, 128, f)) break;
    for (i = 0; i < 127 && n[i]; i++) if (isspace(n[i])) { n[i] = 0; break; }
    n[i] = 0;
    if ((len=wordlen=strlen((char*)n)) == 0 || len > 15) continue;
    if (reverse)
      for (i = 0, pos = 0; n[pos]; i++, pos+=len)
        ni[i] = utf8(n+wordlen-1-pos, &len);
    else
      for (i = 0, pos = 0; n[pos]; i++, pos+=len) ni[i] = utf8(n+pos, &len);
    ni[i] = 0;
    for (i = 0; ni[i]; i++) ni[i] = up(ni[i]);
    ni[i] = 0;
    add_word(&dict, ni);
  }
  fclose(f);
  return;
}

int fgetc_utf8(FILE *f)
{
  int _;
  unsigned char v[4];
  int c = fgetc(f);
  if (c == EOF) return c;
  if (c < 128) return c;
  v[0] = c;
  if ((c & 0xe0) == 0xc0) {
    v[1] = fgetc(f); if (v[1] == EOF) abort();
    return utf8(v, &_);
  }
  if ((c & 0xf0) == 0xe0) {
    v[1] = fgetc(f); if (v[1] == EOF) abort();
    v[2] = fgetc(f); if (v[2] == EOF) abort();
    return utf8(v, &_);
  }
  if ((c & 0xf8) == 0xf0) {
    v[1] = fgetc(f); if (v[1] == EOF) abort();
    v[2] = fgetc(f); if (v[2] == EOF) abort();
    v[3] = fgetc(f); if (v[3] == EOF) abort();
    return utf8(v, &_);
  }
  abort();
}

char target;

void print_stack(void)
{
  int i;
  if (reverse)
    for (i = wlen-1; i >= 0; i--) printf("%c", w[i]+'a'-1);
  else
    for (i = 0; i < wlen; i++) printf("%c", w[i]+'a'-1);
  printf("\n");
}

void dig(dict_letter *cur)
{
  while (cur) {
    if (cur->l == target && cur->w) print_stack();
    cur = cur->next;
  }
}

void go(dict_letter *cur)
{
  if (cur == NULL) return;

  while (cur) {
    push(cur->l);
    if (cur->w) dig(cur->child);
    go(cur->child);
    pop();
    cur = cur->next;
  }
}

char *low(int c)
{
  static char r[2];
  if (c <= 26) { r[0] = c + 'a' - 1; r[1] = 0; return r; }
  switch(c) {
  case 27: return "ç";
  case 28: return "ş";
  case 29: return "ğ";
  case 30: return "ı";
  case 31: return "ö";
  case 32: return "ü";
  }
  abort();
}

char *high(int c)
{
  static char r[2];
  if (c <= 26) { r[0] = c + 'A' - 1; r[1] = 0; return r; }
  switch(c) {
  case 27: return "Ç";
  case 28: return "Ş";
  case 29: return "Ğ";
  case 30: return "İ";
  case 31: return "Ö";
  case 32: return "Ü";
  }
  abort();
}

void usage(void)
{
  printf("usage: contest <dictionary> <letter to check> [options]\n");
  printf("options:\n");
  printf("    -r\n");
  printf("        reverse (check head of word)\n");
  exit(1);
}

int main(int n, char **v)
{
  int i;
  char *fdict = NULL;
  char *letter = NULL;

  for (i = 1; i < n; i++) {
    if (!strcmp(v[i], "-r")) { reverse = 1; continue; }
    if (fdict == NULL) { fdict = v[i]; continue; }
    if (letter == NULL) { letter = v[i]; continue; }
    usage();
  }
  if (fdict == NULL || letter == NULL) usage();
  target = up(letter[0]);
  new_dict(&dict);
  read_dict(fdict);
  go(dict.top);
  return 0;
}
