#include "defs.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdint.h>

static int is_positive_letter(int board[15][15], int row, int col)
{
  return row >= 0 && row < 15 && col >= 0 && col < 15 && board[row][col] > 0;
}

static int other_word_ok(game *g, int l, int i, int j,
                         int inci, int incj)
{
  extern dictionary dict;
  int si = i, sj = j;
  int word[16];
  int w;
  int ii = 1 - inci;
  int jj = 1 - incj;
  int starti = i, startj = j;
  int endi, endj;
  g->board[si][sj] = l;
  /* find start/end pos of word */
  while (starti >= 0 && startj >= 0 && g->board[starti][startj] != 0)
    starti -= ii, startj -= jj;
  starti += ii; startj += jj;
  endi = starti; endj = startj;
  while (endi < 15 && endj < 15 && g->board[endi][endj] != 0)
    endi += ii, endj += jj;
  endi -= ii; endj -= jj;
  /* get word */
  for (w=0, i=starti, j=startj; i!=endi+ii || j!=endj+jj; w++, i+=ii, j+=jj) {
    int letter = abs(g->board[i][j]); if (letter > 26) letter -= 26;
    word[w] = letter;
  }
  word[w] = 0;
  g->board[si][sj] = 0;
  return w == 1 || find_word(&dict, word);
}

static int other_positive_word_ok(game *g, int l, int i, int j,
                                  int inci, int incj)
{
  extern dictionary dict;
  int si = i, sj = j;
  int word[16];
  int w;
  int ii = 1 - inci;
  int jj = 1 - incj;
  int starti = i, startj = j;
  int endi, endj;
  g->board[si][sj] = l;
  /* find start/end pos of word */
  while (starti >= 0 && startj >= 0 && g->board[starti][startj] > 0)
    starti -= ii, startj -= jj;
  starti += ii; startj += jj;
  endi = starti; endj = startj;
  while (endi < 15 && endj < 15 && g->board[endi][endj] > 0)
    endi += ii, endj += jj;
  endi -= ii; endj -= jj;
  /* get word */
  for (w=0, i=starti, j=startj; i!=endi+ii || j!=endj+jj; w++, i+=ii, j+=jj) {
    int letter = abs(g->board[i][j]); if (letter > 26) letter -= 26;
    word[w] = letter;
  }
  word[w] = 0;
  g->board[si][sj] = 0;
  return w == 1 || find_word(&dict, word);
}

typedef struct {
  int row, col, vert;
  int len;
  int cur[15];
  int preboard[15];
} wordat;

typedef struct {
  wordat *w;
  int wlen;
  game *g;
  letterstack *s;
  movelist *ml;
} world;

int valid_board(world *w)
{
  /* for each wordat:
   *   no letter before / after
   *   if letter X not on board, no near letter allowed in the other direction
   */
  int row, col, incrow, inccol, l;
  int i;
  wordat *at;
  for (i = 0; i < w->wlen; i++) {
    at = &w->w[i];
    row = at->row;
    col = at->col;
    if (at->vert) { incrow = 1; inccol = 0; } else { incrow = 0; inccol = 1; }
    if (is_positive_letter(w->g->board, row-incrow, col-inccol)) return 0;
    if (is_positive_letter(w->g->board,
                           row+incrow*at->len, col+inccol*at->len)) return 0;
    for (l = 0; l < at->len; l++, row+=incrow, col+=inccol) {
      if (w->g->board[row][col] >= 0) continue;
      /* letter in *other* direction must be empty (or smthing < 0) */
      if (is_positive_letter(w->g->board, row-inccol, col-incrow)) return 0;
      if (is_positive_letter(w->g->board, row+inccol, col+incrow)) return 0;
    }
  }
  return 1;
}

int putit(world *w, wordat *at)
{
  int count[27];
  int row, col, incrow, inccol, l;
  if (at->vert) { incrow = 1; inccol = 0; } else { incrow = 0; inccol = 1; }
  /* check we can put it */
  for (l = 0; l < 27; l++) count[l] = w->g->lcount[l];
  for (row=at->row,col=at->col,l=0;l<at->len;l++,row+=incrow,col+=inccol) {
    if (w->g->board[row][col] && w->g->board[row][col]!=-at->cur[l]) return 0;
    if (w->g->board[row][col]) continue;
    if (count[at->cur[l]] == 0) return 0;
    count[at->cur[l]]--;
  }
  /* put it */
  for (row=at->row,col=at->col,l=0;l<at->len;l++,row+=incrow,col+=inccol) {
    at->preboard[l] = w->g->board[row][col];
    if (!at->preboard[l]) w->g->lcount[at->cur[l]]--;
    w->g->board[row][col] = -at->cur[l];
  }
  return 1;
}

void unputit(world *w, wordat *at)
{
  int row, col, incrow, inccol, l;
  if (at->vert) { incrow = 1; inccol = 0; } else { incrow = 0; inccol = 1; }
  for (row=at->row,col=at->col,l=0;l<at->len;l++,row+=incrow,col+=inccol) {
    w->g->board[row][col] = at->preboard[l];
    if (!at->preboard[l]) w->g->lcount[at->cur[l]]++;
  }
}

static int word_ok(game *g, int i, int j, int inci, int incj)
{
  extern dictionary dict;
  int word[16];
  int w;
  int ii = inci;
  int jj = incj;
  int starti = i, startj = j;
  int endi, endj;
  /* find start/end pos of word */
  while (starti >= 0 && startj >= 0 && g->board[starti][startj] != 0)
    starti -= ii, startj -= jj;
  starti += ii; startj += jj;
  endi = starti; endj = startj;
  while (endi < 15 && endj < 15 && g->board[endi][endj] != 0)
    endi += ii, endj += jj;
  endi -= ii; endj -= jj;
  /* get word */
  for (w=0, i=starti, j=startj; i!=endi+ii || j!=endj+jj; w++, i+=ii, j+=jj) {
    int letter = abs(g->board[i][j]); if (letter > 26) letter -= 26;
    word[w] = letter;
  }
  word[w] = 0;
  return w == 1 || find_word(&dict, word);
}

static int positive_word_ok(game *g, int i, int j, int inci, int incj)
{
  extern dictionary dict;
  int word[16];
  int w;
  int ii = inci;
  int jj = incj;
  int starti = i, startj = j;
  int endi, endj;
  /* find start/end pos of word */
  while (starti >= 0 && startj >= 0 && g->board[starti][startj] > 0)
    starti -= ii, startj -= jj;
  starti += ii; startj += jj;
  endi = starti; endj = startj;
  while (endi < 15 && endj < 15 && g->board[endi][endj] > 0)
    endi += ii, endj += jj;
  endi -= ii; endj -= jj;
  /* get word */
  for (w=0, i=starti, j=startj; i!=endi+ii || j!=endj+jj; w++, i+=ii, j+=jj) {
    int letter = abs(g->board[i][j]); if (letter > 26) letter -= 26;
    word[w] = letter;
  }
  word[w] = 0;
  return w == 1 || find_word(&dict, word);
}

#if 0
int word_possible(world *w, wordat *at)
{
  int row, col, incrow, inccol, l;
  row = at->row;
  col = at->col;
  if (at->vert) { incrow = 1; inccol = 0; } else { incrow = 0; inccol = 1; }
#if 0
  /* check we don't have something before/after */
  if (letterat(w->g->board, row-incrow, col-inccol)) return 0;
  if (letterat(w->g->board, row+incrow*at->len, col+inccol*at->len)) return 0;
#endif
  for (row=at->row,col=at->col,l=0;l<at->len;l++,row+=incrow,col+=inccol) {
    if (w->g->board[row][col] <= 0) {
      w->g->board[row][col] = at->cur[l];
      if (!word_ok(w->g, row,col,inccol,incrow)) return 0;
    } else {
      if (abs(w->g->board[row][col]) != at->cur[l]) return 0;
    }
  }
  if (!word_ok(w->g, at->row,at->col,incrow,inccol)) return 0;
  return 1;
}

int future_possible(world *w)
{
  int ret = 1;
  int b[15][15];
  int i;
  wordat a;

  memcpy(b, w->g->board, 15*15*sizeof(int));

  for (i = 0; i < w->wlen; i++) {
    wordat *at = w->w + i;
    if (!word_possible(w, at)) { ret = 0; goto end; }
  }

  a = (wordat){ 14, 0, 0, 15, {O,X,Y,P,H,E,N,B,U,T,A,Z,O,N,E} };
  if (!word_possible(w, &a)) { ret = 0; goto end; }

end:
  memcpy(w->g->board, b, 15*15*sizeof(int));
  return ret;
}
#endif

static void dpush(letterstack *s, game *g, dict_letter *n, int hl, int new)
{
  s->stack[s->stacklen].l = n;
  s->stack[s->stacklen].hl = hl;
  s->stack[s->stacklen].new = new;
  s->stacklen++;
}

static void dpop(letterstack *s)
{
  s->stacklen--;
}

static void check_word(game *g, letterstack *s, movelist *ml,
                int starti, int startj, int inci, int incj)
{
  move m;
  int i, j, l, ok = 0;
  /* word must contain at least 1 new letter */
  for (l=0, i=starti, j=startj; l<s->stacklen; l++, i+=inci, j+=incj)
    if (!is_positive_letter(g->board, i, j)) { ok = 1; break; }
  if (!ok) return;
  ok = 0;
  for (l=0, i=starti, j=startj; l<s->stacklen; l++, i+=inci, j+=incj)
    if (is_positive_letter(g->board, i, j) ||             /* letter on board */
        is_positive_letter(g->board, i-incj, j-inci) ||   /* touch on board */
        is_positive_letter(g->board, i+incj, j+inci) ||   /* touch on board */
        (i == 7 && j == 7))            /* word is at center pos (first word) */
    { ok = 1; break; }
  if (!ok) return;
  for (l = 0; l < 7; l++) m.next_hand[l] = g->hand[l];
  for (l = 0; l < s->stacklen; l++) {
    m.word[l] = s->stack[l].l->l;
    if (s->stack[l].hl == 0) m.word[l] += 26;
    if (s->stack[l].new) m.word[l] = -m.word[l];
  }
  m.wordlen = s->stacklen;
  m.row = starti;
  m.col = startj;
  m.vert = inci ? 1 : 0;
  m.score = score(g, m.word, m.wordlen, starti, startj, inci, incj);
  add_movelist(ml, &m);
}

void gogo(world *w, dict_letter *l,
          int starti, int startj, int i, int j, int inci, int incj)
{
  int h;
  int hletter;
  uint32_t letter_done = 0, joker_done = 0;

  if (l==NULL) return;
  if (i==15||j==15) return;
again:
  if (w->g->board[i][j] > 0) {
    int letter = w->g->board[i][j]; if (letter > 26) letter -= 26;
    if (l->l == letter) {
      dpush(w->s, w->g, l, w->g->board[i][j] > 26 ? 0 : w->g->board[i][j], 0);
      if (l->w && (i+inci==15||j+incj==15||w->g->board[i+inci][j+incj]==0))
        check_word(w->g, w->s, w->ml, starti, startj, inci, incj);
      gogo(w, l->child, starti, startj, i + inci, j + incj, inci, incj);
      dpop(w->s);
      goto next;
    }
    goto next;
  }
  if (w->g->board[i][j] < 0) {
    int letter = abs(w->g->board[i][j]); if (letter > 26) {letter-=26;abort();}
    if (l->l != letter) goto next;
    /* reserved letter, must be in hand */
    for (h = 0; h < 7; h++) if (w->g->hand[h] == -w->g->board[i][j]) break;
    if (h == 7) goto next;
    hletter = w->g->hand[h];
    if (letter_done & (1 << hletter)) goto next;
    /* other (already there) word must be ok */
    if (!word_ok(w->g,i, j, incj, inci)) goto next;
    {
      /* to check with positive letter, we must make it >0 */
      int positive;
      w->g->board[i][j] = abs(w->g->board[i][j]);
      positive = positive_word_ok(w->g,i, j, incj, inci);
      w->g->board[i][j] = -w->g->board[i][j];
      if (!positive) goto next;
    }
    w->g->hand[h] = -1;
    letter_done |= 1 << hletter;
    /* no lcount decrement, letter reservation has already done that */
    dpush(w->s, w->g, l, hletter, 1);
    if (l->w && (i+inci==15||j+incj==15||w->g->board[i+inci][j+incj]==0))
      check_word(w->g, w->s, w->ml, starti, startj, inci, incj);
    gogo(w, l->child, starti, startj, i+inci, j+incj, inci, incj);
    dpop(w->s);
    w->g->hand[h] = hletter;
    goto next;
  }

  /* no letter on board, no reserved letter */
  for (h = 0; h < 7; h++) if (w->g->hand[h]==l->l || w->g->hand[h]==0) break;
  if (h == 7) goto next;
  hletter = w->g->hand[h];
  /* must not have been tried before */
  if (hletter && (letter_done & (1 << hletter))) goto next;
  if (!hletter && (joker_done & (1 << l->l))) goto next;
  /* if not a joker, letter must be in lcount */
  if (/*hletter && */!w->g->lcount[hletter]) goto next;
  if (!other_word_ok(w->g, l->l, i, j, inci, incj)) goto next;
  if (!other_positive_word_ok(w->g, l->l, i, j, inci, incj)) goto next;
  w->g->hand[h] = -1;
  if (hletter) letter_done |= 1 << hletter;
  if (!hletter) joker_done |= 1 << l->l;
  w->g->lcount[hletter]--;
  dpush(w->s, w->g, l, hletter, 1);
  if (l->w && (i+inci==15||j+incj==15||w->g->board[i+inci][j+incj]==0))
    check_word(w->g, w->s, w->ml, starti, startj, inci, incj);
  gogo(w, l->child, starti, startj, i+inci, j+incj, inci, incj);
  dpop(w->s);
  w->g->lcount[hletter]++;
  w->g->hand[h] = hletter;

next:
  l = l->next;
  if (l != NULL) goto again;
}

void get_dmoves(world *w)
{
  extern dictionary dict;
  int row, col;
  for (row = 0; row < 15; row++) for (col = 0; col < 15; col++) {
    if (!is_positive_letter(w->g->board, row-1, col))
      gogo(w, dict.top, row, col, row, col, 1, 0);
    if (!is_positive_letter(w->g->board, row, col-1))
      gogo(w, dict.top, row, col, row, col, 0, 1);
  }
}

int move_distance(const void *_a, const void *_b)
{
  const move *a = _a;
  const move *b = _b;
  if (a->distance < b->distance) return -1;
  if (a->distance > b->distance) return 1;
  return 0;
}

int move_length(const void *_a, const void *_b)
{
  const move *a = _a;
  const move *b = _b;
  if (a->score < b->score) return -1;
  if (a->score > b->score) return 1;
  if (a->wordlen < b->wordlen) return -1;
  if (a->wordlen > b->wordlen) return 1;
  return 0;
}

void compute_distance(world *w, move *m)
{
  int i, j;
  int distance = 0;
  play_move_noneg(w->g, m);
  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++)
    if (w->g->board[i][j] < 0) distance--;
  m->distance = distance;
  unplay_move(w->g, m);
}

int distance(int *w, int wlen, int *_hand, int *_bag, int bagpos)
{
  int ret = 0;
  int i, j;
  int hand[7];
  int bag[100];
  memcpy(hand, _hand, 7 * sizeof(int));
  memcpy(bag, _bag, 100 * sizeof(int));
  for (i = 0; i < wlen; i++) {
    if (w[i] > 0) continue;
    /* is letter in hand */
    for (j = 0; j < 7; j++) if (hand[j] == -w[i]) { hand[j] = -1; break; }
    if (j != 7) continue;
    /* where is it in bag */
    for (j = bagpos; j < 100; j++) if (bag[j] == -w[i]) {
      if (j-bagpos+1 > ret) ret = j-bagpos+1;
      bag[j] = -1;
      break;
    }
    if (j == 100) abort();
  }
  return ret;
}

void doit(world *w, int pos);

void target_play(world *w, move *m, int pos)
{
  movelist ml = { NULL, 0, 0 };
  movelist *saved_ml;
  int lcount[27];
  int saved_hand[7];
  int i;
  int nb_letters;
  int row, col, incrow, inccol, l;

  if (m->vert) { incrow = 1; inccol = 0; } else { incrow = 0; inccol = 1; }
  row = m->row;
  col = m->col;
  for (l = 0; l < m->wordlen; l++, row+=incrow, col+=inccol)
    if (w->g->board[row][col] <= 0) break;
  if (l == m->wordlen) { doit(w, pos+1); return; }

  if (!m->distance) {
    play_move_noneg(w->g, m);
    doit(w, pos+1);
    unplay_move(w->g, m);
    return;
  }

  if (options.z) return;

  /* letters already in hand are removed for get_dmoves to not use them */
  memset(lcount, 0, 27*sizeof(int));
  for (i = 0; i < m->wordlen; i++) {
    /* for future, check joker stuff and crash (for now) */
    if (m->word[i] < 0 && abs(m->word[i]) > 26) abort();
    if (m->word[i] < 0) lcount[abs(m->word[i])]++;
  }
  memcpy(saved_hand, w->g->hand, 7 * sizeof(int));
  nb_letters = 0;
  for (i = 0; i < 7; i++)
    if (saved_hand[i] != -1 && lcount[saved_hand[i]])
      { lcount[saved_hand[i]]--; w->g->hand[i] = -1; nb_letters++; if (nb_letters >= m->wordlen) abort(); }
  saved_ml = w->ml;
  w->ml = &ml;
  get_dmoves(w);

  /* we should order moves here wrt target m */
  qsort(ml.v, ml.size, sizeof(move), move_length);

  for (i = ml.size-1; i >= 0; i--) {
    play_move_noneg(w->g, &ml.v[i]);
    if (valid_board(w)) {
      /* unplay the move, restore normal hand, play it again and go down
       * to next word */
      int saved_distance = m->distance;
      int saved_word[15];
      memcpy(saved_word, m->word, 15*sizeof(int));
      unplay_move(w->g, &ml.v[i]);
      memcpy(w->g->hand, saved_hand, 7 * sizeof(int));
      play_move_noneg(w->g, &ml.v[i]);
      if (m->vert) { incrow = 1; inccol = 0; } else { incrow = 0; inccol = 1; }
      row = m->row;
      col = m->col;
      for (l = 0; l < m->wordlen; l++, row+=incrow, col+=inccol) {
        /* don't re-reverse already reserved letters */
        if (w->g->board[row][col] <= 0)
          m->word[l] = -abs(m->word[l]);
        else
          m->word[l] = abs(m->word[l]);
      }
      m->distance = distance(m->word, m->wordlen, w->g->hand,
                             w->g->bag, w->g->bagpos);
      m->score = score(w->g, m->word, m->wordlen,
                       m->row, m->col, m->vert, 1-m->vert);
      target_play(w, m, pos);
      m->distance = saved_distance;
      memcpy(m->word, saved_word, 15*sizeof(int));
      unplay_move(w->g, &ml.v[i]);
      goto done;
    }
    unplay_move(w->g, &ml.v[i]);
  }

done:
  free(ml.v);
  w->ml = saved_ml;
  memcpy(w->g->hand, saved_hand, 7 * sizeof(int));
}

void doit(world *w, int pos)
{
static int maxpos = -1; if (pos > maxpos) {printf("pos %d\n",pos);maxpos=pos;}
  movelist ml = { NULL, 0, 0 };
  extern wordlist wlist[16];
  wordat *at;
  wordlist *wl;
  int row, col, incrow, inccol, l;
  int i;
printf("doit level %d hand:", pos); print_hand(w->g->hand);
print_board(w->g->board);
print_moves(w->g);
  /* done? */
  if (pos == w->wlen) {
    /* don't stop if o, p, b, z or e letters are not empty */
    if (w->g->board[14][0] < 0 &&
        w->g->board[14][3] < 0 &&
        w->g->board[14][7] < 0 &&
        w->g->board[14][11] < 0 &&
        w->g->board[14][14]) {
      print_moves(w->g); abort();
    }
    return;
  }
  at = &w->w[pos];
  if (at->vert) { incrow = 1; inccol = 0; } else { incrow = 0; inccol = 1; }
  /* if all letter there, next level */
  for (row=at->row,col=at->col,l=0;l<at->len;l++,row+=incrow,col+=inccol) {
    if (w->g->board[row][col] <= 0) break;
  }
  if (l == at->len) return doit(w, pos+1);
  /* seek all compatible words, put them in a local movelist, compute
   * distance, sort, and process */
  wl = &wlist[at->len];
  for (i = 0; i < wl->size; i++) {
    int lcount[27];
    memcpy(lcount, w->g->lcount, 27*sizeof(int));
    /* word putable? */
    for (row=at->row,col=at->col,l=0;l<at->len;l++,row+=incrow,col+=inccol) {
      if (w->g->board[row][col] == 0) {
        if (lcount[wl->w[i][l]] == 0) break;
        if (!other_positive_word_ok(w->g, wl->w[i][l], row, col,
                                    incrow, inccol)) break;
        lcount[wl->w[i][l]]--;
        continue;
      }
      if (abs(w->g->board[row][col]) != wl->w[i][l]) break;
    }
    if (l != at->len) continue;
    /* word is ok, create a move for it */
    {
      move m;
      int k;
      row = at->row; col = at->col;
      for (k = 0; k < at->len; k++, row+=incrow, col+=inccol) {
        if (w->g->board[row][col] <= 0)
          m.word[k] = -wl->w[i][k];
        else
          m.word[k] = wl->w[i][k];
      }
      m.wordlen = at->len;
      m.row = at->row; m.col = at->col; m.vert = at->vert;
      m.distance=distance(m.word,m.wordlen,w->g->hand,w->g->bag,w->g->bagpos);
      m.score = score(w->g, m.word, m.wordlen,
                      m.row, m.col, m.vert, 1-m.vert);
      add_movelist(&ml, &m);
    }
  }
  qsort(ml.v, ml.size, sizeof(move), move_distance);

  for (i = 0; i < ml.size; i++) {
    int board[15][15];
    int lcount[27];
    /* we must reserve letters on board */
    memcpy(board, w->g->board, 15*15*sizeof(int));
    memcpy(lcount, w->g->lcount, 27*sizeof(int));
    if (ml.v[i].vert)
      { incrow = 1; inccol = 0; } else { incrow = 0; inccol = 1; }
    row = ml.v[i].row;
    col = ml.v[i].col;
    for (l=0;l<ml.v[i].wordlen;l++,row+=incrow,col+=inccol) {
      if (w->g->board[row][col] != 0) continue;
      w->g->board[row][col] = ml.v[i].word[l];  /* letter is <0 in ml.v */
      if (w->g->lcount[-ml.v[i].word[l]] == 0) abort();
      w->g->lcount[-ml.v[i].word[l]]--;
    }
    target_play(w, &ml.v[i], pos);
    memcpy(w->g->board, board, 15*15*sizeof(int));
    memcpy(w->g->lcount, lcount, 27*sizeof(int));
  }

  free(ml.v);
}

int order_a_score(const void *_a, const void *_b)
{
  const move *a = _a;
  const move *b = _b;
  if (a->a < b->a) return -1;
  if (a->a > b->a) return 1;
  if (a->score < b->score) return -1;
  if (a->score > b->score) return 1;
  return 0;
}

int order_inva_score(const void *_a, const void *_b)
{
  const move *a = _a;
  const move *b = _b;
  /* big a come first in list (smaller interest) */
  if (a->a > b->a) return -1;
  if (a->a < b->a) return 1;
  if (a->score < b->score) return -1;
  if (a->score > b->score) return 1;
  return 0;
}

int order_score(const void *_a, const void *_b)
{
  const move *a = _a;
  const move *b = _b;
  if (a->score < b->score) return -1;
  if (a->score > b->score) return 1;
  return 0;
}

int check_vertical(world *w, int col, int letter)
{
  extern wordlist wlist[16];
  int l, i, j;
  for (l = 3; l < 8; l++) {
    if (is_letter(w->g, 15-1-l, col)) continue;
    for (i = 0; i < wlist[l].size; i++) {
      int *word = wlist[l].w[i];
      if (word[l-1] != letter) continue;
      for (j = 0; j < l; j++) {
        int cur = w->g->board[15-l+j][col]; if (cur > 26) cur -= 26;
        if (cur && cur != word[j]) break;
        if (!cur)
          if (!other_word_ok(w->g, word[j], 15-l+j, col, 1, 0)) break;
      }
      if (j == l) return 1; /* one word found */
    }
  }
  return 0;
}

int bottom_reachable(world *w)
{
  int o, i, subword_ok, start;

  /* try all 'cuts' */
  for (o = 0; o < w->g->bottom_cut_size; o++) {
    /* if a letter is set, it must belong to the 'cut' */
    for (i = 0; i < 15; i++)
      if (w->g->board[14][i] && w->g->board[14][i] != w->g->bottom_cut[o][i])
        break;
    if (i != 15) continue;
    /* we found a compatible cut, check that each 'subword' is accessible
     * by a vertical word of length at least 3
     */
    start = 0;
    while (start < 15) {
      /* go to start of subword */
      while (start < 15 && w->g->bottom_cut[o][start] == 0) start++;
      if (start==15) break;
      /* try all letters of subword */
      subword_ok = 0;
      while (start < 15 && w->g->bottom_cut[o][start] != 0) {
        /* if letter of subword is there then it's 'reached',
         * otherwise check vertical word */
        if (w->g->board[14][start] ||
            check_vertical(w, start, w->g->bottom_cut[o][start]))
          { start++; subword_ok = 1; break; }
        start++;
      }
      if (subword_ok == 0) break;
    }
    /* we reached the end? ok */
    if (start == 15) return 1;
  }

  return 0;
}

int new_decision(world *w, decision_stack *dstack, move *m, reason r,
                 movelist *ml, int i)
{
  int l;
  int lcount[27] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 };
  int triple;
  int bottom_count;

  play_move(w->g, m);

  /* check that move is bottom-compatible with bag */
  if (options.nd == 1) {
    /* triple played at same time only */
    triple = (w->g->board[14][0]!=0) +
             (w->g->board[14][7]!=0) +
             (w->g->board[14][14]!=0);
    if (!(triple == 0 || triple == 3)) goto err;
  } else {
    /* triple and double played at same time only */
    triple = (w->g->board[14][0]!=0) +
             (w->g->board[14][3]!=0) +
             (w->g->board[14][7]!=0) +
             (w->g->board[14][11]!=0) +
             (w->g->board[14][14]!=0);
    if (!(triple == 0 || triple == 5)) goto err;
  }
  /* bottom puttable? */
  bottom_count = 0;
  for (l = 0; l < 15; l++) {
    /* count letters */
    if (w->g->board[14][l] <= 0) lcount[w->g->bottom[l]]++;
    /* count # letters already put */
    if (w->g->board[14][l] > 0) bottom_count++;
    /* already there letter must be in bottom word */
    if (w->g->board[14][l] > 0 &&
        w->g->board[14][l] != w->g->bottom[l]) goto err;
  }
  for (l = 0; l < 27; l++) if (lcount[l] > w->g->lcount[l]) goto err;

  /* bottom reachable? */
  if (bottom_count != 15 && !bottom_reachable(w)) goto err;

  /* put bottom and check other word */
  for (l = 0; l < 15; l++) {
    int old = w->g->board[14][l];
    int ok;
    w->g->board[14][l] = w->g->bottom[l];
    ok = other_word_ok(w->g, w->g->bottom[l], 14, l, 0, 1);
    w->g->board[14][l] = old;
    if (!ok) goto err;
  }

  /* add new decision */
  if (dstack->size == dstack->maxsize) {
    dstack->maxsize += 128;
    dstack->e = realloc(dstack->e, dstack->maxsize * sizeof(decision_elem));
    if (dstack->e == NULL) abort();
  }
  dstack->e[dstack->size].m = *m;
  dstack->e[dstack->size].r = r;
  dstack->e[dstack->size].ml = *ml;
  dstack->e[dstack->size].mlpos = i;
  dstack->size++;

#if 0
  printf("new decision %d hand:", r);
  print_hand(w->g->hand);
  print_board(w->g->board);
#endif
  static int best_score = 0;
  if (w->g->score > best_score) {
    best_score = w->g->score;
    for (i = 0; i < dstack->size; i++)
      w->g->moves[i] = dstack->e[i].m;
    w->g->movesize = dstack->size;
    print_moves(w->g);
    printf("hand:"); print_hand(w->g->hand);
    print_board(w->g->board);
    printf("score %d\n", best_score);
  }
  /* reset move list */
  *w->ml = (movelist){ NULL, 0, 0 };
  return 1;

err:
  unplay_move(w->g, m);
  return 0;
}

int compute_distance_bottom(int board[15][15])
{
  int i, j, col, dist, row = 14;
  int mindist[15] = { 99,99,99,99,99,99,99,99,99,99,99,99,99,99,99 };
  for (i=0;i<15;i++) for (j=0;j<15;j++) {
    if (board[i][j] <= 0) continue;
    for (col = 0; col < 15; col++) {
      dist = abs(i-row) + abs(j-col);
      if (dist < mindist[col]) mindist[col] = dist;
    }
  }
  dist = mindist[0];
  for (i = 1; i < 15; i++) dist += mindist[i];
  return dist;
}

void dig(game *g)
{
  letterstack stack = { stacklen : 0 };
  decision_stack dstack = { e:NULL, maxsize:0 };
  movelist ml = { NULL, 0, 0 };
  world w;
  int i, l, count, precount, dist, predist, ready_letters, preready_letters;
  decision_elem decision;
  int current_hand_lcount[27];
  int target_hand_lcount[27];

  w.g = g;
  w.s = &stack;
  w.ml = &ml;

next:
  get_dmoves(&w);

  /* last line ready? */
  count = 0;
  for (i = 0; i < 15; i++) if (g->board[14][i] > 0) count++;
  if (count < 8) goto last_line_not_ready;

  /* hand ready? */
  memset(current_hand_lcount, 0, 27*sizeof(int));
  memset(target_hand_lcount, 0, 27*sizeof(int));
  for (i = 0; i < 15; i++)
    if (g->board[14][i] <= 0) target_hand_lcount[g->bottom[i]]++;
  for (i = 0; i < 7; i++)
    if (g->hand[i] > 0) current_hand_lcount[g->hand[i]]++;
  for (i = 0; i < 27; i++)
    if (target_hand_lcount[i] < current_hand_lcount[i]) break;
  if (i != 27) goto hand_not_ready;

  goto best;

hand_not_ready:
  /* compute # of ready letters in hand */
  memset(current_hand_lcount, 0, 27*sizeof(int));
  for (l = 0; l < 7; l++)
    if (g->hand[l] > 0) current_hand_lcount[g->hand[l]]++;
  ready_letters = 0;
  for (l = 0; l < 27; l++) {
    if (target_hand_lcount[l] == 0) continue;
    if (current_hand_lcount[l] < target_hand_lcount[l])
      ready_letters += current_hand_lcount[l];
    else
      ready_letters += target_hand_lcount[l];
  }
  preready_letters = ready_letters;
  /* get closer to target hand */
  /* set 'a' to # *new* target letters in hand */
  for (i = 0; i < ml.size; i++) {
    play_move(g, &ml.v[i]);
    memset(current_hand_lcount, 0, 27*sizeof(int));
    for (l = 0; l < 7; l++)
      if (g->hand[l] > 0) current_hand_lcount[g->hand[l]]++;
    ready_letters = 0;
    for (l = 0; l < 27; l++) {
      if (target_hand_lcount[l] == 0) continue;
      if (current_hand_lcount[l] < target_hand_lcount[l])
        ready_letters += current_hand_lcount[l];
      else
        ready_letters += target_hand_lcount[l];
    }
    unplay_move(g, &ml.v[i]);
    ml.v[i].a = ready_letters - preready_letters;
  }
  qsort(ml.v, ml.size, sizeof(move), order_a_score);
  i = ml.size-1;
next_closer_to_target_hand:
  for (; i >= 0; i--)
    if (ml.v[i].a != 0)
      if (new_decision(&w, &dstack, &ml.v[i], CLOSER_TO_HAND, &ml, i))
        goto next;

last_line_not_ready:
  /* check if we can feed last line */
  /* set 'a' to # new letters */
  precount = count;
  for (i = 0; i < ml.size; i++) {
    int l;
    play_move(g, &ml.v[i]);
    count = 0; for (l = 0; l < 15; l++) if (g->board[14][l] > 0) count++;
    unplay_move(g, &ml.v[i]);
    ml.v[i].a = count - precount;
  }
  qsort(ml.v, ml.size, sizeof(move), order_a_score);
  i = ml.size-1;
next_feed_last_line:
  for (; i >= 0; i--)
    if (ml.v[i].a != 0)
      if (new_decision(&w, &dstack, &ml.v[i], FEED_LAST_LINE, &ml, i))
        goto next;

  /* check if we can get closer to bottom */
  predist = compute_distance_bottom(g->board);
  for (i = 0; i < ml.size; i++) {
    play_move(g, &ml.v[i]);
    dist = compute_distance_bottom(g->board);
    unplay_move(g, &ml.v[i]);
    ml.v[i].a = dist;
  }
  qsort(ml.v, ml.size, sizeof(move), order_inva_score);
  /* precompute decision ('predist' may be dead at next loop) */
  for (i = 0; i < ml.size; i++) ml.v[i].a = ml.v[i].a < predist;
  i = ml.size-1;
next_closer_to_bottom:
  for (; i >= 0; i--)
    if (ml.v[i].a)
      if (new_decision(&w, &dstack, &ml.v[i], CLOSER_TO_BOTTOM, &ml, i))
        goto next;

best:
  /* play best move (score, maybe check letters used wrt bottom) (only one) */
  qsort(ml.v, ml.size, sizeof(move), order_score);
  for (i = ml.size-1; i >= 0; i--)
  if (new_decision(&w, &dstack, &ml.v[i], BEST, &ml, i))
    goto next;

  /* tried it all, go back up */
pop:
  if (dstack.size == 0) abort();
  decision = dstack.e[dstack.size-1];
  dstack.size--;
  if (ml.maxsize) free(ml.v);
  ml = decision.ml;
  w.ml = &ml;
  unplay_move(g, &decision.m);
  switch (decision.r) {
  default: abort();
  case FEED_LAST_LINE:
    i = decision.mlpos - 1;
    goto next_feed_last_line;
  case CLOSER_TO_BOTTOM:
    i = decision.mlpos - 1;
    goto next_closer_to_bottom;
  case CLOSER_TO_HAND:
    i = decision.mlpos - 1;
    goto next_closer_to_target_hand;
  case BEST:
    goto pop;
  }

  exit(1);
}
