#ifndef _DEFS_H_
#define _DEFS_H_

#define _ 0
#define A 1
#define B 2
#define C 3
#define D 4
#define E 5
#define F 6
#define G 7
#define H 8
#define I 9
#define J 10
#define K 11
#define L 12
#define M 13
#define N 14
#define O 15
#define P 16
#define Q 17
#define R 18
#define S 19
#define T 20
#define U 21
#define V 22
#define W 23
#define X 24
#define Y 25
#define Z 26

#include "dict.h"

struct game;
struct constraint;

typedef struct {
  int (*check)(struct game *, struct constraint *);
  int (*is_changeable)(struct game *, struct constraint *);
  int (*change)(struct game *, struct constraint *);
} constraint_main;

/* type of wordat constraints */
#define WORDAT_ABSOLUTE   0
#define WORDAT_CHANGEABLE 1

typedef struct {
  constraint_main common;
  int type;
  int word[16];
  int wordlen;
  int row, col, vert;
} wordat_constraint;

typedef struct {
  constraint_main common;
  int row1, col1;
  int row2, col2;
} play_together_constraint;

typedef struct constraint {
  union {
    constraint_main _1;
    wordat_constraint _2;
    play_together_constraint _3;
  };
} constraint;

typedef struct {
  /* same as board in game struct. >0 values for already used letters */
  int word[15];
  int wordlen;
  int row, col, vert;
  int next_hand[7];
  int pre_hand[7];
  int preboard[15];
  unsigned char prelcount[27];
  int pre_bagpos;
  int pre_handsize;
  int score;
  int distance;
  int a, b, c, d;
} move;

typedef struct {
  move *v;
  int size;
  int maxsize;
} movelist;

typedef int cut[15];

typedef struct game {
  /* value board[row][col] is:
   *     0               : free
   *     1     .. 26     : A-Z
   *     1+26  .. 26+26  : joker with given letter
   *     -1    .. -26    : A-Z new (for a current move or whatever)
   *     -1-26 .. -26-26 : joker new (for a current move or whatever)
   */
  int board[15][15];

  /*
   * 0     is a joker
   * 1..26 is a letter
   * -1    is an empty position (only in hand, not in bag)
   */
  int hand[7];
  int handsize;
  int bag[100];
  int bagpos;
  int lcount[27];

  move moves[100];
  int movesize;

  constraint **constraints;
  int csize;
  int cmaxsize;

  int score;

  /* top target word */
  int top[15];
  cut *top_cut;
  int top_cut_size;
} game;

typedef struct {
  dict_letter *l;
  int hl;         /* hand's value (to check for jokers) */
  int new;
} letterstack_elem;

typedef struct {
  letterstack_elem stack[15];
  int stacklen;
} letterstack;

typedef enum {
  FEED_FIRST_LINE, CLOSER_TO_TOP, BEST, CLOSER_TO_HAND
} reason;

typedef struct {
  move m;
  reason r;
  movelist ml;
  int mlpos;
} decision_elem;

typedef struct {
  decision_elem *e;
  int size;
  int maxsize;
} decision_stack;

void init_game(game *, char *bag);
void play_move(game *g, move *m);
void play_move_noneg(game *g, move *m);
void unplay_move(game *g, move *m);
void game_preplay(game *, char *);
void print_board(int [15][15]);
void print_hand(int *);
void print_moves(game *g);

void add_constraint_wordat(game *, int type, int *word, int row, int col,
                           int vert);
void add_constraint_play_together(game *, int, int, int, int);
void solve_constraints(game *);

void init_movelist(movelist *);
void free_movelist(movelist *);
void add_movelist(movelist *, move *);
void get_moves(game *, movelist *, dictionary *);

int is_letter(game *g, int i, int j);
int score(game *, int *, int wlen, int starti,int startj,int inci,int incj);

extern struct options {
  int nd;

  /* below: dead stuff */
  int z;
  int ats;
} options;

#endif /* _DEFS_H_ */
