#include "dict.h"
#include "defs.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

dictionary dict;
wordlist wlist[16];

void usage(void)
{
  printf("usage: [options] <dictionary> <bag> <top word> <bottom word>");
  printf("  <top cut> <bottom cut>\n");
  printf("    <word_top_min> <word_top_max> <cut_top_min> <cut_top_max>\n");
  printf("    <word_bottom_min> <word_bottom_max>");
  printf(" <cut_bottom_min> <cut_bottom_max>\n");
  printf("option:\n");
  printf("    -ndt\n");
  printf("        don't care of double letters in top word\n");
  printf("    -ndb\n");
  printf("        don't care of double letters in bottom word\n");
  exit(0);
}

struct options options;

int main(int n, char **v)
{
  game g;
  char *fdict = NULL;
  char *bag = NULL;
  char *top = NULL;
  char *bottom = NULL;
  char *topcut = NULL;
  char *bottomcut = NULL;
  char *premoves = NULL;
  int word_top_min=-1, word_top_max=-1, cut_top_min=-1, cut_top_max=-1;
  int word_bottom_min=-1, word_bottom_max=-1,
      cut_bottom_min=-1, cut_bottom_max=-1;
  int i;
  int wordcount;

  for (i = 1; i < n; i++) {
    if (!strcmp(v[i],"-ndt"))       { options.ndt = 1; continue; }
    if (!strcmp(v[i],"-ndb"))       { options.ndb = 1; continue; }
    if (fdict == NULL) { fdict = v[i]; continue; }
    if (bag == NULL) { bag = v[i]; continue; }
    if (top == NULL) { top = v[i]; continue; }
    if (bottom == NULL) { bottom = v[i]; continue; }
    if (topcut == NULL) { topcut = v[i]; continue; }
    if (bottomcut == NULL) { bottomcut = v[i]; continue; }
#define a(x) if (x == -1) { x = atoi(v[i]); continue; }
    a(word_top_min); a(word_top_max); a(cut_top_min); a(cut_top_max);
    a(word_bottom_min); a(word_bottom_max);
    a(cut_bottom_min); a(cut_bottom_max);
#undef a
    usage();
  }
  if (fdict == NULL || bag == NULL || top == NULL || bottom == NULL) usage();
  if (topcut == NULL || bottomcut == NULL) usage();
  if (word_top_min == -1 || word_top_max == -1 ||
      cut_top_min == -1 || cut_top_max == -1) usage();
  if (word_bottom_min == -1 || word_bottom_max == -1 ||
      cut_bottom_min == -1 || cut_bottom_max == -1) usage();

  new_dict(&dict);
  read_dict(fdict, &dict, 0);
  for (i = 2; i <= 15; i++) {
    init_wordlist(&wlist[i], i);
    dict_to_wordlist(&dict, &wlist[i], i);
  }

  wordcount = 0;
  for (i = 2; i <= 15; i++) {
    printf("words of length %d [%d]\n", i, wlist[i].size);
    wordcount += wlist[i].size;
  }
  printf("%d words in dict\n", wordcount);

  init_game(&g, bag);

  if (premoves != NULL) game_preplay(&g, premoves);

  if (strlen(top) != 15) usage();
  for (i = 0; i < 15; i++) {
    int l = top[i];
    if      (l >= 'a' && l <= 'z') l = l - 'a' + 1;
    else if (l >= 'A' && l <= 'Z') l = l - 'A' + 1;
    else usage();
    g.top[i] = l;
  }
  cut *getcut(int word[15], int *retlen, int no_double);
  static int tc[15];
  int tc_nb_holes = 0;
  for (i = 0; i < 15; i++) {
    int l = topcut[i];
    if      (l >= 'a' && l <= 'z') l = l - 'a' + 1;
    else if (l >= 'A' && l <= 'Z') l = l - 'A' + 1;
    else if (l == '.') { l = 0; tc_nb_holes++; }
    else usage();
    tc[i] = l;
  }
  g.top_cut = &tc; //getcut(g.top, &g.top_cut_size, options.ndt);
  g.top_cut_size = 1;
  g.tc_nb_holes = tc_nb_holes;

  if (strlen(bottom) != 15) usage();
  for (i = 0; i < 15; i++) {
    int l = bottom[i];
    if      (l >= 'a' && l <= 'z') l = l - 'a' + 1;
    else if (l >= 'A' && l <= 'Z') l = l - 'A' + 1;
    else usage();
    g.bottom[i] = l;
  }
  static int bc[15];
  int bc_nb_holes = 0;
  for (i = 0; i < 15; i++) {
    int l = bottomcut[i];
    if      (l >= 'a' && l <= 'z') l = l - 'a' + 1;
    else if (l >= 'A' && l <= 'Z') l = l - 'A' + 1;
    else if (l == '.') { l = 0; bc_nb_holes++; }
    else usage();
    bc[i] = l;
  }
  g.bottom_cut = &bc; //getcut(g.bottom, &g.bottom_cut_size, options.ndt);
  g.bottom_cut_size = 1;
  //g.bottom_cut = getcut(g.bottom, &g.bottom_cut_size, options.ndb);
  g.bc_nb_holes = bc_nb_holes;

  g.word_top_min = word_top_min;
  g.word_top_max = word_top_max;
  g.cut_top_min = cut_top_min;
  g.cut_top_max = cut_top_max;
  g.word_bottom_min = word_bottom_min;
  g.word_bottom_max = word_bottom_max;
  g.cut_bottom_min = cut_bottom_min;
  g.cut_bottom_max = cut_bottom_max;

  void dig(game *);
  dig(&g);

  return 0;
}
