#include "defs.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdint.h>

static int is_positive_letter(letter board[15][15], int row, int col)
{
  return row >= 0 && row < 15 && col >= 0 && col < 15 &&
         board[row][col].bagpos != -1 && board[row][col].value > 0;
}

static int other_word_ok(game *g, letter l, int i, int j,
                         int inci, int incj)
{
  extern dictionary dict;
  int si = i, sj = j;
  int word[16];
  int w;
  int ii = 1 - inci;
  int jj = 1 - incj;
  int starti = i, startj = j;
  int endi, endj;
  g->board[si][sj] = l;
  /* find start/end pos of word */
  while (starti >= 0 && startj >= 0 && g->board[starti][startj].bagpos != -1)
    starti -= ii, startj -= jj;
  starti += ii; startj += jj;
  endi = starti; endj = startj;
  while (endi < 15 && endj < 15 && g->board[endi][endj].bagpos != -1)
    endi += ii, endj += jj;
  endi -= ii; endj -= jj;
  /* get word */
  for (w=0, i=starti, j=startj; i!=endi+ii || j!=endj+jj; w++, i+=ii, j+=jj) {
    word[w] = abs(g->board[i][j].value);
  }
  word[w] = 0;
  g->board[si][sj].bagpos = -1;
  return w == 1 || find_word(&dict, word);
}

static int other_positive_word_ok(game *g, letter l, int i, int j,
                                  int inci, int incj)
{
  extern dictionary dict;
  int si = i, sj = j;
  int word[16];
  int w;
  int ii = 1 - inci;
  int jj = 1 - incj;
  int starti = i, startj = j;
  int endi, endj;
  g->board[si][sj] = l;
  /* find start/end pos of word */
  while (starti >= 0 && startj >= 0 && g->board[starti][startj].bagpos != -1 &&
         g->board[starti][startj].value > 0)
    starti -= ii, startj -= jj;
  starti += ii; startj += jj;
  endi = starti; endj = startj;
  while (endi < 15 && endj < 15 && g->board[endi][endj].bagpos != -1 &&
         g->board[starti][startj].value > 0)
    endi += ii, endj += jj;
  endi -= ii; endj -= jj;
  /* get word */
  for (w=0, i=starti, j=startj; i!=endi+ii || j!=endj+jj; w++, i+=ii, j+=jj) {
    word[w] = abs(g->board[i][j].value);
  }
  word[w] = 0;
  g->board[si][sj].bagpos = -1;
  return w == 1 || find_word(&dict, word);
}

typedef struct {
  int row, col, vert;
  int len;
  int cur[15];
  int preboard[15];
} wordat;

typedef struct {
  wordat *w;
  int wlen;
  game *g;
  letterstack *s;
  movelist *ml;
} world;

#if 0
int valid_board(world *w)
{
  /* for each wordat:
   *   no letter before / after
   *   if letter X not on board, no near letter allowed in the other direction
   */
  int row, col, incrow, inccol, l;
  int i;
  wordat *at;
  for (i = 0; i < w->wlen; i++) {
    at = &w->w[i];
    row = at->row;
    col = at->col;
    if (at->vert) { incrow = 1; inccol = 0; } else { incrow = 0; inccol = 1; }
    if (is_positive_letter(w->g->board, row-incrow, col-inccol)) return 0;
    if (is_positive_letter(w->g->board,
                           row+incrow*at->len, col+inccol*at->len)) return 0;
    for (l = 0; l < at->len; l++, row+=incrow, col+=inccol) {
      if (w->g->board[row][col] >= 0) continue;
      /* letter in *other* direction must be empty (or smthing < 0) */
      if (is_positive_letter(w->g->board, row-inccol, col-incrow)) return 0;
      if (is_positive_letter(w->g->board, row+inccol, col+incrow)) return 0;
    }
  }
  return 1;
}
#endif

static void dpush(letterstack *s, game *g, dict_letter *n, int bp, int new)
{
  s->stack[s->stacklen].l = n;
  s->stack[s->stacklen].bagpos = bp;
  s->stack[s->stacklen].new = new;
  s->stacklen++;
}

static void dpop(letterstack *s)
{
  s->stacklen--;
}

static void check_word(game *g, letterstack *s, movelist *ml,
                int starti, int startj, int inci, int incj)
{
  move m;
  int i, j, l, ok = 0;
  /* word must contain at least 1 new letter */
  for (l=0, i=starti, j=startj; l<s->stacklen; l++, i+=inci, j+=incj)
    if (!is_positive_letter(g->board, i, j)) { ok = 1; break; }
  if (!ok) return;
  ok = 0;
  for (l=0, i=starti, j=startj; l<s->stacklen; l++, i+=inci, j+=incj)
    if (is_positive_letter(g->board, i, j) ||             /* letter on board */
        is_positive_letter(g->board, i-incj, j-inci) ||   /* touch on board */
        is_positive_letter(g->board, i+incj, j+inci) ||   /* touch on board */
        (i == 7 && j == 7))            /* word is at center pos (first word) */
    { ok = 1; break; }
  if (!ok) return;
  for (l = 0; l < 7; l++) m.next_hand[l] = g->hand[l];
  for (l = 0; l < s->stacklen; l++) {
    m.word[l].value = s->stack[l].l->l;
    m.word[l].bagpos = s->stack[l].bagpos;
    if (s->stack[l].new) m.word[l].value = -m.word[l].value;
  }
  m.wordlen = s->stacklen;
  m.row = starti;
  m.col = startj;
  m.vert = inci ? 1 : 0;
  m.score = score(g, m.word, m.wordlen, starti, startj, inci, incj);
  add_movelist(ml, &m);
}

void gogo(world *w, dict_letter *l,
          int starti, int startj, int i, int j, int inci, int incj)
{
  int h;
  int hletter, hpos;
  uint32_t letter_done = 0, joker_done = 0;
  //uint32_t constrained_done = 0;

  if (l==NULL) return;
  if (i==15||j==15) return;
again:
  if (w->g->board[i][j].bagpos != -1) {
    int letter = w->g->board[i][j].value;
    if (l->l == letter) {
      dpush(w->s, w->g, l, w->g->board[i][j].bagpos, 0);
      if (l->w && (i+inci==15||j+incj==15||
                   w->g->board[i+inci][j+incj].bagpos==-1))
        check_word(w->g, w->s, w->ml, starti, startj, inci, incj);
      gogo(w, l->child, starti, startj, i + inci, j + incj, inci, incj);
      dpop(w->s);
      goto next;
    }
    goto next;
  }

  /* no letter on board, no reserved letter */
  for (h = 0; h < 7; h++) if (w->g->hand[h] != -1) {
    if (w->g->con[i][j].on) {
      if (w->g->con[i][j].value == l->l &&
          w->g->con[i][j].bagpos == w->g->hand[h])
      if (w->g->bag[w->g->hand[h]]==l->l || w->g->bag[w->g->hand[h]]==0) break;
    } else if (w->g->bag_constraint[w->g->hand[h]].on) {
      if (w->g->bag_constraint[w->g->hand[h]].row == i &&
          w->g->bag_constraint[w->g->hand[h]].col == j)
      if (w->g->bag[w->g->hand[h]]==l->l || w->g->bag[w->g->hand[h]]==0) break;
    } else {
      if (w->g->bag[w->g->hand[h]]==l->l || w->g->bag[w->g->hand[h]]==0) break;
    }
  }
  if (h == 7) goto next;
  hpos = w->g->hand[h];
  hletter = w->g->bag[hpos];
  /* must not have been tried before */
  if (hletter && (letter_done & (1 << hletter))) goto next;
  if (!hletter && (joker_done & (1 << l->l))) goto next;
  /* if not a joker, letter must be in lcount */
  if (/*hletter && */!w->g->lcount[hletter]) goto next;
  if (!other_word_ok(w->g, (letter){value:l->l,bagpos:hpos},
                     i, j, inci, incj)) goto next;
  if (!other_positive_word_ok(w->g, (letter){value:l->l,bagpos:hpos},
                              i, j, inci, incj)) goto next;
  w->g->hand[h] = -1;
  if (hletter) letter_done |= 1 << hletter;
  if (!hletter) joker_done |= 1 << l->l;
  w->g->lcount[hletter]--;
  dpush(w->s, w->g, l, hpos, 1);
  if (l->w && (i+inci==15||j+incj==15||w->g->board[i+inci][j+incj].bagpos==-1))
    check_word(w->g, w->s, w->ml, starti, startj, inci, incj);
  gogo(w, l->child, starti, startj, i+inci, j+incj, inci, incj);
  dpop(w->s);
  w->g->lcount[hletter]++;
  w->g->hand[h] = hpos;

next:
  l = l->next;
  if (l != NULL) goto again;
}

void get_dmoves(world *w)
{
  extern dictionary dict;
  int row, col;
  for (row = 0; row < 15; row++) for (col = 0; col < 15; col++) {
    if (!is_positive_letter(w->g->board, row-1, col))
      gogo(w, dict.top, row, col, row, col, 1, 0);
    if (!is_positive_letter(w->g->board, row, col-1))
      gogo(w, dict.top, row, col, row, col, 0, 1);
  }
}

int move_distance(const void *_a, const void *_b)
{
  const move *a = _a;
  const move *b = _b;
  if (a->distance < b->distance) return -1;
  if (a->distance > b->distance) return 1;
  return 0;
}

int move_length(const void *_a, const void *_b)
{
  const move *a = _a;
  const move *b = _b;
  if (a->score < b->score) return -1;
  if (a->score > b->score) return 1;
  if (a->wordlen < b->wordlen) return -1;
  if (a->wordlen > b->wordlen) return 1;
  return 0;
}

int order_a_score(const void *_a, const void *_b)
{
  const move *a = _a;
  const move *b = _b;
  if (a->a < b->a) return -1;
  if (a->a > b->a) return 1;
  if (a->score < b->score) return -1;
  if (a->score > b->score) return 1;
  return 0;
}

int order_inva_score(const void *_a, const void *_b)
{
  const move *a = _a;
  const move *b = _b;
  /* big a come first in list (smaller interest) */
  if (a->a > b->a) return -1;
  if (a->a < b->a) return 1;
  if (a->score < b->score) return -1;
  if (a->score > b->score) return 1;
  return 0;
}

int order_score(const void *_a, const void *_b)
{
  const move *a = _a;
  const move *b = _b;
  if (a->score < b->score) return -1;
  if (a->score > b->score) return 1;
  return 0;
}

int check_vertical_bottom(world *w, int col, int lt)
{
  extern wordlist wlist[16];
  int l, i, j;
  for (l = 3; l < 8; l++) {
    if (is_letter(w->g, 15-1-l, col)) continue;
    for (i = 0; i < wlist[l].size; i++) {
      int *word = wlist[l].w[i];
      if (word[l-1] != lt) continue;
      for (j = 0; j < l; j++) {
        int cur = w->g->board[15-l+j][col].value;
        if (w->g->board[15-l+j][col].bagpos != -1 && cur != word[j]) break;
        if (w->g->board[15-l+j][col].bagpos == -1)
          /* bagpos value meaningless, must jut be != -1 */
          if (!other_word_ok(w->g, (letter){value:word[j],bagpos:0},
                             15-l+j, col, 1, 0)) break;
      }
      if (j == l) return 1; /* one word found */
    }
  }
  return 0;
}

int check_vertical_top(world *w, int col, int lt)
{
  extern wordlist wlist[16];
  int l, i, j;
  for (l = 3; l < 8; l++) {
    if (is_letter(w->g, l, col)) continue;
    for (i = 0; i < wlist[l].size; i++) {
      int *word = wlist[l].w[i];
      if (word[0] != lt) continue;
      for (j = 0; j < l; j++) {
        int cur = w->g->board[j][col].value;
        if (w->g->board[j][col].bagpos != -1 && cur != word[j]) break;
        if (w->g->board[j][col].bagpos == -1)
          /* bagpos value meaningless, must jut be != -1 */
          if (!other_word_ok(w->g, (letter){value:word[j],bagpos:0},
                             j, col, 1, 0)) break;
      }
      if (j == l) return 1; /* one word found */
    }
  }
  return 0;
}

int bottom_reachable(world *w)
{
  int o, i, subword_ok, start;

  /* try all 'cuts' */
  for (o = 0; o < w->g->bottom_cut_size; o++) {
    /* if a letter is set, it must belong to the 'cut' */
    for (i = 0; i < 15; i++)
      if (w->g->board[14][i].bagpos != -1 &&
          w->g->board[14][i].value != w->g->bottom_cut[o][i])
        break;
    if (i != 15) continue;
    /* we found a compatible cut, check that each 'subword' is accessible
     * by a vertical word of length at least 3
     */
    start = 0;
    while (start < 15) {
      /* go to start of subword */
      while (start < 15 && w->g->bottom_cut[o][start] == 0) start++;
      if (start==15) break;
      /* try all letters of subword */
      subword_ok = 0;
      while (start < 15 && w->g->bottom_cut[o][start] != 0) {
        /* if letter of subword is there then it's 'reached',
         * otherwise check vertical word */
        if (w->g->board[14][start].bagpos != -1 ||
            check_vertical_bottom(w, start, w->g->bottom_cut[o][start]))
          { start++; subword_ok = 1; break; }
        start++;
      }
      if (subword_ok == 0) break;
    }
    /* we reached the end? ok */
    if (start == 15) return 1;
  }

  return 0;
}

int top_reachable(world *w)
{
  int o, i, subword_ok, start;

  /* try all 'cuts' */
  for (o = 0; o < w->g->top_cut_size; o++) {
    /* if a letter is set, it must belong to the 'cut' */
    for (i = 0; i < 15; i++)
      if (w->g->board[0][i].bagpos != -1 &&
          w->g->board[0][i].value != w->g->top_cut[o][i])
        break;
    if (i != 15) continue;
    /* we found a compatible cut, check that each 'subword' is accessible
     * by a vertical word of length at least 3
     */
    start = 0;
    while (start < 15) {
      /* go to start of subword */
      while (start < 15 && w->g->top_cut[o][start] == 0) start++;
      if (start==15) break;
      /* try all letters of subword */
      subword_ok = 0;
      while (start < 15 && w->g->top_cut[o][start] != 0) {
        /* if letter of subword is there then it's 'reached',
         * otherwise check vertical word */
        if (w->g->board[0][start].bagpos != -1 ||
            check_vertical_top(w, start, w->g->top_cut[o][start]))
          { start++; subword_ok = 1; break; }
        start++;
      }
      if (subword_ok == 0) break;
    }
    /* we reached the end? ok */
    if (start == 15) return 1;
  }

  return 0;
}

int new_decision(world *w, decision_stack *dstack, move *m, reason r,
                 movelist *ml, int i)
{
  int l;
  int lcount[27] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 };
  int triple;
  int top_count, bottom_count;

  play_move(w->g, m);

#if 0
  /* check that move is bottom-compatible with bag */
  if (options.ndb == 1) {
    /* triple played at same time only */
    triple = (w->g->board[14][0].bagpos!=-1) +
             (w->g->board[14][7].bagpos!=-1) +
             (w->g->board[14][14].bagpos!=-1);
    if (!(triple == 0 || triple == 3)) goto err;
  } else {
    /* triple and double played at same time only */
    triple = (w->g->board[14][0].bagpos!=-1) +
             (w->g->board[14][3].bagpos!=-1) +
             (w->g->board[14][7].bagpos!=-1) +
             (w->g->board[14][11].bagpos!=-1) +
             (w->g->board[14][14].bagpos!=-1);
    if (!(triple == 0 || triple == 5)) goto err;
  }
  /* bottom puttable? */
  bottom_count = 0;
  for (l = 0; l < 15; l++) {
    /* count letters */
    if (w->g->board[14][l].bagpos == -1) lcount[w->g->bottom[l]]++;
    /* count # letters already put */
    if (w->g->board[14][l].bagpos != -1) bottom_count++;
    /* already there letter must be in bottom word */
    if (w->g->board[14][l].bagpos != -1 &&
        w->g->board[14][l].value != w->g->bottom[l]) goto err;
  }
  for (l = 0; l < 27; l++) if (lcount[l] > w->g->lcount[l]) goto err;

  /* bottom reachable? */
  if (bottom_count != 15 && !bottom_reachable(w)) goto err;

  /* put bottom and check other word */
  for (l = 0; l < 15; l++) {
    letter old = w->g->board[14][l];
    int ok;
    /* bagpos meaningless, must be != -1 */
    w->g->board[14][l].bagpos = 0;
    w->g->board[14][l].value = w->g->bottom[l];
    ok = other_word_ok(w->g, (letter){value:w->g->bottom[l],bagpos:0},
                       14, l, 0, 1);
    w->g->board[14][l] = old;
    if (!ok) goto err;
  }
#endif

  /* check that move is top-compatible with bag */
  if (options.ndt == 1) {
    /* triple played at same time only */
    triple = (w->g->board[0][0].bagpos!=-1) +
             (w->g->board[0][7].bagpos!=-1) +
             (w->g->board[0][14].bagpos!=-1);
    if (!(triple == 0 || triple == 3)) goto err;
  } else {
    /* triple and double played at same time only */
    triple = (w->g->board[0][0].bagpos!=-1) +
             (w->g->board[0][3].bagpos!=-1) +
             (w->g->board[0][7].bagpos!=-1) +
             (w->g->board[0][11].bagpos!=-1) +
             (w->g->board[0][14].bagpos!=-1);
    if (!(triple == 0 || triple == 5)) goto err;
  }
  /* top puttable? */
  top_count = 0;
  for (l = 0; l < 15; l++) {
    /* count letters */
    if (w->g->board[0][l].bagpos == -1) lcount[w->g->top[l]]++;
    /* count # letters already put */
    if (w->g->board[0][l].bagpos != -1) top_count++;
    /* already there letter must be in top word */
    if (w->g->board[0][l].bagpos != -1 &&
        w->g->board[0][l].value != w->g->top[l]) goto err;
  }
  for (l = 0; l < 27; l++) if (lcount[l] > w->g->lcount[l]) goto err;

  /* top reachable? */
  if (top_count != 15 && !top_reachable(w)) goto err;

  /* put top and check other word */
  for (l = 0; l < 15; l++) {
    letter old = w->g->board[0][l];
    int ok;
    /* bagpos value meaningless, must be != -1 */
    w->g->board[0][l].bagpos = 0;
    w->g->board[0][l].value = w->g->top[l];
    ok = other_word_ok(w->g, (letter){value:w->g->top[l],bagpos:0},
                       0, l, 0, 1);
    w->g->board[0][l] = old;
    if (!ok) goto err;
  }

  /* add new decision */
  if (dstack->size == dstack->maxsize) {
    dstack->maxsize += 128;
    dstack->e = realloc(dstack->e, dstack->maxsize * sizeof(decision_elem));
    if (dstack->e == NULL) abort();
  }
  dstack->e[dstack->size].m = *m;
  dstack->e[dstack->size].r = r;
  dstack->e[dstack->size].ml = *ml;
  dstack->e[dstack->size].mlpos = i;
  dstack->size++;

  static int best_score = 0;
  if (1 || w->g->score > best_score) {
    best_score = w->g->score;
    for (i = 0; i < dstack->size; i++)
      w->g->moves[i] = dstack->e[i].m;
    w->g->movesize = dstack->size;
    print_moves(w->g, w->g->bag);
    printf("hand:"); print_hand(w->g->hand, w->g->bag);
    printf("bag:"); print_bag(w->g->bag, w->g->bagpos);
    print_board(w->g->board, w->g->bag);
    printf("score %d\n", best_score);
  }
  /* reset move list */
  *w->ml = (movelist){ NULL, 0, 0 };
  return 1;

err:
  unplay_move(w->g, m);
  return 0;
}

int compute_distance_both(letter board[15][15])
{
  int i, j, col, dist, row;
  int mindist_top[15] = { 99999,99999,99999,99999,99999,99999,99999,99999,99999,99999,99999,99999,99999,99999,99999 };
  int mindist_bottom[15] = { 99999,99999,99999,99999,99999,99999,99999,99999,99999,99999,99999,99999,99999,99999,99999 };
  for (i=0;i<15;i++) for (j=0;j<15;j++) {
    if (board[i][j].bagpos == -1) continue;
    row = 0;
    for (col = 0; col < 15; col++) {
      dist = abs(i-row) + abs(j-col);
      if (j-col) dist += 1;
      if (dist < mindist_top[col]) mindist_top[col] = dist;
    }
    row = 14;
    for (col = 0; col < 15; col++) {
      dist = abs(i-row) + abs(j-col);
      if (j-col) dist += 1;
      if (dist < mindist_bottom[col]) mindist_bottom[col] = dist;
    }
  }
  dist = 0;
  for (i = 0; i < 15; i++) dist += mindist_top[i];
  //for (i = 0; i < 15; i++) dist += mindist_bottom[i];
  for (i = 0; i < 15; i++) {
    if (board[0][i].bagpos == -1) dist += 100;
    //if (board[14][i].bagpos == -1) dist += 100;
  }

  return dist;
}

void dig(game *g)
{
  letterstack stack = { stacklen : 0 };
  decision_stack dstack = { e:NULL, maxsize:0 };
  movelist ml = { NULL, 0, 0 };
  world w;
  int i, l, count, precount, dist, predist, ready_letters, preready_letters;
  decision_elem decision;
  int current_hand_lcount[27];
  int target_hand_lcount[27];

  w.g = g;
  w.s = &stack;
  w.ml = &ml;

next:
  get_dmoves(&w);

  /* minimise distance */
  predist = compute_distance_both(g->board);
  for (i = 0; i < ml.size; i++) {
    play_move(g, &ml.v[i]);
    dist = compute_distance_both(g->board);
    unplay_move(g, &ml.v[i]);
    ml.v[i].a = dist;
  }
  qsort(ml.v, ml.size, sizeof(move), order_inva_score);
  /* precompute decision ('predist' may be dead at next loop) */
  //for (i = 0; i < ml.size; i++) ml.v[i].a = ml.v[i].a < predist;
  i = ml.size-1;
next_closer_to_top:
  for (; i >= 0; i--)
    if (ml.v[i].a)
      if (new_decision(&w, &dstack, &ml.v[i], CLOSER_TO_TOP, &ml, i))
        goto next;

#if 0
best:
  /* play best move (score, maybe check letters used wrt bottom) (only one) */
  qsort(ml.v, ml.size, sizeof(move), order_score);
  for (i = ml.size-1; i >= 0; i--)
  if (new_decision(&w, &dstack, &ml.v[i], BEST, &ml, i))
    goto next;
#endif

  /* tried it all, go back up */
pop:
  if (dstack.size == 0) abort();
  decision = dstack.e[dstack.size-1];
  dstack.size--;
  if (ml.maxsize) free(ml.v);
  ml = decision.ml;
  w.ml = &ml;
  unplay_move(g, &decision.m);
  switch (decision.r) {
  default: abort();
  case CLOSER_TO_TOP:
    i = decision.mlpos - 1;
    goto next_closer_to_top;
  case BEST:
    goto pop;
  }

  exit(1);
}
