#include "dict.h"
#include "defs.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

dictionary dict;
wordlist wlist[16];

void usage(void)
{
  printf("usage: [options] <dictionary> <bag> <constraint file>\n");
  printf("    <top word> <bottom word> <top cut> <bottom cut>\n");
  printf("    <numbers (15 * 2)>\n");
  printf("15 first numbers for positions of top word, >=100 if in the cut\n");
  printf("constraint file contains lines like:\n");
  printf("    <row> <col> <bapos> <value (A-Z no joker allowed, a-z yes)>\n");
  printf("option:\n");
  printf("    -ndt\n");
  printf("        don't care of double letters in top word\n");
  printf("    -ndb\n");
  printf("        don't care of double letters in bottom word\n");
  printf("    -preplay <moves>\n");
  exit(0);
}

struct options options;

void read_constraints(game *g, char *fn)
{
  FILE *f = fopen(fn, "r"); if (f == NULL) { perror(fn); exit(1); }
  char s[512];
  int row, col, bagpos;
  char value;
  for (bagpos=0; bagpos<100; bagpos++) g->bag_constraint[bagpos].on = 0;
  for (row = 0; row < 15; row++) for (col = 0; col < 15; col++)
    g->con[row][col].on = 0;
  while (1) {
    if (fgets(s, 512, f) == NULL) break;
    if (s[0] == '#' || s[0] == '\n') continue;
    if (sscanf(s, "%d %d %d %c", &row, &col, &bagpos, &value) != 4) abort();
    if (row < 0 || row > 14 || col < 0 || col > 14 || bagpos < 0 ||
        bagpos > 99 || !((value >= 'a' && value <= 'z') ||
        (value >= 'A' && value <= 'Z'))) abort();
    g->con[row][col].on = 1;
    g->con[row][col].value = value >= 'a' && value <= 'z' ? value - 'a' + 1 :
                             value - 'A' + 1;
    g->con[row][col].bagpos = bagpos;
    g->bag_constraint[bagpos].on = 1;
    g->bag_constraint[bagpos].row = row;
    g->bag_constraint[bagpos].col = col;
  }
  fclose(f);
}

int main(int n, char **v)
{
  game g;
  char *fdict = NULL;
  char *bag = NULL;
  char *top = NULL;
  char *bottom = NULL;
  char *topcut = NULL;
  char *bottomcut = NULL;
  char *premoves = NULL;
  char *constraint = NULL;
  int i;
  int letter = 0;
  int top_bottom_letters[30];
  int wordcount;

  for (i = 1; i < n; i++) {
    if (!strcmp(v[i],"-ndt"))       { options.ndt = 1; continue; }
    if (!strcmp(v[i],"-ndb"))       { options.ndb = 1; continue; }
    if (!strcmp(v[i],"-preplay")) {if(i>n-2)usage();premoves=v[++i];continue;}
    if (fdict == NULL) { fdict = v[i]; continue; }
    if (bag == NULL) { bag = v[i]; continue; }
    if (constraint == NULL) { constraint = v[i]; continue; }
    if (top == NULL) { top = v[i]; continue; }
    if (bottom == NULL) { bottom = v[i]; continue; }
    if (topcut == NULL) { topcut = v[i]; continue; }
    if (bottomcut == NULL) { bottomcut = v[i]; continue; }
    if (letter < 30) { top_bottom_letters[letter++] = atoi(v[i]); continue; }
    usage();
  }
  if (fdict == NULL || bag == NULL || top == NULL || bottom == NULL ||
      constraint == NULL || topcut == NULL || bottomcut == NULL) usage();
  if (letter != 30) usage();

  new_dict(&dict);
  read_dict(fdict, &dict, 0);
  for (i = 2; i <= 15; i++) {
    init_wordlist(&wlist[i], i);
    dict_to_wordlist(&dict, &wlist[i], i);
  }

  wordcount = 0;
  for (i = 2; i <= 15; i++) {
    printf("words of length %d [%d]\n", i, wlist[i].size);
    wordcount += wlist[i].size;
  }
  printf("%d words in dict\n", wordcount);

  init_game(&g, bag);

  if (premoves != NULL) game_preplay(&g, premoves);

  if (strlen(top) != 15) usage();
  for (i = 0; i < 15; i++) {
    int l = top[i];
    if      (l >= 'a' && l <= 'z') l = l - 'a' + 1;
    else if (l >= 'A' && l <= 'Z') l = l - 'A' + 1;
    else usage();
    g.top[i] = l;
  }
  cut *getcut(int word[15], int *retlen, int no_double);
  static int tc[15];
  int tc_nb_holes = 0;
  for (i = 0; i < 15; i++) {
    int l = topcut[i];
    if      (l >= 'a' && l <= 'z') l = l - 'a' + 1;
    else if (l >= 'A' && l <= 'Z') l = l - 'A' + 1;
    else if (l == '.') { l = 0; tc_nb_holes++; }
    else usage();
    tc[i] = l;
  }
  g.top_cut = &tc; //getcut(g.top, &g.top_cut_size, options.ndt);
  g.top_cut_size = 1;

  if (strlen(bottom) != 15) usage();
  for (i = 0; i < 15; i++) {
    int l = bottom[i];
    if      (l >= 'a' && l <= 'z') l = l - 'a' + 1;
    else if (l >= 'A' && l <= 'Z') l = l - 'A' + 1;
    else usage();
    g.bottom[i] = l;
  }
  static int bc[15];
  int bc_nb_holes = 0;
  for (i = 0; i < 15; i++) {
    int l = bottomcut[i];
    if      (l >= 'a' && l <= 'z') l = l - 'a' + 1;
    else if (l >= 'A' && l <= 'Z') l = l - 'A' + 1;
    else if (l == '.') { l = 0; bc_nb_holes++; }
    else usage();
    bc[i] = l;
  }
  g.bottom_cut = &bc; //getcut(g.bottom, &g.bottom_cut_size, options.ndt);
  g.bottom_cut_size = 1;
  //g.bottom_cut = getcut(g.bottom, &g.bottom_cut_size, options.ndb);

  read_constraints(&g, constraint);

  memset(g.top_bottom_letter, 0, 100*sizeof(int));
  /* top word constraints */
  for (i = 0; i < 15; i++) {
    g.top_bottom_letter[top_bottom_letters[i]] = 1;
    if (g.top_cut[i] != 0) continue;
    g.con[0][i].on = 1;
    g.con[0][i].value = g.bag[top_bottom_letters[i]];
    g.con[0][i].bagpos = top_bottom_letters[i];
    g.bag_constraint[top_bottom_letters[i]].on = 1;
    g.bag_constraint[top_bottom_letters[i]].row = 0;
    g.bag_constraint[top_bottom_letters[i]].col = i;
  }
  /* bottom word constraints */
  for (i = 15; i < 30; i++) {
    g.top_bottom_letter[top_bottom_letters[i]] = 1;
    if (g.top_cut[i] != 0) continue;
    g.con[14][i-15].on = 1;
    g.con[14][i-15].value = g.bag[top_bottom_letters[i]];
    g.con[14][i-15].bagpos = top_bottom_letters[i];
    g.bag_constraint[top_bottom_letters[i]].on = 1;
    g.bag_constraint[top_bottom_letters[i]].row = 14;
    g.bag_constraint[top_bottom_letters[i]].col = i;
  }

  void dig(game *);
  dig(&g);

  return 0;
}
