#include "defs.h"
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>

void init_game(game *g, char *b)
{
  int row, col, i;
  for (row = 0; row < 15; row++) for (col = 0; col < 15; col++)
    g->board[row][col].bagpos = -1;
  if (strlen(b) != 100) abort();
  for (i = 0; i < 27; i++) g->lcount[i] = 0;
  for (i = 0; i < 100; i++) {
    if (b[i] >= 'a' && b[i] <= 'z') g->bag[i] = b[i] - 'a' + 1;
    else if (b[i] >= 'A' && b[i] <= 'Z') g->bag[i] = b[i] - 'A' + 1;
    else if (b[i] == '?') g->bag[i] = 0;
else if (b[i] == '.') { g->bag[i] = -1; continue; }
    else abort();
    g->lcount[g->bag[i]]++;
  }
  g->bagpos = 7;
  for (i = 0; i < 7; i++) g->hand[i] = i;
  g->handsize = 7;
  g->movesize = 0;
  g->constraints = NULL;
  g->csize = g->cmaxsize = 0;
  g->score = 0;
}

void play_move(game *g, move *m)
{
  int row, col, incrow, inccol, k, l;
  if (m->vert) { incrow = 1; inccol = 0; } else { incrow = 0; inccol = 1; }
  for (k = 0; k < 7; k++) m->pre_hand[k] = g->hand[k];
  for (k = 0; k < 27; k++) m->prelcount[k] = g->lcount[k];
  m->pre_bagpos = g->bagpos;
  m->pre_handsize = g->handsize;
  for (row=m->row,col=m->col,l=0; l<m->wordlen; row+=incrow,col+=inccol,l++) {
    if (row < 0 || row > 15 || col < 0 || col > 15) abort();
    m->preboard[l] = g->board[row][col];
    if (m->word[l].value < 0) {
      int bagpos = m->word[l].bagpos;
      if (g->lcount[g->bag[bagpos]] == 0) abort();
      g->lcount[g->bag[bagpos]]--;
      for (k = 0; k < 7; k++)
        if (g->hand[k] == bagpos) { g->hand[k] = -100; break; }
      if (k == 7) abort();
    }
    if (m->word[l].value > 0 &&
        (g->board[row][col].value != m->word[l].value ||
         g->board[row][col].bagpos != m->word[l].bagpos)) abort();
    if (m->word[l].value < 0 && g->board[row][col].bagpos != -1) abort();
    g->board[row][col].bagpos = m->word[l].bagpos;
    g->board[row][col].value = abs(m->word[l].value);
  }
  for (k = 0; k < 7; k++)
    if (g->hand[k] == -100) {
      if (g->bagpos != 100) {
        g->hand[k] = g->bagpos;
        g->bagpos++;
      }
      else g->hand[k] = -1;
    }
  g->handsize = 0;
  for (k = 0; k < 7; k++) if (g->hand[k] != -1) g->handsize++;
  g->score += m->score;
  if (g->movesize == 100) abort();
  g->moves[g->movesize] = *m;
  g->movesize++;
}

void unplay_move(game *g, move *m)
{
  int new_letters = 0, i;
  int row, col, incrow, inccol, l, k;
  for (i = 0; i < m->wordlen; i++) if (m->word[i].value < 0) new_letters++;
  if (g->bagpos - new_letters < 0) abort();
  if (m->vert) { incrow = 1; inccol = 0; } else { incrow = 0; inccol = 1; }
  for (row=m->row,col=m->col,l=0;l<m->wordlen;row+=incrow,col+=inccol,l++) {
    //int letter;
    if (m->word[l].value >= 0) continue;
    if (g->board[row][col].value != -m->word[l].value) abort();
    if (g->board[row][col].bagpos != m->word[l].bagpos) abort();
    g->board[row][col] = m->preboard[l];
  }
  g->bagpos = m->pre_bagpos;
  g->handsize = m->pre_handsize;
  for (k = 0; k < 7; k++) g->hand[k] = m->pre_hand[k];
  for (k = 0; k < 27; k++) g->lcount[k] = m->prelcount[k];
  g->score -= m->score;
  g->movesize--;
}

void game_preplay(game *g, char *moves)
{
#if 0
  int truescore;
  move m;
  char *p = moves;
  while (*p) {
    int row, col, incrow, inccol, vert, l, dascore = -1;
    char colc;
    while (*p && (isspace(*p) || *p == ',')) p++;
    if (*p >= '0' && *p <= '9') {
      vert = 0; incrow = 0; inccol = 1;
      if (sscanf(p, "%d%c", &row, &colc) != 2) abort();
    } else {
      vert = 1; incrow = 1; inccol = 0;
      if (sscanf(p, "%c%d", &colc, &row) != 2) abort();
    }
    row--;
    col = colc >= 'a' && colc <= 'z' ? colc - 'a' : colc - 'A';
    if (row < 0 || row > 15 || col < 0 || col > 15) abort();
    while (*p && (!isspace(*p))) p++;
    while (*p && isspace(*p)) p++;
    m.wordlen = 0;
    while ((*p >= 'a' && *p <= 'z') || (*p >= 'A' && *p <= 'Z')) {
      if (m.wordlen > 15) abort();
      if (*p >= 'a' && *p <= 'z') m.word[m.wordlen] = *p - 'a' + 1 + 26;
      else                        m.word[m.wordlen] = *p - 'A' + 1;
      m.wordlen++;
      p++;
    }
    sscanf(p, "%d", &dascore);
    truescore = score(g, m.word, m.wordlen, row, col, incrow, inccol);
    if (dascore > 0 && dascore != truescore) abort();
    while (*p && *p != ',') p++;
    m.row = row;
    m.col = col;
    m.vert = vert;
    m.score = truescore;
    for (l=0; l<m.wordlen; row+=incrow,col+=inccol,l++) {
      if (row < 0 || row > 15 || col < 0 || col > 15) abort();
      if (m.word[l] != g->board[row][col]) m.word[l] = -m.word[l];
    }
    play_move_noneg(g, &m);
  }
#endif
}

void print_board(letter b[15][15], int *bag)
{
  int i, j;
  for (i = 0; i < 15; i++) {
    for (j = 0; j < 15; j++) {
      int l = b[i][j].value;
      if (l < 0) printf("\x1b[32m");
      printf("%c", b[i][j].bagpos != -1 ?
                    bag[b[i][j].bagpos] == 0 ?
                      abs(b[i][j].value) + 'a' - 1 :
                      abs(b[i][j].value) + 'A' - 1 : '.');
      if (l < 0) printf("\x1b[0m");
    }
    printf("\n");
  }
}

void print_hand(int *h, int *bag)
{
  int i;
  for (i = 0; i < 7; i++)
    printf("%c[%d] ",
        h[i] == -1 ? '.' : bag[h[i]] ? bag[h[i]] + 'A' - 1 : '?', h[i]);
  printf("\n");
}

void print_bag(int *b, int bpos)
{
  int i;
  if (bpos == 100) { printf("(empty)\n"); return; }
  for (i = bpos; i < 100; i++)
    printf("%c", b[i] == -1 ? '.' : b[i] ? b[i] + 'A' - 1 : '?');
  printf("\n");
}

void print_moves(game *g, int *bag)
{
  int i, j;
  for (i = 0; i < g->movesize; i++) {
    int row = g->moves[i].row;
    int col = g->moves[i].col;
    int len = g->moves[i].wordlen; 
    int vert = g->moves[i].vert;
    int score = g->moves[i].score;
    int incrow, inccol;
    if (!vert) {
      printf("%d%c ", row+1, col+'A');
      incrow = 0;
      inccol = 1;
    } else {
      printf("%c%d ", col+'A', row+1);
      incrow = 1;
      inccol = 0;
    }
    for (j = 0; j < len; j++, row += incrow, col += inccol) {
      int c = g->board[row][col].value;
      int joker = bag[g->board[row][col].bagpos] == 0;
      printf("%c", c+'A'-1 + joker * ('a'-'A'));
    }
    printf(" %d", score);
    if (i == g->movesize-1) printf("\n"); else printf(",");
  }
}
