#include "dict.h"
#include "defs.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

dictionary dict;
wordlist wlist[16];

void usage(void)
{
  printf("usage: [options] <dictionary> <bag> <top word> <bottom word>");
  printf(" <constraint file>\n");
  printf("constraint file contains lines like:\n");
  printf("    <row> <col> <bapos> <value (A-Z no joker allowed, a-z yes)>\n");
  printf("option:\n");
  printf("    -ndt\n");
  printf("        don't care of double letters in top word\n");
  printf("    -ndb\n");
  printf("        don't care of double letters in bottom word\n");
  exit(0);
}

struct options options;

void read_constraints(game *g, char *fn)
{
  FILE *f = fopen(fn, "r"); if (f == NULL) { perror(fn); exit(1); }
  char s[512];
  int row, col, bagpos;
  char value;
  for (bagpos=0; bagpos<100; bagpos++) g->bag_constraint[bagpos].on = 0;
  for (row = 0; row < 15; row++) for (col = 0; col < 15; col++)
    g->con[row][col].on = 0;
  while (1) {
    if (fgets(s, 512, f) == NULL) break;
    if (s[0] == '#' || s[0] == '\n') continue;
    if (sscanf(s, "%d %d %d %c", &row, &col, &bagpos, &value) != 4) abort();
    if (row < 0 || row > 14 || col < 0 || col > 14 || bagpos < 0 ||
        bagpos > 99 || !((value >= 'a' && value <= 'z') ||
        (value >= 'A' && value <= 'Z'))) abort();
    g->con[row][col].on = 1;
    g->con[row][col].value = value >= 'a' && value <= 'z' ? value - 'a' + 1 :
                             value - 'A' + 1;
    g->con[row][col].bagpos = bagpos;
    g->bag_constraint[bagpos].on = 1;
    g->bag_constraint[bagpos].row = row;
    g->bag_constraint[bagpos].col = col;
  }
  fclose(f);
}

int main(int n, char **v)
{
  game g;
  char *fdict = NULL;
  char *bag = NULL;
  char *top = NULL;
  char *bottom = NULL;
  char *premoves = NULL;
  char *constraint = NULL;
  int i;
  int wordcount;

  for (i = 1; i < n; i++) {
    if (!strcmp(v[i],"-ndt"))       { options.ndt = 1; continue; }
    if (!strcmp(v[i],"-ndb"))       { options.ndb = 1; continue; }
    if (fdict == NULL) { fdict = v[i]; continue; }
    if (bag == NULL) { bag = v[i]; continue; }
    if (top == NULL) { top = v[i]; continue; }
    if (bottom == NULL) { bottom = v[i]; continue; }
    if (constraint == NULL) { constraint = v[i]; continue; }
    usage();
  }
  if (fdict == NULL || bag == NULL || top == NULL || bottom == NULL ||
      constraint == NULL) usage();

  new_dict(&dict);
  read_dict(fdict, &dict, 0);
  for (i = 2; i <= 15; i++) {
    init_wordlist(&wlist[i], i);
    dict_to_wordlist(&dict, &wlist[i], i);
  }

  wordcount = 0;
  for (i = 2; i <= 15; i++) {
    printf("words of length %d [%d]\n", i, wlist[i].size);
    wordcount += wlist[i].size;
  }
  printf("%d words in dict\n", wordcount);

  init_game(&g, bag);

  if (premoves != NULL) game_preplay(&g, premoves);

  if (strlen(top) != 15) usage();
  for (i = 0; i < 15; i++) {
    int l = top[i];
    if      (l >= 'a' && l <= 'z') l = l - 'a' + 1;
    else if (l >= 'A' && l <= 'Z') l = l - 'A' + 1;
    else usage();
    g.top[i] = l;
  }
  cut *getcut(int word[15], int *retlen, int no_double);
  g.top_cut = getcut(g.top, &g.top_cut_size, options.ndt);

  if (strlen(bottom) != 15) usage();
  for (i = 0; i < 15; i++) {
    int l = bottom[i];
    if      (l >= 'a' && l <= 'z') l = l - 'a' + 1;
    else if (l >= 'A' && l <= 'Z') l = l - 'A' + 1;
    else usage();
    g.bottom[i] = l;
  }
  g.bottom_cut = getcut(g.bottom, &g.bottom_cut_size, options.ndb);

  read_constraints(&g, constraint);

  void dig(game *);
  dig(&g);

  return 0;
}
