#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "dict.h"
#include "scrabble.h"

typedef struct {
  dict_letter *l;
  int hl;         /* hand's value (to check for jokers) */
  int i, j;
} stack_elem;

typedef struct {
  /* from main.c */
  int is_first;
  int board[15][15];
  int is_joker[15][15];
  int hand[7];
  int starti, startj, inci, incj;
  stack_elem stack[15];
  int stacklen;
  /* from contest.c */
  int next_letter;
  /* internal to this structure (not restored on pop) */
  int cumul_score;
  int steps[100];  // access with STEP_XX defined below
  int steps_size;
  int played_board[15][15];
  int played_is_joker[15][15];
} game_state;

#define G_STEP_I(x)        ( (x)       & 0x0f)
#define G_STEP_J(x)        (((x) >> 4) & 0x0f)
#define G_STEP_LEN(x)      (((x) >> 8) & 0x0f)
#define G_STEP_VERTICAL(x) (((x) >> 12) & 0x01)
#define G_STEP_SCORE(x)    (((x) >> 16) & 0xffff)

#define S_STEP_I(x, v)        x = ((x) &   ~0x0f) |  (v)
#define S_STEP_J(x, v)        x = ((x) &   ~0xf0) | ((v) << 4)
#define S_STEP_LEN(x, v)      x = ((x) &  ~0xf00) | ((v) << 8)
#define S_STEP_VERTICAL(x, v) x = ((x) & ~0x1000) | ((v) << 12)
#define S_STEP_SCORE(x, v)    x = ((x) & ~0xffff0000) | ((v) << 16)

static char letters[200];
static int translated_letters[200];
static int next_letter;

game_state game_stack[1024];
int game_stack_size;

extern int is_first;
extern int board[15][15];
extern int is_joker[15][15];
extern int hand[7];
extern int starti, startj, inci, incj;
extern stack_elem stack[15];
extern int stacklen;
extern int nb_input_letters;

int my_isalpha(int n);
int up(int n);

void add_step(game_state *g, int i, int j, int inci, int len, int score)
{
  if (g->steps_size == 100) abort();
  g->steps[g->steps_size] = 0;
  S_STEP_I(g->steps[g->steps_size], i);
  S_STEP_J(g->steps[g->steps_size], j);
  S_STEP_VERTICAL(g->steps[g->steps_size], inci);
  S_STEP_LEN(g->steps[g->steps_size], len);
  S_STEP_SCORE(g->steps[g->steps_size], score);
  g->steps_size++;
}

void translate_letters(void)
{
  int i;
  for (i = 0; letters[i]; i++) {
    int c = letters[i];
    if (c == '?') c = 0;                 /* joker */
    else { if (!my_isalpha(c)) abort(); c = up(c); }
    translated_letters[i] = c;
  }
}

void push_game(void)
{
  game_state *g;
  if (game_stack_size == 1024) abort();
  g = &game_stack[game_stack_size];
  game_stack_size++;
  g->is_first = is_first;
  memcpy(g->board, board, 15*15*sizeof(int));
  memcpy(g->is_joker, is_joker, 15*15*sizeof(int));
  memcpy(g->hand, hand, 7*sizeof(int));
  g->starti = starti;
  g->startj = startj;
  g->inci = inci;
  g->incj = incj;
  memcpy(g->stack, stack, 15*sizeof(stack_elem));
  g->stacklen = stacklen;
  g->next_letter = next_letter;
}

void pop_game(void)
{
  game_state *g;
  if (game_stack_size == 0) abort();
  game_stack_size--;
  g = &game_stack[game_stack_size];
  is_first = g->is_first;
  memcpy(board, g->board, 15*15*sizeof(int));
  memcpy(is_joker, g->is_joker, 15*15*sizeof(int));
  memcpy(hand, g->hand, 7*sizeof(int));
  starti = g->starti;
  startj = g->startj;
  inci = g->inci;
  incj = g->incj;
  memcpy(stack, g->stack, 15*sizeof(stack_elem));
  stacklen = g->stacklen;
  next_letter = g->next_letter;
}

extern dictionary dict;

int is_letter(int i, int j);
void read_hand(char *h);
void go(dict_letter *cur, int i, int j);

/* global "static" variable, set by first_hand/next_hand, used by eat_letter */
/* not stored in push_game/pop_game */
static int bagpos[7];

char eat_letter(void)
{
  char r = letters[next_letter];
  /* we don't use letters in bagpos, they are the target */
  if (next_letter == bagpos[0] || next_letter == bagpos[1] ||
      next_letter == bagpos[2] || next_letter == bagpos[3] ||
      next_letter == bagpos[4] || next_letter == bagpos[5] ||
      next_letter == bagpos[6]) {
    next_letter++;
    return ':';
  }
  if (!r) return '.';
  next_letter++;
  if (r == '?') return '_';
  return r;
}

/* local "static" variable, set by first_hand/next_hand, used by contest */
static int init_hand[7];
/* copy of above before first_hand/next_hand modifies it */
static int init_hand_saved[7];

void contest(char *bag)
{
  //char cur_hand[8];
  int i, j;

#if 0
  if (strlen(bag) > 199) abort();
  strcpy(letters, bag);
  translate_letters();
#endif

#if 0
  for (i = 0; i < 7; i++) cur_hand[i] = eat_letter();
  cur_hand[7] = 0;
  read_hand(cur_hand);
#endif
  nb_input_letters = 0;
  for (i = 0; i < 7; i++) {
    hand[i] = init_hand[i];
    if (hand[i] < 0x1000000) nb_input_letters++;
  }

  inci = 0; incj = 1;
  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) {
    starti = i; startj = j;
    /* must be a start of a word (no letter before) */
    if (is_letter(starti-inci, startj-incj)) continue;
    is_first = 1;
    go(dict.top, i, j);
  }
  inci = 1; incj = 0;
  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) {
    starti = i; startj = j;
    /* must be a start of a word (no letter before) */
    if (is_letter(starti-inci, startj-incj)) continue;
    is_first = 1;
    go(dict.top, i, j);
  }
}

int best_score = -10000000;

int total_score(int score)
{
  extern letter english_letters[];
  int i;
  for (i = 0; i < 7; i++)
    if (hand[i] < 0x1000000) score -= english_letters[hand[i]].value;
  for (i = next_letter; letters[i]; i++)
    score -= english_letters[translated_letters[i]].value;
  return score;
}

void print_board(void)
{
  int i, j;
  for (i = 0; i < 15; i++) {
    for (j = 0; j < 15; j++) printf("%c", board[i][j]!=-10000000 ? 'A' + board[i][j]-1 : '.');
    printf("\n");
  }
}

void print_steps(game_state *g)
{
  int i, j;
  for (i = 0; i < g->steps_size; i++) {
    int row = G_STEP_I(g->steps[i]);
    int col = G_STEP_J(g->steps[i]);
    int len = G_STEP_LEN(g->steps[i]);
    int vert = G_STEP_VERTICAL(g->steps[i]);
    int score = G_STEP_SCORE(g->steps[i]);
    int incrow, inccol;
    if (!vert) {
      printf("%d%c ", row+1, col+'A');
      incrow = 0;
      inccol = 1;
    } else {
      printf("%c%d ", col+'A', row+1);
      incrow = 1;
      inccol = 0;
    }
    for (j = 0; j < len; j++, row += incrow, col += inccol) {
      int c = g->played_board[row][col];
      int is_joker = g->played_is_joker[row][col];
      printf("%c", c+'A'-1 + is_joker * ('a'-'A'));
    }
    printf(" %d", score);
    if (i == g->steps_size-1) printf("\n"); else printf(",");
  }
}

static int _find15(dict_letter *l, int *w)
{
  if (*w == 0) return 0;
  if (l == NULL) return 0;
next:
  if (l->l == *w || *w == -10000000) {
    if (l->w) return 1;
    if (_find15(l->child, w+1)) return 1;
    if (l->l == *w) return 0;
  }
  l = l->next;
  if (l != NULL) goto next;
  return 0;
}

int find15(int *w)
{
  extern dictionary dict15;
  return _find15(dict15.top, w);
}

/* global static variable, set by preplay, used by put_word */
static int starting_score;

void put_word(int score, int _starti, int _startj, int vert, int *letters)
{
  int i, j;
  int current_total_score;
  int len;
  int top=0, bottom=0, left=0, right=0;
  int nb_target_letters;

//  printf("before\n");
//  print_board();

  if (!game_stack_size) {
    game_stack[0].cumul_score = score + starting_score;
  } else {
    game_stack[game_stack_size].cumul_score =
           score + game_stack[game_stack_size-1].cumul_score;
  }

  push_game();

  /* "play" on the board */
  /* board[i][j] = letter (A=1) or -10000000 if empty */
  /* is_joker[i][j] = 1 if letter at (i,j) is a joker */
  /* hand[k] = 1 (A) .. 26 (Z), or 0 for joker or 0x1000000 if no letter
   * + 0x1000000 if used
   */
  int new_letters = 0;
  for (i = 0; i < stacklen; i++) {
    if (board[stack[i].i][stack[i].j] == -10000000) new_letters++;
    if (stack[i].i == 0) top = 1;
    if (stack[i].i == 14) bottom = 1;
    if (stack[i].j == 0) left = 1;
    if (stack[i].j == 14) right = 1;
    board[stack[i].i][stack[i].j] = stack[i].l->l;
    if (stack[i].hl == 0) is_joker[stack[i].i][stack[i].j] = 1;
  }
  /* recompute hand */
  nb_input_letters = 0;
  nb_target_letters = 0;
  for (i = 0; i < 7; i++) {
    int c;
    /* 0x1100000 means letter cannot be played nor moved */
    if (hand[i] == 0x1100000) { nb_target_letters++; continue; }
    if (hand[i] < 0x1000000) { nb_input_letters++; continue; }
    c = eat_letter();
    if (c != '.' && c != ':') nb_input_letters++;
    if (c == '_') c = 0;                 /* joker */
    else if (c == '.') c = 0x1000000;  /* no letter, mark it as used */
    else if (c == ':') c = 0x1100000;  /* ununsable letter */
    else { if (!my_isalpha(c)) abort(); c = up(c); }
    hand[i] = c;
    if (c == 0x1100000) nb_target_letters++;
  }

  /* save played game in state */
  if (game_stack_size == 1) {
    game_stack[0].steps_size = 0;
  } else {
    game_stack[game_stack_size-1].steps_size =
           game_stack[game_stack_size-2].steps_size;
    memcpy(game_stack[game_stack_size-1].steps,
           game_stack[game_stack_size-2].steps,
           game_stack[game_stack_size-1].steps_size * sizeof(int));
  }
  memcpy(game_stack[game_stack_size-1].played_board, board, 15*15*sizeof(int));
  memcpy(game_stack[game_stack_size-1].played_is_joker, is_joker,
      15*15*sizeof(int));

  /* compute len of played word */
  for (len = 0; letters[len]; len++) /* do nothing */;
  add_step(&game_stack[game_stack_size-1], _starti, _startj, vert, len, score);

  current_total_score = total_score(game_stack[game_stack_size-1].cumul_score);
  if (current_total_score > best_score) {
    best_score = current_total_score;
    printf("current score %d (next_letter %d score %d)\n",
        best_score, next_letter, game_stack[game_stack_size-1].cumul_score);
    print_board();
    print_steps(&game_stack[game_stack_size-1]);
    printf("hand:");
    for (i = 0; i < 7; i++) printf(" %d", hand[i]);
    printf("\n");
    printf("init_hand:");
    for (i = 0; i < 7; i++) printf(" %d", init_hand[i]);
    printf("\n");
  }

  extern int border;
  int badplay=0;
  switch (border) {
  case 0: badplay=top; break;
  case 1: badplay=right; break;
  case 2: badplay=bottom; break;
  case 3: badplay=left; break;
  }

  //if (board[13][14] != -10000000) goto over;
  //if (board[12][14] != -10000000) goto over;
  //if (board[13][4] != -10000000) goto over;
  //if (board[13][3] != -10000000) goto over;
  //if (board[13][2] != -10000000) goto over;

  if (!badplay && nb_target_letters == 7) {
    print_board();
    print_steps(&game_stack[game_stack_size-1]);
    printf("DONE\n");
    abort();
  }

//  printf("after\n");
//  print_board();

  /* clear all variables */
  stacklen = 0;

  /* check oxyphenbutazone compatible */
#if 0
  if (board[0][14] != -10000000 &&  board[0][14]  != 'o'-'a'+1) goto over;
  if (board[1][14] != -10000000 &&  board[1][14]  != 'x'-'a'+1) goto over;
  if (board[2][14] != -10000000 &&  board[2][14]  != 'y'-'a'+1) goto over;
  if (board[3][14] != -10000000 &&  board[3][14]  != 'p'-'a'+1) goto over;
  if (board[4][14] != -10000000 &&  board[4][14]  != 'h'-'a'+1) goto over;
  if (board[5][14] != -10000000 &&  board[5][14]  != 'e'-'a'+1) goto over;
  if (board[6][14] != -10000000 &&  board[6][14]  != 'n'-'a'+1) goto over;
  if (board[7][14] != -10000000 &&  board[7][14]  != 'b'-'a'+1) goto over;
  if (board[8][14] != -10000000 &&  board[8][14]  != 'u'-'a'+1) goto over;
  if (board[9][14] != -10000000 &&  board[9][14]  != 't'-'a'+1) goto over;
  if (board[10][14] != -10000000 && board[10][14] != 'a'-'a'+1) goto over;
  if (board[11][14] != -10000000 && board[11][14] != 'z'-'a'+1) goto over;
  if (board[12][14] != -10000000 && board[12][14] != 'o'-'a'+1) goto over;
  if (board[13][14] != -10000000 && board[13][14] != 'n'-'a'+1) goto over;
  if (board[14][14] != -10000000 && board[14][14] != 'e'-'a'+1) goto over;
#endif

#if 0
  if (board[0][14] != -10000000 || board[7][14] != -10000000 ||
      board[14][14] != -10000000) {
    if (board[0][14] == -10000000 || board[7][14] == -10000000 ||
        board[14][14] == -10000000) goto over;
    print_steps(&game_stack[game_stack_size-1]);
    print_board(); abort();
  }
  if (board[0][0] != -10000000 || board[7][0] != -10000000 ||
      board[14][0] != -10000000) {
    if (board[0][0] == -10000000 || board[7][0] == -10000000 ||
        board[14][0] == -10000000) goto over;
    print_steps(&game_stack[game_stack_size-1]);
    print_board(); abort();
  }
  if (board[0][0] != -10000000 || board[0][7] != -10000000 ||
      board[0][14] != -10000000) {
    if (board[0][0] == -10000000 || board[0][7] == -10000000 ||
        board[0][14] == -10000000) goto over;
    print_steps(&game_stack[game_stack_size-1]);
    print_board(); abort();
  }
  if (board[14][0] != -10000000 || board[14][7] != -10000000 ||
      board[14][14] != -10000000) {
    if (board[14][0] == -10000000 || board[14][7] == -10000000 ||
        board[14][14] == -10000000) goto over;
    print_steps(&game_stack[game_stack_size-1]);
    print_board(); abort();
  }
#endif

#if 0
  /* on the left */
  for (i = 0; i < 15; i++)
    if (board[i][1] != -10000000 && board[i][0] == -10000000) goto over;
  /* on the right */
  for (i = 0; i < 15; i++)
    if (board[i][13] != -10000000 && board[i][14] == -10000000) goto over;
  /* on the top */
  for (i = 0; i < 15; i++)
    if (board[1][i] != -10000000 && board[0][i] == -10000000) goto over;
  /* on the bottom */
  for (i = 0; i < 15; i++)
    if (board[13][i] != -10000000 && board[14][i] == -10000000) goto over;
#endif

  switch (border) {
  case 0: if (top) goto over; break;
  case 1: if (right) goto over; break;
  case 2: if (bottom) goto over; break;
  case 3: if (left) goto over; break;
  }

#if 0
  switch(border) {
  case 3:
  /* on the left */
  for (i = 0; i < 15; i++)
    if (board[i][1] != -10000000 && board[i][0] == -10000000) goto over;
  break;
  case 1:
  /* on the right */
  for (i = 0; i < 15; i++)
    if (board[i][13] != -10000000 && board[i][14] == -10000000) goto over;
  break;
  case 0:
  /* on the top */
  for (i = 0; i < 15; i++)
    if (board[1][i] != -10000000 && board[0][i] == -10000000) goto over;
  break;
  case 2:
  /* on the bottom */
  for (i = 0; i < 15; i++)
    if (board[13][i] != -10000000 && board[14][i] == -10000000) goto over;
  break;
  }
#endif

#if 0
  if (top) {
    int w[15];
    int bl = 0;
    for (i = 0; i < 15; i++) { w[i] = board[0][i]; if(w[i]!=-10000000)bl++;}
//print_board(); printf("top found? %d\n", find15(w)); getchar();
    if (!find15(w)) goto over;
  }
  if (bottom) {
    int w[15];
    int bl = 0;
    for (i = 0; i < 15; i++) { w[i] = board[14][i]; if(w[i]!=-10000000)bl++;}
//print_board(); printf("bottom found? %d\n", find15(w)); getchar();
    if (!find15(w)) goto over;
  }
  if (left) {
    int w[15];
    int bl = 0;
    for (i = 0; i < 15; i++) { w[i] = board[i][0]; if(w[i]!=-10000000)bl++;}
//print_board(); printf("left found? %d\n", find15(w)); getchar();
    if (!find15(w)) goto over;
  }
  if (right) {
    int w[15];
    int bl = 0;
    for (i = 0; i < 15; i++) { w[i] = board[i][14]; if(w[i]!=-10000000)bl++;}
//print_board(); printf("right found? %d\n", find15(w)); getchar();
    if (!find15(w)) goto over;
  }
#endif

//  if (next_letter > 40) goto over;
//  if (new_letters < 4 - 2*(top + left + bottom + right)) goto over;

  /* run next hand */
  inci = 0; incj = 1;
  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) {
    starti = i; startj = j;
    /* must be a start of a word (no letter before) */
    if (is_letter(starti-inci, startj-incj)) continue;
    is_first = 0;
    go(dict.top, i, j);
  }
  inci = 1; incj = 0;
  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) {
    starti = i; startj = j;
    /* must be a start of a word (no letter before) */
    if (is_letter(starti-inci, startj-incj)) continue;
    is_first = 0;
    go(dict.top, i, j);
  }

over:
  pop_game();
}

static int bagsize;

void preplay(char *moves, char *bag)
{
  int i;
  int score = 0, move_score;

  if (strlen(bag) > 199) abort();
  strcpy(letters, bag);
  translate_letters();

  bagsize = strlen(bag);

  for (i = 0; i < 7; i++) init_hand[i] = bag[i]=='?'?0:bag[i]-'A'+1;
  next_letter = 7;

  while (*moves) {
    char *start = moves;
    char a;
    char w[16];
    int row, col, vert;
    char colc;
    a = *moves++;
    while (*moves != ' ') moves++; moves++;
    i = 0; while (*moves != ' ') w[i++] = *moves++;
    w[i] = 0;
    sscanf(moves, "%d", &move_score); score += move_score;
    while (*moves && *moves != ',') moves++; if (*moves) moves++;
    if (a >= '0' && a <= '9') {
      vert = 0;
      sscanf(start, "%d%c", &row, &colc);
      col = colc - 'A';
      row--;
    } else {
      vert = 1;
      sscanf(start, "%c%d", &colc, &row);
      col = colc - 'A';
      row--;
    }
    printf("%s at (%d, %d) %s\n", w, row, col, vert?"vert":"horiz");
    i = 0;
    while (w[i]) {
      if (board[row][col] == -10000000) {
        int k;
        int val=w[i]>='a'&&w[i]<='z'?0:w[i]-'A'+1;
        for (k = 0; k < 7; k++) if (val == init_hand[k]) {
          init_hand[k] = bag[next_letter]=='?'?0:bag[next_letter]-'A'+1;
          break;
        }
        if (k == 7) abort();
        next_letter++; if (bag[next_letter] == 0) abort();
      }
      if (w[i] >= 'a' && w[i] <= 'z') {
        board[row][col] = w[i] - 'a' + 1;
        is_joker[row][col] = 1;
      } else {
        board[row][col] = w[i] - 'A' + 1;
      }
      if (vert) row++; else col++;
      i++;
    }
  }
  printf("next_letter %d score %d\n", next_letter, score);
  printf("init hand: ");
  for (i = 0; i < 7; i++) printf("%c", init_hand[i]?init_hand[i]-1+'A':'?');
  printf("\n");
  print_board();
  starting_score = score;

  for (i = 0; i < 7; i++) init_hand_saved[i] = init_hand[i];
}

static int target_hand[7];
static int busy[200];

int next_hand(void)
{
  int l = 6;
  int k;
  int pos;

next:
  if (l == -1) return 0;

  pos = bagpos[l];
  busy[pos] = 0;
  pos++;
  for (; pos < bagsize; pos++) {
    if (busy[pos]) continue;
    if (!(/*translated_letters[pos] == 0 ||*/
          translated_letters[pos] == target_hand[l])) continue;
    busy[pos] = 1;
    bagpos[l] = pos;
    break;
  }
  if (pos == bagsize) { l--; goto next; }

  l++;
  for (k = l; k < 7; k++) { /*busy[bagpos[k]] = 0;*/ bagpos[k] = next_letter-7-1; }

  if (l != 7) goto next;

  /* remove unusable letters already in hand */
  for (k = 0; k < 7; k++)
    if (busy[next_letter-7+k]) init_hand[k] = 0x1100000;
    else                       init_hand[k] = init_hand_saved[k];

  return 1;
}

void first_hand(char *_target_hand)
{
  int i;
  int pos = next_letter-7; if (pos<0) abort();

  /* put current hand "back" in bag at position pos */
  for (i = 0; i < 7; i++) translated_letters[pos+i] = init_hand[i];

  /* set target_hand */
  if (strlen(_target_hand) != 7) abort();
  for (i = 0; i < 7; i++) {
    if (_target_hand[i] < 'A' || _target_hand[i] > 'Z') abort();
    target_hand[i] = _target_hand[i] -'A'+1;
  }

  /* set init pos to pos-1 and call next_hand until all hand decided */
  for (i = 0; i < 7; i++) bagpos[i] = pos-1;
  while (1) {
    /* check if all hand decided */
    for (i = 0; i < 7; i++) if (bagpos[i] == pos-1) break;
    if (i == 7) break;
    if (!next_hand()) { printf("no possible\n"); abort(); }
  }

#if 0
  /* get first possible position for each letter in target hand */
  for (i = 0; i < 7; i++) {
    for (k = pos; k < bagsize; k++) {
      if (busy[k]) continue;
      if (translated_letters[k] != target_hand[i]) continue;
      bagpos[i] = k;
      busy[k] = 1;
      break;
    }
    if (k == bagsize) {
      /* no free letter, try a joker */
      for (k = pos; k < bagsize; k++) {
        if (busy[k]) continue;
        if (translated_letters[k] != 0) continue;
        bagpos[i] = k;
        busy[k] = 1;
        break;
      }
      if (k == bagsize) { printf("no hand possible\n"); abort(); }
    }
  }

  /* remove unusable letters already in hand */
  for (i = 0; i < 7; i++) if (busy[pos+i]) init_hand[i] = 0x1100000;
#endif
}

void print_first_next(void)
{
  int i;
  printf("busy:");
  for (i = 0; i < 200; i++) if (busy[i]) printf(" %d", i);
  printf(" ih:");
  for (i = 0; i < 7; i++) printf(" %2.2d", init_hand[i]);
  printf("\n");
  printf("          bagpos:");
  for (i = 0; i < 7; i++) printf(" %2.2d", bagpos[i]);
  printf("\n");
}
