#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "common/hole15.h"

typedef int cut[15];

typedef struct dict_letter {
  struct dict_letter *next;
  struct dict_letter *child;
  int l;
  int w;
} dict_letter;

typedef struct {
  dict_letter *top;
} dictionary;

dictionary dict;

int utf8(unsigned char *n, int *_len)
{
  int v = 0;
  if (*n == 0xc0 || *n == 0xc1 ||
      *n >= 0xf5) abort();
  if (!(*n & 0x80)) {
    v = *n;
    *_len = 1;
  } else
  if ((*n & 0xe0) == 0xc0) { /* 110xxxxx */
    if ((*(n+1) & 0xc0) != 0x80) abort();
    v = ((*n & 0x1f) << 6) | (*(n+1) & 0x3f);
    if (v < 0x80 || v > 0x7ff) abort();
    *_len = 2;
  } else
  if ((*n & 0xf0) == 0xe0) { /* 1110xxxx */
    if ((*(n+1) & 0xc0) != 0x80 ||
        (*(n+2) & 0xc0) != 0x80) abort();
    v = ((*n & 0x0f) << (6*2)) | ((*(n+1) & 0x3f) << 6) | (*(n+2) & 0x3f);
    if (v < 0x800 || v > 0xffff) abort();
    *_len = 3;
  } else
  if ((*n & 0xf8) == 0xf0) { /* 11110xxx */
    if ((*(n+1) & 0xc0) != 0x80 ||
        (*(n+2) & 0xc0) != 0x80 ||
        (*(n+3) & 0xc0) != 0x80) abort();
    v = ((*n & 0x07) << (6*3)) | ((*(n+1) & 0x3f) << (6*2)) |
        ((*(n+2) & 0x3f) << 6) | (*(n+3) & 0x3f);
    if (v < 0x10000 || v > 0x10ffff) abort();
    *_len = 4;
  } else abort();
  return v;
}

int up(int n)
{
  if (isascii(n)) return toupper(n)-'A'+1;
  abort();
}

static void go(dict_letter *l, int *w)
{
  dict_letter *t;

  if (*w == 0) return;

more:
  if (l->l == *w) {
    if (w[1] == 0) { l->w = 1; return; }
    if (l->child) { go(l->child, w+1); return; }
    w++;
    goto suffix;
  }
  if (l->next) { l = l->next; goto more; }

  t = l->next = malloc(sizeof(dict_letter)); if (!t) abort();

next:
  t->next = 0;
  t->child = 0;
  t->l = *w;
  t->w = w[1] ? 0 : 1;
  l = t;
  w++;
  if (!*w) return;
suffix:
  t = l->child = malloc(sizeof(dict_letter)); if (!t) abort();
  goto next;
}

void add_word(dictionary *d, int *w)
{
  go(d->top, w);
}

void new_dict(dictionary *d)
{
  d->top = malloc(sizeof(dict_letter)); if (!d->top) abort();
  d->top->next = 0;
  d->top->child = 0;
  d->top->l = -1;
  d->top->w = 0;
}

static int find(dict_letter *l, int *w)
{
  if (*w == 0) return 0;
  if (l == NULL) return 0;
next:
  if (l->l == *w) {
    if (l->w && w[1] == 0) return 1;
    return find(l->child, w+1);
  }
  l = l->next;
  if (l != NULL) goto next;
  return 0;
}

int find_word(dictionary *d, int *w)
{
//printf("yo %s %d\n", get_word(w), find(d->top, w));
  return find(d->top, w);
}

void read_dict(char *fn, dictionary *dict, int reverse)
{
  FILE *f = fopen(fn, "r"); if (!f) { perror(fn); abort(); }
  while (1) {
    unsigned char n[128];
    int ni[128];
    int pos;
    int i;
    int len;
    int wordlen;
    if (!fgets((char*)n, 128, f)) break;
    for (i = 0; i < 127 && n[i]; i++) if (isspace(n[i])) { n[i] = 0; break; }
    n[i] = 0;
    if ((len=wordlen=strlen((char*)n)) == 0 || len > 15) continue;
    if (reverse)
      for (i = 0, pos = 0; n[pos]; i++, pos+=len)
        ni[i] = utf8(n+wordlen-1-pos, &len);
    else
      for (i = 0, pos = 0; n[pos]; i++, pos+=len) ni[i] = utf8(n+pos, &len);
    ni[i] = 0;
    for (i = 0; ni[i]; i++) ni[i] = up(ni[i]);
    ni[i] = 0;
    add_word(dict, ni);
  }
  fclose(f);
  return;
}

int hands[26][100];

int (*words)[15];
int words_size;

int fail(void) { abort(); }
void read_words(char *fn)
{
  char l[256];
  int i;
  FILE *f = fopen(fn, "r"); if (!f) { perror(fn); exit(1); }
  while (1) {
    if (fgets(l, 256, f) == NULL) break;
    words_size++;
    words = realloc(words, words_size * 15 * sizeof(int));
    if (words == NULL) abort();
    for (i = 0; i<15; i++) words[words_size-1][i] =
      l[i] >= 'a' && l[i] <= 'z' ? l[i] - 'a' + 1 :
      l[i] >= 'A' && l[i] <= 'Z' ? l[i] - 'A' + 1 :
      fail();
  }
  fclose(f);
}

void read_hands(char *fn)
{
  FILE *f = fopen(fn, "r"); if (!f) { perror(fn); exit(1); }
  int i;
  for (i = 0; i < 26; i++) {
    int c;
    int j;
    for (j = 0; j < 100; j++) {
      c = fgetc(f); if (c == EOF) abort();
      hands[i][j] = c == '?' ? 0 : c-'A'+1;
    }
    fgetc(f);
  }
  fclose(f);
}

static cut *cutword;
static int cutword_size;
static int no_double;

void addword(int w[15])
{
  cutword_size++;
  cutword = realloc(cutword, cutword_size * sizeof(cut));
  if (cutword == NULL) abort();
  memcpy(cutword[cutword_size-1], w, 15*sizeof(int));
}

int check_word(int *dw)
{
  int w[16];
  int i, p = 0;
  if (dw[0] || dw[7] || dw[14]) return 0;
  if (no_double == 0) {
    if (dw[3]) return 0;
    if (dw[11]) return 0;
  }
#if 0
  if (dw[1]) return 0;
  if (dw[2]) return 0;
#endif
  for (i = 0; i < 15; i++) {
    w[p] = dw[i];
    if (dw[i] == 0) {
      if (p > 1 && !find_word(&dict, w)) return 0;
      p = -1;
    }
    p++;
  }
  w[p] = 0;
  if (p > 1 && !find_word(&dict, w)) return 0;
  return 1;
}

void print_dword(int *dw)
{
  int i;
  for (i = 0; i < 15; i++)
    printf("%c", dw[i] ? dw[i]+'A'-1 : ' ');
  printf("\n");
}

void doword(int *w)
{
  int dw[15];
  int i, j, k;
  //printf("do "); print_dword(w);
  for (i = 0; i < 7; i++) for (j = 0; j < _15[i].len; j++) {
    for (k = 0; k < 15; k++)
      if ((*_15[i].h)[j][k] == 0) dw[k] = w[k]; else dw[k] = 0;
    if (check_word(dw)) { /*printf("%d: ", i+1); print_dword(dw);*/ addword(dw); }
  }
}

cut *getcut(int word[15], int *retlen, int _no_double)
{
  cut *ret;
  no_double = _no_double;
  if (cutword_size) free(cutword);
  cutword_size = 0;
  cutword = NULL;
  doword(word);
  *retlen = cutword_size;
  ret = malloc(cutword_size * sizeof(cut)); if (!ret) abort();
  memcpy(ret, cutword, cutword_size * sizeof(cut));
  return ret;
}

void pwc(int *w)
{
  int i;
  for (i = 0; i < 15; i++) printf("%c", w[i] ? w[i] + 'a' -1 : '.');
}

int compatible(int *word1, int *cut1, int *word2, int *cut2, int *hand)
{
  int used[100];
  int bagpos1[15];
  int bagpos2[15];
  int word1min = 200;
  int word2min = 200;
  int word1max = -200;
  int word2max = -200;
  int cut1min = 200;
  int cut2min = 200;
  int cut1max = -200;
  int cut2max = -200;
  int i, j;
  int letter;
  memset(used, 0, sizeof(int) * 100);
  /* all word1 must come before last move of word2 (the cut) */
  /* get farest play for the cut of word2 */
  /* jokers not used, should we? */
  for (i = 0; i < 15; i++) {
    if (cut2[i]) continue;
    /* get last unused letter */
    letter = word2[i];
    for (j = 99; j >= 0; j--) {
      if (hand[j] != letter) continue;
      if (used[j]) continue;
      used[j] = 1;
      bagpos2[i] = j;
      if (j < cut2min) cut2min = j;
      if (j > cut2max) cut2max = j;
      break;
    }
    if (j == -1) abort();
  }
  /* try to put the cut of word1 *before* cut2min */
  for (i = 0; i < 15; i++) {
    if (cut1[i]) continue;
    /* get last unused letter */
    letter = word1[i];
    for (j = cut2min-1; j >= 0; j--) {
      if (hand[j] != letter) continue;
      if (used[j]) continue;
      used[j] = 1;
      bagpos1[i] = j;
      if (j < cut1min) cut1min = j;
      if (j > cut1max) cut1max = j;
      break;
    }
    if (j == -1) return 0;
  }
  //if (cut2min < 80) return 0;
  if (cut1min < 80) return 0;
  /* try to put all remaining letters of word2 *before* its cut */
  for (i = 0; i < 15; i++) {
    letter = cut2[i];
    if (!letter) continue;
    for (j = cut2min-1; j >= 0; j--) {
      if (hand[j] != letter) continue;
      if (used[j]) continue;
      used[j] = 1;
      bagpos2[i] = j;
      if (j < word2min) word2min = j;
      if (j > word2max) word2max = j;
      break;
    }
    if (j == -1) return 0;
  }
#if 0
printf("cut2: "); pwc(cut2); printf("\n");
printf("cut1: "); pwc(cut1); printf("\n");
printf("%c\n", hand[cut2min] + 'A' - 1);
printf("word 2 %d %d - cut %d %d\n", word2min, word2max, cut2min, cut2max);
#endif

#if 0
printf("\n");
printf("cut2: "); pwc(cut2); printf("\n");
printf("cut1: "); pwc(cut1); printf("\n");
printf("%c\n", hand[cut2min] + 'A' - 1);
printf("word 2 %d %d - cut %d %d\n", word2min, word2max, cut2min, cut2max);
printf("word 1 %d %d - cut %d %d\n", word1min, word1max, cut1min, cut1max);
#endif
  /* try to put all letters of word1 *before* its cut */
  for (i = 0; i < 15; i++) {
    letter = cut1[i];
    if (!letter) continue;
    for (j = cut1min-1; j >= 0; j--) {
      if (hand[j] != letter) continue;
      if (used[j]) continue;
      used[j] = 1;
      bagpos1[i] = j;
      if (j < word1min) word1min = j;
      if (j > word1max) word1max = j;
      break;
    }
    if (j == -1) return 0;
  }
#define min(a,b) ((a)<(b)?(a):(b))
#define max(a,b) ((a)>(b)?(a):(b))
if (word1max > min(cut1min, cut2min)) return 0;
if (word2max > min(cut1min, cut2min)) return 0;
//  if (min(word1min,word2min) < 25 || max(word1min, word2min) < 30) return 0;
printf("\n  word1 [%d %d] cut1 [%d %d] word2 [%d %d] cut2 [%d %d]\n  ",
    word1min, word1max, cut1min, cut1max,
    word2min, word2max, cut2min, cut2max);
pwc(word1); printf("\n  ");
pwc(word2); printf("\n  ");
pwc(cut1); printf("\n  ");
pwc(cut2); printf("\n");

printf("  "); pwc(word1); printf(" "); pwc(word2); printf(" "); pwc(cut1);
printf(" "); pwc(cut2);
for (i = 0; i < 15; i++) printf(" %d", bagpos1[i] + (cut1[i] ? 0 : 0));
printf("    ");
for (i = 0; i < 15; i++) printf(" %d", bagpos2[i] + (cut2[i] ? 0 : 0));
printf("\n");
printf("  "); pwc(word2); printf(" "); pwc(word1); printf(" "); pwc(cut2);
printf(" "); pwc(cut1);
for (i = 0; i < 15; i++) printf(" %d", bagpos2[i] + (cut2[i] ? 0 : 0));
printf("    ");
for (i = 0; i < 15; i++) printf(" %d", bagpos1[i] + (cut1[i] ? 0 : 0));
printf("\n");
  return 1;
}

int putword(int *lcount, int *word)
{
  int i;
  for (i = 0; i < 15; i++)
    { if (lcount[word[i]] == 0) return 0; lcount[word[i]]--; }
  return 1;
}

int main(int n, char **v)
{
  int w1, w2, h;
  cut **wcut;
  int *wcut_size;
  cut *w1cut, *w2cut;
  int w1cut_size, w2cut_size;
  int w1c, w2c;
  int k;

  if (n!=4)
    { printf("gimme <word list> <hand list> <dictionnary>\n"); return 1; }
  read_words(v[1]);
  read_hands(v[2]);
  new_dict(&dict);
  read_dict(v[3], &dict, 0);

  printf("word list size: %d\n", words_size);

  wcut = malloc(words_size * sizeof(cut *)); if (wcut==NULL) abort();
  wcut_size = malloc(words_size * sizeof(int)); if (wcut_size==NULL)abort();

  for (k = 0; k < words_size; k++)
    wcut[k] = getcut(words[k], &wcut_size[k], 1);

  for (h = 0; h < 26; h++) {
    printf("hand %c\n", hands[h][0] + 'A' - 1);
    for (w1 = 0; w1 < words_size; w1++) for (w2 = w1; w2 < words_size; w2++) {
      int lcount[27]={ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 };
      for (k = 0; k < 100; k++) lcount[hands[h][k]]++;
      if (!putword(lcount, words[w1])) continue;
      if (!putword(lcount, words[w2])) continue;

      w1cut = wcut[w1]; //getcut(words[w1], &w1cut_size, 1);
      w1cut_size = wcut_size[w1];
      w2cut = wcut[w2]; //getcut(words[w2], &w2cut_size, 1);
      w2cut_size = wcut_size[w2];

      for (w1c=0; w1c<w1cut_size; w1c++) for (w2c=0; w2c<w2cut_size; w2c++) {
        if (compatible(words[w1],w1cut[w1c],words[w2],w2cut[w2c],hands[h])) {
#if 0
          printf("%c ", hands[h][0] + 'A' - 1);
          pwc(words[w1]);
          printf(" ");
          pwc(words[w2]);
          printf(" '");
          pwc(w1cut[w1c]);
          printf("' '");
          pwc(w2cut[w2c]);
          printf("'\n");
#endif
        }
        if (compatible(words[w2],w2cut[w2c],words[w1],w1cut[w1c],hands[h])) {
#if 0
          printf("%c ", hands[h][0] + 'A' - 1);
          pwc(words[w2]);
          printf(" ");
          pwc(words[w1]);
          printf(" '");
          pwc(w2cut[w2c]);
          printf("' '");
          pwc(w1cut[w1c]);
          printf("'\n");
#endif
        }
      }

      //if (w1cut != NULL) free(w1cut);
      //if (w2cut != NULL) free(w2cut);
    }
  }
  return 0;
}
