#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "dict.h"
#include "scrabble.h"

typedef struct {
  dict_letter *l;
  int bagpos;
//  int hl;         /* hand's value (to check for jokers) */
  int i, j;
} stack_elem;

typedef struct {
  /* from main.c */
  int is_first;
  int board[15][15];
  int is_joker[15][15];
  int hand[7];
  int starti, startj, inci, incj;
  stack_elem stack[15];
  int stacklen;
  /* from contest.c */
  int next_letter;
  /* internal to this structure (not restored on pop) */
  int cumul_score;
  int steps[100];  // access with STEP_XX defined below
  int steps_size;
  int played_board[15][15];
  int played_is_joker[15][15];
} game_state;

#define G_STEP_I(x)        ( (x)       & 0x0f)
#define G_STEP_J(x)        (((x) >> 4) & 0x0f)
#define G_STEP_LEN(x)      (((x) >> 8) & 0x0f)
#define G_STEP_VERTICAL(x) (((x) >> 12) & 0x01)
#define G_STEP_SCORE(x)    (((x) >> 16) & 0xffff)

#define S_STEP_I(x, v)        x = ((x) &   ~0x0f) |  (v)
#define S_STEP_J(x, v)        x = ((x) &   ~0xf0) | ((v) << 4)
#define S_STEP_LEN(x, v)      x = ((x) &  ~0xf00) | ((v) << 8)
#define S_STEP_VERTICAL(x, v) x = ((x) & ~0x1000) | ((v) << 12)
#define S_STEP_SCORE(x, v)    x = ((x) & ~0xffff0000) | ((v) << 16)

static char letters[200];
int translated_letters[200];
int next_letter;

game_state game_stack[1024];
int game_stack_size;

extern int is_first;
extern int board[15][15];
extern int is_joker[15][15];
extern int hand[7];
extern int starti, startj, inci, incj;
extern stack_elem stack[15];
extern int stacklen;
extern int nb_input_letters;

int my_isalpha(int n);
int up(int n);

void add_step(game_state *g, int i, int j, int inci, int len, int score)
{
  if (g->steps_size == 100) abort();
  g->steps[g->steps_size] = 0;
  S_STEP_I(g->steps[g->steps_size], i);
  S_STEP_J(g->steps[g->steps_size], j);
  S_STEP_VERTICAL(g->steps[g->steps_size], inci);
  S_STEP_LEN(g->steps[g->steps_size], len);
  S_STEP_SCORE(g->steps[g->steps_size], score);
  g->steps_size++;
}

void translate_letters(void)
{
  int i;
  for (i = 0; letters[i]; i++) {
    int c = letters[i];
    if (c == '?') c = 0;                 /* joker */
    else { if (!my_isalpha(c)) abort(); c = up(c); }
    translated_letters[i] = c;
  }
}

void push_game(void)
{
  game_state *g;
  if (game_stack_size == 1024) abort();
  g = &game_stack[game_stack_size];
  game_stack_size++;
  g->is_first = is_first;
  memcpy(g->board, board, 15*15*sizeof(int));
  memcpy(g->is_joker, is_joker, 15*15*sizeof(int));
  memcpy(g->hand, hand, 7*sizeof(int));
  g->starti = starti;
  g->startj = startj;
  g->inci = inci;
  g->incj = incj;
  memcpy(g->stack, stack, 15*sizeof(stack_elem));
  g->stacklen = stacklen;
  g->next_letter = next_letter;
}

void pop_game(void)
{
  game_state *g;
  if (game_stack_size == 0) abort();
  game_stack_size--;
  g = &game_stack[game_stack_size];
  is_first = g->is_first;
  memcpy(board, g->board, 15*15*sizeof(int));
  memcpy(is_joker, g->is_joker, 15*15*sizeof(int));
  memcpy(hand, g->hand, 7*sizeof(int));
  starti = g->starti;
  startj = g->startj;
  inci = g->inci;
  incj = g->incj;
  memcpy(stack, g->stack, 15*sizeof(stack_elem));
  stacklen = g->stacklen;
  next_letter = g->next_letter;
}

extern dictionary dict;

int is_letter(int i, int j);
void go(dict_letter *cur, int i, int j);

int eat_letter(void)
{
  if (letters[next_letter] == 0) return 0x1000000;
  next_letter++;
  return next_letter-1;
}

void init_contest(char *bag)
{
  if (strlen(bag) > 199) abort();
  strcpy(letters, bag);
  translate_letters();
}

void contest(void)
{
  int i, j;

  for (i = 0; i < 7; i++) hand[i] = eat_letter();
  /* we suppose a bag of at least 7 letters */
  nb_input_letters = 7;

  inci = 0; incj = 1;
  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) {
    starti = i; startj = j;
    /* must be a start of a word (no letter before) */
    if (is_letter(starti-inci, startj-incj)) continue;
    is_first = 1;
    go(dict.top, i, j);
  }
}

int best_score = -10000000;

int total_score(int score)
{
  extern letter english_letters[];
  int i;
  for (i = 0; i < 7; i++)
    if (hand[i] < 0x1000000)
      score -= english_letters[translated_letters[hand[i]]].value;
  for (i = next_letter; letters[i]; i++)
    score -= english_letters[translated_letters[i]].value;
  return score;
}

void print_board(void)
{
  int i, j;
  for (i = 0; i < 15; i++) {
    for (j = 0; j < 15; j++)
      if (is_joker[i][j])
        printf("%c", board[i][j]!=-10000000 ? 'A' + ((board[i][j]>>8)&255)-1 : '.');
      else
        printf("%c", board[i][j]!=-10000000 ? 'A' + translated_letters[board[i][j]-1]-1 : '.');
    printf("\n");
  }
}

void print_steps(game_state *g)
{
  int i, j;
  for (i = 0; i < g->steps_size; i++) {
    int row = G_STEP_I(g->steps[i]);
    int col = G_STEP_J(g->steps[i]);
    int len = G_STEP_LEN(g->steps[i]);
    int vert = G_STEP_VERTICAL(g->steps[i]);
    int score = G_STEP_SCORE(g->steps[i]);
    int incrow, inccol;
    if (!vert) {
      printf("%d%c ", row+1, col+'A');
      incrow = 0;
      inccol = 1;
    } else {
      printf("%c%d ", col+'A', row+1);
      incrow = 1;
      inccol = 0;
    }
    for (j = 0; j < len; j++, row += incrow, col += inccol) {
      int is_joker = g->played_is_joker[row][col];
      int c;
      if (is_joker)
        c = (g->played_board[row][col]>>8)&255;
      else
        c = translated_letters[g->played_board[row][col]-1];
      printf("%c", c+'A'-1 + is_joker * ('a'-'A'));
    }
    printf(" %d", score);
    if (i == g->steps_size-1) printf("\n"); else printf(",");
  }
}

void put_word(int score, int _starti, int _startj, int vert, int *letters)
{
  int i, j;
  int current_total_score;
  int len;

//  printf("before\n");
//  print_board();

  if (!game_stack_size) {
    game_stack[0].cumul_score = score;
  } else {
    game_stack[game_stack_size].cumul_score =
           score + game_stack[game_stack_size-1].cumul_score;
  }

  push_game();

  /* "play" on the board */
  /* board[i][j] = position in bag+1, includes chosen letter<<8 if joker
   * or -10000000 if empty
   */
  /* is_joker[i][j] = 1 if letter at (i,j) is a joker */
  /* hand[k] = position in bag or 0x1000000 if no letter
   * + 0x1000000 if used
   */
  for (i = 0; i < stacklen; i++) {
    board[stack[i].i][stack[i].j] = stack[i].bagpos+1;
    if (translated_letters[stack[i].bagpos] == 0) {
      is_joker[stack[i].i][stack[i].j] = 1;
      board[stack[i].i][stack[i].j] |= stack[i].l->l << 8;
    }
  }
  /* recompute hand */
  nb_input_letters = 0;
  for (i = 0; i < 7; i++) {
    int c;
    if (hand[i] < 0x1000000) { nb_input_letters++; continue; }
    c = eat_letter();
    if (c != 0x1000000) nb_input_letters++;
    hand[i] = c;
  }

  /* save played game in state */
  if (game_stack_size == 1) {
    game_stack[0].steps_size = 0;
  } else {
    game_stack[game_stack_size-1].steps_size =
           game_stack[game_stack_size-2].steps_size;
    memcpy(game_stack[game_stack_size-1].steps,
           game_stack[game_stack_size-2].steps,
           game_stack[game_stack_size-1].steps_size * sizeof(int));
  }
  memcpy(game_stack[game_stack_size-1].played_board, board, 15*15*sizeof(int));
  memcpy(game_stack[game_stack_size-1].played_is_joker, is_joker,
      15*15*sizeof(int));

  /* compute len of played word */
  for (len = 0; letters[len]; len++) /* do nothing */;
  add_step(&game_stack[game_stack_size-1], _starti, _startj, vert, len, score);

  current_total_score = total_score(game_stack[game_stack_size-1].cumul_score);
  if (current_total_score > best_score) {
    best_score = current_total_score;
    printf("current score %d (next_letter %d score %d)\n", best_score, next_letter, game_stack[game_stack_size-1].cumul_score);
    print_board();
    print_steps(&game_stack[game_stack_size-1]);
  }

//  printf("after\n");
//  print_board();

  /* clear all variables */
  stacklen = 0;

  /* run next hand */
  inci = 0; incj = 1;
  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) {
    starti = i; startj = j;
    /* must be a start of a word (no letter before) */
    if (is_letter(starti-inci, startj-incj)) continue;
    is_first = 0;
    go(dict.top, i, j);
  }
  inci = 1; incj = 0;
  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) {
    starti = i; startj = j;
    /* must be a start of a word (no letter before) */
    if (is_letter(starti-inci, startj-incj)) continue;
    is_first = 0;
    go(dict.top, i, j);
  }

  pop_game();
}
