#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include "dict.h"
#include "scrabble.h"

typedef struct {
  dict_letter *l;
  int hl;         /* hand's value (to check for jokers) */
  int i, j;
} stack_elem;

typedef struct {
  /* solver state, from main.c */
  int board[15][15];
  int is_joker[15][15];
  int hand[7];
  int starti, startj, inci, incj;
  stack_elem stack[15];
  int stacklen;
  int nb_input_letters;
  /* local state, from contest.c */
  int next_letter;
  int cumul_score;
  int steps[100];
  int steps_size;
  int wordscore;
  int border_letters;
} game_state;

#define G_STEP_I(x)        ( (x)       & 0x0f)
#define G_STEP_J(x)        (((x) >> 4) & 0x0f)
#define G_STEP_LEN(x)      (((x) >> 8) & 0x0f)
#define G_STEP_VERTICAL(x) (((x) >> 12) & 0x01)
#define G_STEP_SCORE(x)    (((x) >> 16) & 0xffff)

#define S_STEP_I(x, v)        x = ((x) &   ~0x0f) |  (v)
#define S_STEP_J(x, v)        x = ((x) &   ~0xf0) | ((v) << 4)
#define S_STEP_LEN(x, v)      x = ((x) &  ~0xf00) | ((v) << 8)
#define S_STEP_VERTICAL(x, v) x = ((x) & ~0x1000) | ((v) << 12)
#define S_STEP_SCORE(x, v)    x = ((x) & ~0xffff0000) | ((v) << 16)

static char letters[200];
static int translated_letters[200];

/* local state */
static int next_letter;
static int cumul_score;
int steps[100];
int steps_size;
int wordscore;
int border_letters;

game_state game_stack[1024];
int game_stack_size;

/* solver state */
extern int board[15][15];
extern int is_joker[15][15];
extern int hand[7];
extern int starti, startj, inci, incj;
extern stack_elem stack[15];
extern int stacklen;
extern int nb_input_letters;

int my_isalpha(int n);
int up(int n);

void add_step(int i, int j, int inci, int len, int score)
{
  if (steps_size == 100) abort();
  steps[steps_size] = 0;
  S_STEP_I(steps[steps_size], i);
  S_STEP_J(steps[steps_size], j);
  S_STEP_VERTICAL(steps[steps_size], inci);
  S_STEP_LEN(steps[steps_size], len);
  S_STEP_SCORE(steps[steps_size], score);
  steps_size++;
}

void translate_letters(void)
{
  int i;
  for (i = 0; letters[i]; i++) {
    int c = letters[i];
    if (c == '?') c = 0;                 /* joker */
    else { if (!my_isalpha(c)) abort(); c = up(c); }
    translated_letters[i] = c;
  }
}

void push_game(void)
{
  game_state *g;
  if (game_stack_size == 1024) abort();
  g = &game_stack[game_stack_size];
  game_stack_size++;
  /* solver state */
  memcpy(g->board, board, 15*15*sizeof(int));
  memcpy(g->is_joker, is_joker, 15*15*sizeof(int));
  memcpy(g->hand, hand, 7*sizeof(int));
  g->starti = starti;
  g->startj = startj;
  g->inci = inci;
  g->incj = incj;
  memcpy(g->stack, stack, 15*sizeof(stack_elem));
  g->stacklen = stacklen;
  g->nb_input_letters = nb_input_letters;
  /* local state */
  g->next_letter = next_letter;
  g->cumul_score = cumul_score;
  memcpy(g->steps, steps, 100*sizeof(int));
  g->steps_size = steps_size;
  g->wordscore = wordscore;
  g->border_letters = border_letters;
}

void pop_game(void)
{
  game_state *g;
  if (game_stack_size == 0) abort();
  game_stack_size--;
  g = &game_stack[game_stack_size];
  /* solver state */
  memcpy(board, g->board, 15*15*sizeof(int));
  memcpy(is_joker, g->is_joker, 15*15*sizeof(int));
  memcpy(hand, g->hand, 7*sizeof(int));
  starti = g->starti;
  startj = g->startj;
  inci = g->inci;
  incj = g->incj;
  memcpy(stack, g->stack, 15*sizeof(stack_elem));
  stacklen = g->stacklen;
  nb_input_letters = g->nb_input_letters;
  /* local state */
  next_letter = g->next_letter;
  cumul_score = g->cumul_score;
  memcpy(steps, g->steps, 100*sizeof(int));
  steps_size = g->steps_size;
  wordscore = g->wordscore;
  border_letters = g->border_letters;
}

extern dictionary dict;

int is_letter(int i, int j);
void read_hand(char *h);
void go(dict_letter *cur, int i, int j);

char eat_letter(void)
{
  char r = letters[next_letter];
  if (!r) return '.';
  next_letter++;
  if (r == '?') return '_';
  return r;
}

/* list to store found words by solver */
#define MAXLIST 500

static game_state list[MAXLIST];
static game_state _list[MAXLIST];
static int list_size;

int cmp_list(const void *_a, const void *_b)
{
  const game_state *a = _a;
  const game_state *b = _b;

  if (a->border_letters < b->border_letters) return -1;
  if (a->border_letters > b->border_letters) return 1;
  if (a->next_letter > b->next_letter) return -1;
  if (a->next_letter < b->next_letter) return 1;
  if (a->cumul_score < b->cumul_score) return -1;
  if (a->cumul_score > b->cumul_score) return 1;
return 0;

  /* order by cumul_score */
  if (a->cumul_score < b->cumul_score) return -1;
  if (a->cumul_score > b->cumul_score) return 1;
  /* other metrics? */
  return 0;
}

void sort_list(void)
{
  qsort(list, list_size, sizeof(game_state), cmp_list);
}

void put_in_list(void)
{
  if (list_size == MAXLIST) {
    int i;
    int min;
    push_game();
    min = 0;
    for (i = 1; i < MAXLIST; i++)
      if (cmp_list(&list[i], &list[min]) < 0) min = i;
    if (cmp_list(&game_stack[game_stack_size-1], &list[min]) > 0)
      memcpy(&list[min], &game_stack[game_stack_size-1], sizeof(game_state));
    pop_game();
    return;
  }
  push_game();
  memcpy(&list[list_size], &game_stack[game_stack_size-1], sizeof(game_state));
  pop_game();
  list_size++;
}

int timeout_level = 1;
volatile int timed_out;

void signal_timeout(int _)
{
  //printf("timeout!\n");
  //timed_out = 1;
  //alarm(1);
}

void do_list(int level)
{
  int _list_size;
  int i, j;
  int k;

  //if (level <= 4) printf("level %d\n", level);

  while (1) {
    memcpy(_list, list, list_size * sizeof(game_state));
    _list_size = list_size;

    if (_list_size == 0) break;
level++;
printf("level %d, to process %d\n", level, _list_size);

    list_size = 0;

    //if (level >= 3 && _list_size > 1) _list_size = 1;

    for (k = _list_size-1; k>=0; k--) {
      game_state *g;

      push_game();

      /* to set game state to the one in the list, simulate push and then pop */
      if (game_stack_size == 1024) abort();
      g = &_list[k];
      memcpy(&game_stack[game_stack_size], g, sizeof(game_state));
      game_stack_size++;

      pop_game();

      stacklen = 0;

      /* run next hand */
      inci = 0; incj = 1;
      for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) {
        starti = i; startj = j;
        /* must be a start of a word (no letter before) */
        if (is_letter(starti-inci, startj-incj)) continue;
        if (timed_out && timeout_level <= level) goto over;
        go(dict.top, i, j);
      }
      inci = 1; incj = 0;
      for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) {
        starti = i; startj = j;
        /* must be a start of a word (no letter before) */
        if (is_letter(starti-inci, startj-incj)) continue;
        if (timed_out && timeout_level <= level) goto over;
        go(dict.top, i, j);
      }

over:
      pop_game();

      if (timed_out && timeout_level <= level) break;
    }

    if (timeout_level >= level) timed_out = 0;

    sort_list();
  }
}

void contest(char *bag)
{
  char cur_hand[8];
  int i, j;

  signal(SIGALRM, signal_timeout);
  alarm(5);

  if (strlen(bag) > 199) abort();
  strcpy(letters, bag);
  translate_letters();

  for (i = 0; i < 7; i++) cur_hand[i] = eat_letter();
  cur_hand[7] = 0;
  read_hand(cur_hand);

  inci = 0; incj = 1;
  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) {
    starti = i; startj = j;
    /* must be a start of a word (no letter before) */
    if (is_letter(starti-inci, startj-incj)) continue;
    go(dict.top, i, j);
  }

  sort_list();
  do_list(0);
}

int best_score = -10000000;

int total_score(int score)
{
  extern letter english_letters[];
  int i;
  for (i = 0; i < 7; i++)
    if (hand[i] < 0x1000000) score -= english_letters[hand[i]].value;
  for (i = next_letter; letters[i]; i++)
    score -= english_letters[translated_letters[i]].value;
  return score;
}

void print_board(void)
{
  int i, j;
  for (i = 0; i < 15; i++) {
    for (j = 0; j < 15; j++) printf("%c", board[i][j]!=-10000000 ? 'A' + board[i][j]-1 : '.');
    printf("\n");
  }
}

void print_steps(void)
{
  int i, j;
  for (i = 0; i < steps_size; i++) {
    int row = G_STEP_I(steps[i]);
    int col = G_STEP_J(steps[i]);
    int len = G_STEP_LEN(steps[i]);
    int vert = G_STEP_VERTICAL(steps[i]);
    int score = G_STEP_SCORE(steps[i]);
    int incrow, inccol;
    if (!vert) {
      printf("%d%c ", row+1, col+'A');
      incrow = 0;
      inccol = 1;
    } else {
      printf("%c%d ", col+'A', row+1);
      incrow = 1;
      inccol = 0;
    }
    for (j = 0; j < len; j++, row += incrow, col += inccol) {
      int c = board[row][col];
      int joker = is_joker[row][col];
      printf("%c", c+'A'-1 + joker * ('a'-'A'));
    }
    printf(" %d", score);
    if (i == steps_size-1) printf("\n"); else printf(",");
  }
}

static int _find15(dict_letter *l, int *w)
{
  if (*w == 0) return 0;
  if (l == NULL) return 0;
next:
  if (l->l == *w || *w == -10000000) {
    if (l->w) return 1;
    if (_find15(l->child, w+1)) return 1;
    if (l->l == *w) return 0;
  }
  l = l->next;
  if (l != NULL) goto next;
  return 0;
}

int find15(int *w)
{
  extern dictionary dict15;
  return _find15(dict15.top, w);
}

void put_word(int score, int _starti, int _startj, int vert, int *letters)
{
  int i;
  int current_total_score;
  int len;
  int top=0, bottom=0, left=0, right=0;
  push_game();

  wordscore = score;
  cumul_score += score;

  /* "play" on the board */
  /* board[i][j] = letter (A=1) or -10000000 if empty */
  /* is_joker[i][j] = 1 if letter at (i,j) is a joker */
  /* hand[k] = 1 (A) .. 26 (Z), or 0 for joker or 0x1000000 if no letter
   * + 0x1000000 if used
   */
  int new_letters = 0;
  for (i = 0; i < stacklen; i++) {
    if (board[stack[i].i][stack[i].j] == -10000000) new_letters++;
    if (stack[i].i == 0) top = 1;
    if (stack[i].i == 14) bottom = 1;
    if (stack[i].j == 0) left = 1;
    if (stack[i].j == 14) right = 1;
    board[stack[i].i][stack[i].j] = stack[i].l->l;
    if (stack[i].hl == 0) is_joker[stack[i].i][stack[i].j] = 1;
  }
  /* recompute hand */
  nb_input_letters = 0;
  for (i = 0; i < 7; i++) {
    int c;
    if (hand[i] < 0x1000000) { nb_input_letters++; continue; }
    c = eat_letter();
    if (c != '.') nb_input_letters++;
    if (c == '_') c = 0;                 /* joker */
    else if (c == '.') c = 0x1000000;  /* no letter, mark it as used */
    else { if (!my_isalpha(c)) abort(); c = up(c); }
    hand[i] = c;
  }

  /* compute len of played word */
  for (len = 0; letters[len]; len++) /* do nothing */;
  add_step(_starti, _startj, vert, len, score);

  current_total_score = total_score(cumul_score);
  if (current_total_score > best_score) {
    best_score = current_total_score;
    printf("current score %d (next_letter %d score %d)\n", best_score, next_letter, cumul_score);
    print_board();
    print_steps();
  }

  /* check oxyphenbutazone compatible */
#if 0
  if (board[0][14] != -10000000 &&  board[0][14]  != 'o'-'a'+1) goto over;
  if (board[1][14] != -10000000 &&  board[1][14]  != 'x'-'a'+1) goto over;
  if (board[2][14] != -10000000 &&  board[2][14]  != 'y'-'a'+1) goto over;
  if (board[3][14] != -10000000 &&  board[3][14]  != 'p'-'a'+1) goto over;
  if (board[4][14] != -10000000 &&  board[4][14]  != 'h'-'a'+1) goto over;
  if (board[5][14] != -10000000 &&  board[5][14]  != 'e'-'a'+1) goto over;
  if (board[6][14] != -10000000 &&  board[6][14]  != 'n'-'a'+1) goto over;
  if (board[7][14] != -10000000 &&  board[7][14]  != 'b'-'a'+1) goto over;
  if (board[8][14] != -10000000 &&  board[8][14]  != 'u'-'a'+1) goto over;
  if (board[9][14] != -10000000 &&  board[9][14]  != 't'-'a'+1) goto over;
  if (board[10][14] != -10000000 && board[10][14] != 'a'-'a'+1) goto over;
  if (board[11][14] != -10000000 && board[11][14] != 'z'-'a'+1) goto over;
  if (board[12][14] != -10000000 && board[12][14] != 'o'-'a'+1) goto over;
  if (board[13][14] != -10000000 && board[13][14] != 'n'-'a'+1) goto over;
  if (board[14][14] != -10000000 && board[14][14] != 'e'-'a'+1) goto over;
#endif
  if (board[0][14] != -10000000 || board[7][14] != -10000000 ||
      board[14][14] != -10000000) {
    if (board[0][14] == -10000000 || board[7][14] == -10000000 ||
        board[14][14] == -10000000) goto over;
    print_board(); print_steps(); abort();
  }
  if (board[0][0] != -10000000 || board[7][0] != -10000000 ||
      board[14][0] != -10000000) {
    if (board[0][0] == -10000000 || board[7][0] == -10000000 ||
        board[14][0] == -10000000) goto over;
    print_board(); print_steps(); abort();
  }
  if (board[0][0] != -10000000 || board[0][7] != -10000000 ||
      board[0][14] != -10000000) {
    if (board[0][0] == -10000000 || board[0][7] == -10000000 ||
        board[0][14] == -10000000) goto over;
    print_board(); print_steps(); abort();
  }
  if (board[14][0] != -10000000 || board[14][7] != -10000000 ||
      board[14][14] != -10000000) {
    if (board[14][0] == -10000000 || board[14][7] == -10000000 ||
        board[14][14] == -10000000) goto over;
    print_board(); print_steps(); abort();
  }

  /* on the left */
  for (i = 0; i < 15; i++)
    if (board[i][1] != -10000000 && board[i][0] == -10000000) goto over;
  /* on the right */
  for (i = 0; i < 15; i++)
    if (board[i][13] != -10000000 && board[i][14] == -10000000) goto over;
  /* on the top */
  for (i = 0; i < 15; i++)
    if (board[1][i] != -10000000 && board[0][i] == -10000000) goto over;
  /* on the bottom */
  for (i = 0; i < 15; i++)
    if (board[13][i] != -10000000 && board[14][i] == -10000000) goto over;

  if (top) {
    int w[15];
    int bl = 0;
    for (i = 0; i < 15; i++) { w[i] = board[0][i]; if(w[i]!=-10000000)bl++;}
//print_board(); printf("top found? %d\n", find15(w)); getchar();
    if (bl > border_letters) border_letters = bl;
    if (!find15(w)) goto over;
if (bl >= 8) { print_board(); print_steps(); abort(); }
  }
  if (bottom) {
    int w[15];
    int bl = 0;
    for (i = 0; i < 15; i++) { w[i] = board[14][i]; if(w[i]!=-10000000)bl++;}
//print_board(); printf("bottom found? %d\n", find15(w)); getchar();
    if (bl > border_letters) border_letters = bl;
    if (!find15(w)) goto over;
if (bl >= 8) { print_board(); print_steps(); abort(); }
  }
  if (left) {
    int w[15];
    int bl = 0;
    for (i = 0; i < 15; i++) { w[i] = board[i][0]; if(w[i]!=-10000000)bl++;}
//print_board(); printf("left found? %d\n", find15(w)); getchar();
    if (bl > border_letters) border_letters = bl;
    if (!find15(w)) goto over;
if (bl >= 8) { print_board(); print_steps(); abort(); }
  }
  if (right) {
    int w[15];
    int bl = 0;
    for (i = 0; i < 15; i++) { w[i] = board[i][14]; if(w[i]!=-10000000)bl++;}
//print_board(); printf("right found? %d\n", find15(w)); getchar();
    if (bl > border_letters) border_letters = bl;
    if (!find15(w)) goto over;
if (bl >= 8) { print_board(); print_steps(); abort(); }
  }

//  if (new_letters < 4 - 2 * (top+left+right+bottom)) goto over;

#if 0
  border_letters = 0;
  for (i = 0; i < 15; i++) {
    if (board[i][0] != -10000000) border_letters++;
    if (board[i][14] != -10000000) border_letters++;
    if (board[0][i] != -10000000) border_letters++;
    if (board[14][i] != -10000000) border_letters++;
  }
#endif

//  if (score > 4)
  put_in_list();

over:
  pop_game();
}
