#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define DL 1
#define TL 2
#define DW 3
#define TW 4

int value[15][15] = {
{ TW,  0,  0, DL,  0,  0,  0, TW,  0,  0,  0, DL,  0,  0, TW  },
{  0, DW,  0,  0,  0, TL,  0,  0,  0, TL,  0,  0,  0, DW,  0  },
{  0,  0, DW,  0,  0,  0, DL,  0, DL,  0,  0,  0, DW,  0,  0  },
{ DL,  0,  0, DW,  0,  0,  0, DL,  0,  0,  0, DW,  0,  0, DL  },
{  0,  0,  0,  0, DW,  0,  0,  0,  0,  0, DW,  0,  0,  0,  0  },
{  0, TL,  0,  0,  0, TL,  0,  0,  0, TL,  0,  0,  0, TL,  0  },
{  0,  0, DL,  0,  0,  0, DL,  0, DL,  0,  0,  0, DL,  0,  0  },
{ TW,  0,  0, DL,  0,  0,  0, DW,  0,  0,  0, DL,  0,  0, TW  },
{  0,  0, DL,  0,  0,  0, DL,  0, DL,  0,  0,  0, DL,  0,  0  },
{  0, TL,  0,  0,  0, TL,  0,  0,  0, TL,  0,  0,  0, TL,  0  },
{  0,  0,  0,  0, DW,  0,  0,  0,  0,  0, DW,  0,  0,  0,  0  },
{ DL,  0,  0, DW,  0,  0,  0, DL,  0,  0,  0, DW,  0,  0, DL  },
{  0,  0, DW,  0,  0,  0, DL,  0, DL,  0,  0,  0, DW,  0,  0  },
{  0, DW,  0,  0,  0, TL,  0,  0,  0, TL,  0,  0,  0, DW,  0  },
{ TW,  0,  0, DL,  0,  0,  0, TW,  0,  0,  0, DL,  0,  0, TW  }};

typedef struct {
  char l;
  int score;
  int count;
} letter;

/* english letters */
letter letters[] = {
  {   0, 0, 2 },           /* joker */
  { 'A', 1, 9 },
  { 'B', 3, 2 },
  { 'C', 3, 2 },
  { 'D', 2, 4 },
  { 'E', 1, 12 },
  { 'F', 4, 2 },
  { 'G', 2, 3 },
  { 'H', 4, 2 },
  { 'I', 1, 9 },
  { 'J', 8, 1 },
  { 'K', 5, 1 },
  { 'L', 1, 4 },
  { 'M', 3, 2 },
  { 'N', 1, 6 },
  { 'O', 1, 8 },
  { 'P', 3, 2 },
  { 'Q', 10, 1 },
  { 'R', 1, 6 },
  { 'S', 1, 4 },
  { 'T', 1, 6 },
  { 'U', 1, 4 },
  { 'V', 4, 2 },
  { 'W', 4, 2 },
  { 'X', 8, 1 },
  { 'Y', 4, 2 },
  { 'Z', 10, 1 },
};

char board[15][15];
char bag[100+1];
int handpos[7];
int next_letter;

int is_letter(int row, int col)
{
  return row >= 0 && row <= 14 && col >= 0 && col <= 14 && board[row][col];
}

int other_word_score(int row, int col, int vert, int letter)
{
  int score = 0;
  int letter_mul = 1;
  int incrow, inccol;
  int r = row, c = col;
  int word_mul = 1;

  /* other direction */
  if (1-vert) { incrow = 1; inccol = 0; } else { incrow = 0; inccol = 1; }

  if (!is_letter(row-incrow, col-inccol) && !is_letter(row+incrow, col+inccol))
    return 0;

  switch(value[row][col]) {
    case DL: letter_mul = 2; break;
    case TL: letter_mul = 3; break;
    case DW: word_mul = 2; break;
    case TW: word_mul = 3; break;
  }

  board[row][col] = letter;

  do { row -= incrow; col -= inccol; } while (is_letter(row, col));
  row += incrow; col += inccol;

  do {
    if (board[row][col] >= 'A' && board[row][col] <= 'Z')
      score += letters[board[row][col]-'A'+1].score;
    row += incrow; col += inccol;
  } while (is_letter(row, col));

  board[r][c] = 0;

  if (letter >= 'A' && letter <= 'Z') letter -= 'A'-1; else letter = 0;
  return word_mul * (score + (letter_mul-1) * letters[letter].score);
}

int compute_score(int row, int col, int vert, char *l)
{
  int other_score = 0;
  int score = 0;
  int new_letters = 0;
  int letter_mul, word_mul;
  int incrow, inccol;

  if (vert) { incrow = 1; inccol = 0; } else { incrow = 0; inccol = 1; }

  word_mul = 1;
  for (; *l; row+=incrow, col+=inccol, l++) {
    letter_mul = 1;
    if (board[row][col] == 0) {
      other_score += other_word_score(row, col, vert, *l);
      new_letters++;
      switch (value[row][col]) {
      case DL: letter_mul = 2; break;
      case TL: letter_mul = 3; break;
      case DW: word_mul *= 2; break;
      case TW: word_mul *= 3; break;
      }
    }
    if (*l >= 'A' && *l <= 'Z') score += letters[*l-'A'+1].score * letter_mul;
  }

  score *= word_mul;
  if (new_letters == 7) score += 50;

  return score + other_score;
}

/* return value in [0..max] inclusive */
int get_random(int max)
{
  return rand() / ((double)RAND_MAX+1) * (max+1);
}

/* how we do?
 * we generate a permutation over the hand
 * when we hae to play a word, we pickup the position of each new
 * letter into the bag via the permutation of the hand
 * say letter 2, we take hand position permute[2] which will
 * give us bag position XX
 * when done with a letter, we 'pickup' a new letter from the
 * bag and put it in the hand at the new empty position
 * (we actually just set hand[permute[2]] to next_letter and
 * do next_letter++)
 * next_letter *will* get bigger than bagsize by at most 7 (the
 * size of full hand). This is useful to know the size of the hand
 * in the end, when it's lower than 7 (no more letters in bag)
 * when we are there, each time we pickup a letter from the end
 * it may create a hole, so we must pack the hand back to the left
 */
void update_bag(int row, int col, int vert, char *l)
{
  int incrow, inccol;
  int r, c, k, pr, pw;
  char *p;
  int permute[7];
  int new_letters = 0;
  int handsize;

  handsize = 107 - next_letter;
  if (handsize > 7) handsize = 7;

  if (vert) { incrow = 1; inccol = 0; } else { incrow = 0; inccol = 1; }

  /* get number of new letters */
  for (k=0, p=l, r=row, c=col; *p; p++, r+=incrow, c+=inccol)
    if (!board[r][c]) new_letters++;

  /* generate a permutation */
  for (k = 0; k < handsize; k++) {
    int k2 = get_random(k);
    permute[k] = permute[k2];
    permute[k2] = k;
  }

  /* update bag and hand */
  for (k=0, p=l, r=row, c=col; *p; p++, r+=incrow, c+=inccol) {
    if (board[r][c]) continue;
    if (k == handsize) abort();
    if (handpos[permute[k]] >= 100) abort();
    bag[handpos[permute[k]]] = *p >= 'A' && *p <= 'Z' ? *p : '?';
    handpos[permute[k]] = next_letter++;
    k++;
  }

  /* no compacting if handsize is 7 */
  /* I think this test could use 101, but I will keep it as is, it doesn't
   * hurt, the following routine might even work for whatever next_letter
   */
  if (next_letter < 100) return;

  /* compact hand to the beginning of it */
  for (pr=0, pw=0; pr < handsize; pr++)
    if (handpos[pr] < 100) { handpos[pw] = handpos[pr]; pw++; }
}

int main(void)
{
  int c;
  char l[16];
  int len;
  int row, col, vert;
  int incrow, inccol, k;
  char *p;
  int score = 0, wordscore;
  int first_word = 1;

  for (next_letter = 0; next_letter < 7; next_letter++)
    handpos[next_letter] = next_letter;

  while (1) {
    /* read move */
    do c = getchar(); while (isspace(c));
    if (c == EOF) break;
    if (c >= 'A' && c <= 'Z') {
      vert = 1;
      col = c - 'A';
      if (scanf("%d", &row) != 1 || row < 1 || row > 15) abort();
      row--;
    } else if (c >= '0' && c <= '9') {
      ungetc(c, stdin);
      if (scanf("%d", &row) != 1 || row < 1 || row > 15) abort();
      row--;
      c = getchar(); if (c == EOF) abort();
      if (!(c >= 'A' && c <= 'Z')) abort();
      col = c - 'A';
      vert = 0;
    } else abort();
    do c = getchar(); while (isspace(c)); if (c == EOF) abort();
    ungetc(c, stdin);
    for (len = 0; len < 15; len++) {
      c = getchar();
      if ((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z')) l[len] = c;
      else break;
    }
    l[len] = 0;
    if ((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z')) c = getchar();
while (c!=',' && c != EOF) c=getchar();
    if (!(isspace(c) || c == ',' || c == EOF)) abort();

    if (vert) { incrow = 1; inccol = 0; } else { incrow = 0; inccol = 1; }

    /* get score */
    score += wordscore = compute_score(row, col, vert, l);

    /* deduce bag stuff */
    update_bag(row, col, vert, l);

    /* put on board */
    if (vert)
      printf("%s%c%d ", first_word?"":",", col+'A', row+1);
    else
      printf("%s%d%c ", first_word?"":",", row+1, col+'A');
    first_word = 0;
    for (k = 0, p = l; *p; k++, p++, row+=incrow, col+=inccol) {
      if (row >= 15 || col >= 15) abort();
      if (board[row][col] && board[row][col] != *p) abort();
      if (board[row][col] == 0) {
        printf("\x1b[32m%c\x1b[0m", *p);
        if (*p >= 'A' && *p <= 'Z') {
          if (letters[*p-'A'+1].count == 0) abort();
          letters[*p-'A'+1].count--;
        } else {
          if (letters[0].count == 0) abort();
          letters[0].count--;
        }
      } else
        printf("%c", *p);
      board[row][col] = *p;
    }
    printf(" %d", wordscore);
  }

  printf("\n\n");
  for (row = 0; row < 15; row++) {
    for (col = 0; col < 15; col++) {
      printf("%c", board[row][col] ? board[row][col] : '.');
    }
    printf("\n");
  }

  printf("score without caring of remaining bag: %d\n", score);

  /* count remaining letters */
  next_letter = 0;

  for (k = 0; k < 27; k++) {
    if (!letters[k].count) continue;
    printf("%c %d\n", k+'A'-1, letters[k].count);
    score -= letters[k].count * letters[k].score;
    while (letters[k].count) {
      if (next_letter >= 100) abort();
      while (bag[next_letter]) next_letter++;
      if (next_letter >= 100) abort();
      bag[next_letter] = letters[k].l==0?'?':letters[k].l;
      next_letter++;
      letters[k].count--;
    }
  }

  if (next_letter != 0 && next_letter != 100) abort();

  printf("final score: %d\n", score);

  bag[100] = 0;
  printf("bag: %s\n", bag);
  return 0;
}
