#include "defs.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdint.h>
#include <X11/Xlib.h>

static int is_positive_letter(int board[15][15], int row, int col)
{
  return row >= 0 && row < 15 && col >= 0 && col < 15 && board[row][col] > 0;
}

static int other_word_ok(game *g, int l, int i, int j,
                         int inci, int incj)
{
  extern dictionary dict;
  int si = i, sj = j;
  int word[16];
  int w;
  int ii = 1 - inci;
  int jj = 1 - incj;
  int starti = i, startj = j;
  int endi, endj;
  g->board[si][sj] = l;
  /* find start/end pos of word */
  while (starti >= 0 && startj >= 0 && g->board[starti][startj] != 0)
    starti -= ii, startj -= jj;
  starti += ii; startj += jj;
  endi = starti; endj = startj;
  while (endi < 15 && endj < 15 && g->board[endi][endj] != 0)
    endi += ii, endj += jj;
  endi -= ii; endj -= jj;
  /* get word */
  for (w=0, i=starti, j=startj; i!=endi+ii || j!=endj+jj; w++, i+=ii, j+=jj) {
    int letter = abs(g->board[i][j]); if (letter > 26) letter -= 26;
    word[w] = letter;
  }
  word[w] = 0;
  g->board[si][sj] = 0;
  return w == 1 || find_word(&dict, word);
}

static int other_positive_word_ok(game *g, int l, int i, int j,
                                  int inci, int incj)
{
  extern dictionary dict;
  int si = i, sj = j;
  int word[16];
  int w;
  int ii = 1 - inci;
  int jj = 1 - incj;
  int starti = i, startj = j;
  int endi, endj;
  g->board[si][sj] = l;
  /* find start/end pos of word */
  while (starti >= 0 && startj >= 0 && g->board[starti][startj] > 0)
    starti -= ii, startj -= jj;
  starti += ii; startj += jj;
  endi = starti; endj = startj;
  while (endi < 15 && endj < 15 && g->board[endi][endj] > 0)
    endi += ii, endj += jj;
  endi -= ii; endj -= jj;
  /* get word */
  for (w=0, i=starti, j=startj; i!=endi+ii || j!=endj+jj; w++, i+=ii, j+=jj) {
    int letter = abs(g->board[i][j]); if (letter > 26) letter -= 26;
    word[w] = letter;
  }
  word[w] = 0;
  g->board[si][sj] = 0;
  return w == 1 || find_word(&dict, word);
}

typedef struct {
  int row, col, vert;
  int len;
  int cur[15];
  int preboard[15];
} wordat;

typedef struct {
  wordat *w;
  int wlen;
  game *g;
  letterstack *s;
  movelist *ml;
} world;

int valid_board(world *w)
{
  /* for each wordat:
   *   no letter before / after
   *   if letter X not on board, no near letter allowed in the other direction
   */
  int row, col, incrow, inccol, l;
  int i;
  wordat *at;
  for (i = 0; i < w->wlen; i++) {
    at = &w->w[i];
    row = at->row;
    col = at->col;
    if (at->vert) { incrow = 1; inccol = 0; } else { incrow = 0; inccol = 1; }
    if (is_positive_letter(w->g->board, row-incrow, col-inccol)) return 0;
    if (is_positive_letter(w->g->board,
                           row+incrow*at->len, col+inccol*at->len)) return 0;
    for (l = 0; l < at->len; l++, row+=incrow, col+=inccol) {
      if (w->g->board[row][col] >= 0) continue;
      /* letter in *other* direction must be empty (or smthing < 0) */
      if (is_positive_letter(w->g->board, row-inccol, col-incrow)) return 0;
      if (is_positive_letter(w->g->board, row+inccol, col+incrow)) return 0;
    }
  }
  return 1;
}

int putit(world *w, wordat *at)
{
  int count[27];
  int row, col, incrow, inccol, l;
  if (at->vert) { incrow = 1; inccol = 0; } else { incrow = 0; inccol = 1; }
  /* check we can put it */
  for (l = 0; l < 27; l++) count[l] = w->g->lcount[l];
  for (row=at->row,col=at->col,l=0;l<at->len;l++,row+=incrow,col+=inccol) {
    if (w->g->board[row][col] && w->g->board[row][col]!=-at->cur[l]) return 0;
    if (w->g->board[row][col]) continue;
    if (count[at->cur[l]] == 0) return 0;
    count[at->cur[l]]--;
  }
  /* put it */
  for (row=at->row,col=at->col,l=0;l<at->len;l++,row+=incrow,col+=inccol) {
    at->preboard[l] = w->g->board[row][col];
    if (!at->preboard[l]) w->g->lcount[at->cur[l]]--;
    w->g->board[row][col] = -at->cur[l];
  }
  return 1;
}

void unputit(world *w, wordat *at)
{
  int row, col, incrow, inccol, l;
  if (at->vert) { incrow = 1; inccol = 0; } else { incrow = 0; inccol = 1; }
  for (row=at->row,col=at->col,l=0;l<at->len;l++,row+=incrow,col+=inccol) {
    w->g->board[row][col] = at->preboard[l];
    if (!at->preboard[l]) w->g->lcount[at->cur[l]]++;
  }
}

static int word_ok(game *g, int i, int j, int inci, int incj)
{
  extern dictionary dict;
  int word[16];
  int w;
  int ii = inci;
  int jj = incj;
  int starti = i, startj = j;
  int endi, endj;
  /* find start/end pos of word */
  while (starti >= 0 && startj >= 0 && g->board[starti][startj] != 0)
    starti -= ii, startj -= jj;
  starti += ii; startj += jj;
  endi = starti; endj = startj;
  while (endi < 15 && endj < 15 && g->board[endi][endj] != 0)
    endi += ii, endj += jj;
  endi -= ii; endj -= jj;
  /* get word */
  for (w=0, i=starti, j=startj; i!=endi+ii || j!=endj+jj; w++, i+=ii, j+=jj) {
    int letter = abs(g->board[i][j]); if (letter > 26) letter -= 26;
    word[w] = letter;
  }
  word[w] = 0;
  return w == 1 || find_word(&dict, word);
}

static int positive_word_ok(game *g, int i, int j, int inci, int incj)
{
  extern dictionary dict;
  int word[16];
  int w;
  int ii = inci;
  int jj = incj;
  int starti = i, startj = j;
  int endi, endj;
  /* find start/end pos of word */
  while (starti >= 0 && startj >= 0 && g->board[starti][startj] > 0)
    starti -= ii, startj -= jj;
  starti += ii; startj += jj;
  endi = starti; endj = startj;
  while (endi < 15 && endj < 15 && g->board[endi][endj] > 0)
    endi += ii, endj += jj;
  endi -= ii; endj -= jj;
  /* get word */
  for (w=0, i=starti, j=startj; i!=endi+ii || j!=endj+jj; w++, i+=ii, j+=jj) {
    int letter = abs(g->board[i][j]); if (letter > 26) letter -= 26;
    word[w] = letter;
  }
  word[w] = 0;
  return w == 1 || find_word(&dict, word);
}

#if 0
int word_possible(world *w, wordat *at)
{
  int row, col, incrow, inccol, l;
  row = at->row;
  col = at->col;
  if (at->vert) { incrow = 1; inccol = 0; } else { incrow = 0; inccol = 1; }
#if 0
  /* check we don't have something before/after */
  if (letterat(w->g->board, row-incrow, col-inccol)) return 0;
  if (letterat(w->g->board, row+incrow*at->len, col+inccol*at->len)) return 0;
#endif
  for (row=at->row,col=at->col,l=0;l<at->len;l++,row+=incrow,col+=inccol) {
    if (w->g->board[row][col] <= 0) {
      w->g->board[row][col] = at->cur[l];
      if (!word_ok(w->g, row,col,inccol,incrow)) return 0;
    } else {
      if (abs(w->g->board[row][col]) != at->cur[l]) return 0;
    }
  }
  if (!word_ok(w->g, at->row,at->col,incrow,inccol)) return 0;
  return 1;
}

int future_possible(world *w)
{
  int ret = 1;
  int b[15][15];
  int i;
  wordat a;

  memcpy(b, w->g->board, 15*15*sizeof(int));

  for (i = 0; i < w->wlen; i++) {
    wordat *at = w->w + i;
    if (!word_possible(w, at)) { ret = 0; goto end; }
  }

  a = (wordat){ 14, 0, 0, 15, {O,X,Y,P,H,E,N,B,U,T,A,Z,O,N,E} };
  if (!word_possible(w, &a)) { ret = 0; goto end; }

end:
  memcpy(w->g->board, b, 15*15*sizeof(int));
  return ret;
}
#endif

static void dpush(letterstack *s, game *g, dict_letter *n, int hl, int new)
{
  s->stack[s->stacklen].l = n;
  s->stack[s->stacklen].hl = hl;
  s->stack[s->stacklen].new = new;
  s->stacklen++;
}

static void dpop(letterstack *s)
{
  s->stacklen--;
}

static void check_word(game *g, letterstack *s, movelist *ml,
                int starti, int startj, int inci, int incj)
{
  move m;
  int i, j, l, ok = 0;
  /* word must contain at least 1 new letter */
  for (l=0, i=starti, j=startj; l<s->stacklen; l++, i+=inci, j+=incj)
    if (!is_positive_letter(g->board, i, j)) { ok = 1; break; }
  if (!ok) return;
  ok = 0;
  for (l=0, i=starti, j=startj; l<s->stacklen; l++, i+=inci, j+=incj)
    if (is_positive_letter(g->board, i, j) ||             /* letter on board */
        is_positive_letter(g->board, i-incj, j-inci) ||   /* touch on board */
        is_positive_letter(g->board, i+incj, j+inci) ||   /* touch on board */
        (i == 7 && j == 7))            /* word is at center pos (first word) */
    { ok = 1; break; }
  if (!ok) return;
  for (l = 0; l < 7; l++) m.next_hand[l] = g->hand[l];
  for (l = 0; l < s->stacklen; l++) {
    m.word[l] = s->stack[l].l->l;
    if (s->stack[l].hl == 0) m.word[l] += 26;
    if (s->stack[l].new) m.word[l] = -m.word[l];
  }
  m.wordlen = s->stacklen;
  m.row = starti;
  m.col = startj;
  m.vert = inci ? 1 : 0;
  m.score = score(g, m.word, m.wordlen, starti, startj, inci, incj);
  add_movelist(ml, &m);
}

void gogo(world *w, dict_letter *l,
          int starti, int startj, int i, int j, int inci, int incj)
{
  int h;
  int hletter;
  uint32_t letter_done = 0, joker_done = 0;

  if (l==NULL) return;
  if (i==15||j==15) return;
again:
  if (w->g->board[i][j] > 0) {
    int letter = w->g->board[i][j]; if (letter > 26) letter -= 26;
    if (l->l == letter) {
      dpush(w->s, w->g, l, w->g->board[i][j] > 26 ? 0 : w->g->board[i][j], 0);
      if (l->w && (i+inci==15||j+incj==15||w->g->board[i+inci][j+incj]==0))
        check_word(w->g, w->s, w->ml, starti, startj, inci, incj);
      gogo(w, l->child, starti, startj, i + inci, j + incj, inci, incj);
      dpop(w->s);
      goto next;
    }
    goto next;
  }
  if (w->g->board[i][j] < 0) {
    int letter = abs(w->g->board[i][j]); if (letter > 26) {letter-=26;abort();}
    if (l->l != letter) goto next;
    /* reserved letter, must be in hand */
    for (h = 0; h < 7; h++) if (w->g->hand[h] == -w->g->board[i][j]) break;
    if (h == 7) goto next;
    hletter = w->g->hand[h];
    if (letter_done & (1 << hletter)) goto next;
    /* other (already there) word must be ok */
    if (!word_ok(w->g,i, j, incj, inci)) goto next;
    {
      /* to check with positive letter, we must make it >0 */
      int positive;
      w->g->board[i][j] = abs(w->g->board[i][j]);
      positive = positive_word_ok(w->g,i, j, incj, inci);
      w->g->board[i][j] = -w->g->board[i][j];
      if (!positive) goto next;
    }
    w->g->hand[h] = -1;
    letter_done |= 1 << hletter;
    /* no lcount decrement, letter reservation has already done that */
    dpush(w->s, w->g, l, hletter, 1);
    if (l->w && (i+inci==15||j+incj==15||w->g->board[i+inci][j+incj]==0))
      check_word(w->g, w->s, w->ml, starti, startj, inci, incj);
    gogo(w, l->child, starti, startj, i+inci, j+incj, inci, incj);
    dpop(w->s);
    w->g->hand[h] = hletter;
    goto next;
  }

  /* no letter on board, no reserved letter */
  for (h = 0; h < 7; h++) if (w->g->hand[h]==l->l || w->g->hand[h]==0) break;
  if (h == 7) goto next;
  hletter = w->g->hand[h];
  /* must not have been tried before */
  if (hletter && (letter_done & (1 << hletter))) goto next;
  if (!hletter && (joker_done & (1 << l->l))) goto next;
  /* if not a joker, letter must be in lcount */
  if (/*hletter && */!w->g->lcount[hletter]) goto next;
  if (!other_word_ok(w->g, l->l, i, j, inci, incj)) goto next;
  if (!other_positive_word_ok(w->g, l->l, i, j, inci, incj)) goto next;
  w->g->hand[h] = -1;
  if (hletter) letter_done |= 1 << hletter;
  if (!hletter) joker_done |= 1 << l->l;
  w->g->lcount[hletter]--;
  dpush(w->s, w->g, l, hletter, 1);
  if (l->w && (i+inci==15||j+incj==15||w->g->board[i+inci][j+incj]==0))
    check_word(w->g, w->s, w->ml, starti, startj, inci, incj);
  gogo(w, l->child, starti, startj, i+inci, j+incj, inci, incj);
  dpop(w->s);
  w->g->lcount[hletter]++;
  w->g->hand[h] = hletter;

next:
  l = l->next;
  if (l != NULL) goto again;
}

void get_dmoves(world *w)
{
  extern dictionary dict;
  int row, col;
  for (row = 0; row < 15; row++) for (col = 0; col < 15; col++) {
    if (!is_positive_letter(w->g->board, row-1, col))
      gogo(w, dict.top, row, col, row, col, 1, 0);
    if (!is_positive_letter(w->g->board, row, col-1))
      gogo(w, dict.top, row, col, row, col, 0, 1);
  }
}

int move_distance(const void *_a, const void *_b)
{
  const move *a = _a;
  const move *b = _b;
  if (a->distance < b->distance) return -1;
  if (a->distance > b->distance) return 1;
  return 0;
}

int move_length(const void *_a, const void *_b)
{
  const move *a = _a;
  const move *b = _b;
  if (a->score < b->score) return -1;
  if (a->score > b->score) return 1;
  if (a->wordlen < b->wordlen) return -1;
  if (a->wordlen > b->wordlen) return 1;
  return 0;
}

void compute_distance(world *w, move *m)
{
  int i, j;
  int distance = 0;
  play_move_noneg(w->g, m);
  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++)
    if (w->g->board[i][j] < 0) distance--;
  m->distance = distance;
  unplay_move(w->g, m);
}

int distance(int *w, int wlen, int *_hand, int *_bag, int bagpos)
{
  int ret = 0;
  int i, j;
  int hand[7];
  int bag[100];
  memcpy(hand, _hand, 7 * sizeof(int));
  memcpy(bag, _bag, 100 * sizeof(int));
  for (i = 0; i < wlen; i++) {
    if (w[i] > 0) continue;
    /* is letter in hand */
    for (j = 0; j < 7; j++) if (hand[j] == -w[i]) { hand[j] = -1; break; }
    if (j != 7) continue;
    /* where is it in bag */
    for (j = bagpos; j < 100; j++) if (bag[j] == -w[i]) {
      if (j-bagpos+1 > ret) ret = j-bagpos+1;
      bag[j] = -1;
      break;
    }
    if (j == 100) abort();
  }
  return ret;
}

int order_a_score(const void *_a, const void *_b)
{
  const move *a = _a;
  const move *b = _b;
  if (a->a < b->a) return -1;
  if (a->a > b->a) return 1;
  if (a->score < b->score) return -1;
  if (a->score > b->score) return 1;
  return 0;
}

int order_inva_score(const void *_a, const void *_b)
{
  const move *a = _a;
  const move *b = _b;
  /* big a come first in list (smaller interest) */
  if (a->a > b->a) return -1;
  if (a->a < b->a) return 1;
  if (a->score < b->score) return -1;
  if (a->score > b->score) return 1;
  return 0;
}

int order_score(const void *_a, const void *_b)
{
  const move *a = _a;
  const move *b = _b;
  if (a->score < b->score) return -1;
  if (a->score > b->score) return 1;
  return 0;
}

int check_vertical_bottom(world *w, int col, int letter)
{
  extern wordlist wlist[16];
  int l, i, j;
  for (l = 3; l < 8; l++) {
    if (is_letter(w->g, 15-1-l, col)) continue;
    for (i = 0; i < wlist[l].size; i++) {
      int *word = wlist[l].w[i];
      if (word[l-1] != letter) continue;
      for (j = 0; j < l; j++) {
        int cur = w->g->board[15-l+j][col]; if (cur > 26) cur -= 26;
        if (cur && cur != word[j]) break;
        if (!cur)
          if (!other_word_ok(w->g, word[j], 15-l+j, col, 1, 0)) break;
      }
      if (j == l) return 1; /* one word found */
    }
  }
  return 0;
}

int check_vertical_top(world *w, int col, int letter)
{
  extern wordlist wlist[16];
  int l, i, j;
  for (l = 3; l < 8; l++) {
    if (is_letter(w->g, l, col)) continue;
    for (i = 0; i < wlist[l].size; i++) {
      int *word = wlist[l].w[i];
      if (word[0] != letter) continue;
      for (j = 0; j < l; j++) {
        int cur = w->g->board[j][col]; if (cur > 26) cur -= 26;
        if (cur && cur != word[j]) break;
        if (!cur)
          if (!other_word_ok(w->g, word[j], j, col, 1, 0)) break;
      }
      if (j == l) return 1; /* one word found */
    }
  }
  return 0;
}

int bottom_reachable(world *w)
{
  int o, i, subword_ok, start;

  /* try all 'cuts' */
  for (o = 0; o < w->g->bottom_cut_size; o++) {
    /* if a letter is set, it must belong to the 'cut' */
    for (i = 0; i < 15; i++)
      if (w->g->board[14][i] && w->g->board[14][i] != w->g->bottom_cut[o][i])
        break;
    if (i != 15) continue;
    /* we found a compatible cut, check that each 'subword' is accessible
     * by a vertical word of length at least 3
     */
    start = 0;
    while (start < 15) {
      /* go to start of subword */
      while (start < 15 && w->g->bottom_cut[o][start] == 0) start++;
      if (start==15) break;
      /* try all letters of subword */
      subword_ok = 0;
      while (start < 15 && w->g->bottom_cut[o][start] != 0) {
        /* if letter of subword is there then it's 'reached',
         * otherwise check vertical word */
        if (w->g->board[14][start] ||
            check_vertical_bottom(w, start, w->g->bottom_cut[o][start]))
          { start++; subword_ok = 1; break; }
        start++;
      }
      if (subword_ok == 0) break;
    }
    /* we reached the end? ok */
    if (start == 15) return 1;
  }

  return 0;
}

int top_reachable(world *w)
{
  int o, i, subword_ok, start;

  /* try all 'cuts' */
  for (o = 0; o < w->g->top_cut_size; o++) {
    /* if a letter is set, it must belong to the 'cut' */
    for (i = 0; i < 15; i++)
      if (w->g->board[0][i] && w->g->board[0][i] != w->g->top_cut[o][i])
        break;
    if (i != 15) continue;
    /* we found a compatible cut, check that each 'subword' is accessible
     * by a vertical word of length at least 3
     */
    start = 0;
    while (start < 15) {
      /* go to start of subword */
      while (start < 15 && w->g->top_cut[o][start] == 0) start++;
      if (start==15) break;
      /* try all letters of subword */
      subword_ok = 0;
      while (start < 15 && w->g->top_cut[o][start] != 0) {
        /* if letter of subword is there then it's 'reached',
         * otherwise check vertical word */
        if (w->g->board[0][start] ||
            check_vertical_top(w, start, w->g->top_cut[o][start]))
          { start++; subword_ok = 1; break; }
        start++;
      }
      if (subword_ok == 0) break;
    }
    /* we reached the end? ok */
    if (start == 15) return 1;
  }

  return 0;
}

int new_decision(world *w, decision_stack *dstack, move *m, reason r,
                 movelist *ml, int i)
{
  int l;
  int lcount[27] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 };
  int triple;
  int top_count, bottom_count;

  play_move(w->g, m);

  /* check that move is bottom-compatible with bag */
  if (options.ndb == 1) {
    /* triple played at same time only */
    triple = (w->g->board[14][0]!=0) +
             (w->g->board[14][7]!=0) +
             (w->g->board[14][14]!=0);
    if (!(triple == 0 || triple == 3)) goto err;
  } else {
    /* triple and double played at same time only */
    triple = (w->g->board[14][0]!=0) +
             (w->g->board[14][3]!=0) +
             (w->g->board[14][7]!=0) +
             (w->g->board[14][11]!=0) +
             (w->g->board[14][14]!=0);
    if (!(triple == 0 || triple == 5)) goto err;
  }
  /* bottom puttable? */
  bottom_count = 0;
  for (l = 0; l < 15; l++) {
    /* count letters */
    if (w->g->board[14][l] <= 0) lcount[w->g->bottom[l]]++;
    /* count # letters already put */
    if (w->g->board[14][l] > 0) bottom_count++;
    /* already there letter must be in bottom word */
    if (w->g->board[14][l] > 0 &&
        w->g->board[14][l] != w->g->bottom[l]) goto err;
  }
  for (l = 0; l < 27; l++) if (lcount[l] > w->g->lcount[l]) goto err;

  /* bottom reachable? */
  if (bottom_count != 15 && !bottom_reachable(w)) goto err;

  /* put bottom and check other word */
  for (l = 0; l < 15; l++) {
    int old = w->g->board[14][l];
    int ok;
    w->g->board[14][l] = w->g->bottom[l];
    ok = other_word_ok(w->g, w->g->bottom[l], 14, l, 0, 1);
    w->g->board[14][l] = old;
    if (!ok) goto err;
  }

  /* check that move is top-compatible with bag */
  if (options.ndt == 1) {
    /* triple played at same time only */
    triple = (w->g->board[0][0]!=0) +
             (w->g->board[0][7]!=0) +
             (w->g->board[0][14]!=0);
    if (!(triple == 0 || triple == 3)) goto err;
  } else {
    /* triple and double played at same time only */
    triple = (w->g->board[0][0]!=0) +
             (w->g->board[0][3]!=0) +
             (w->g->board[0][7]!=0) +
             (w->g->board[0][11]!=0) +
             (w->g->board[0][14]!=0);
    if (!(triple == 0 || triple == 5)) goto err;
  }
  /* top puttable? */
  top_count = 0;
  for (l = 0; l < 15; l++) {
    /* count letters */
    if (w->g->board[0][l] <= 0) lcount[w->g->top[l]]++;
    /* count # letters already put */
    if (w->g->board[0][l] > 0) top_count++;
    /* already there letter must be in top word */
    if (w->g->board[0][l] > 0 &&
        w->g->board[0][l] != w->g->top[l]) goto err;
  }
  for (l = 0; l < 27; l++) if (lcount[l] > w->g->lcount[l]) goto err;

  /* top reachable? */
  if (top_count != 15 && !top_reachable(w)) goto err;

  /* put top and check other word */
  for (l = 0; l < 15; l++) {
    int old = w->g->board[0][l];
    int ok;
    w->g->board[0][l] = w->g->top[l];
    ok = other_word_ok(w->g, w->g->top[l], 0, l, 0, 1);
    w->g->board[0][l] = old;
    if (!ok) goto err;
  }

  /* add new decision */
  if (dstack->size == dstack->maxsize) {
    dstack->maxsize += 128;
    dstack->e = realloc(dstack->e, dstack->maxsize * sizeof(decision_elem));
    if (dstack->e == NULL) abort();
  }
  dstack->e[dstack->size].m = *m;
  dstack->e[dstack->size].r = r;
  dstack->e[dstack->size].ml = *ml;
  dstack->e[dstack->size].mlpos = i;
  dstack->size++;

  static int best_score = 0;
  if (1 || w->g->score > best_score) {
    best_score = w->g->score;
    for (i = 0; i < dstack->size; i++)
      w->g->moves[i] = dstack->e[i].m;
    w->g->movesize = dstack->size;
    print_moves(w->g);
    printf("hand:"); print_hand(w->g->hand);
    printf("bag:"); print_bag(w->g->bag, w->g->bagpos);
    print_board(w->g->board);
    printf("score %d\n", best_score);
  }
  /* reset move list */
  *w->ml = (movelist){ NULL, 0, 0 };
  return 1;

err:
  unplay_move(w->g, m);
  return 0;
}

int compute_distance_both(int board[15][15])
{
  int i, j, col, dist, row;
  int mindist_top[15] = { 99999,99999,99999,99999,99999,99999,99999,99999,99999,99999,99999,99999,99999,99999,99999 };
  int mindist_bottom[15] = { 99999,99999,99999,99999,99999,99999,99999,99999,99999,99999,99999,99999,99999,99999,99999 };
  for (i=0;i<15;i++) for (j=0;j<15;j++) {
    if (board[i][j] <= 0) continue;
    row = 0;
    for (col = 0; col < 15; col++) {
      dist = abs(i-row) + abs(j-col);
      if (j-col) dist += 1;
      if (dist < mindist_top[col]) mindist_top[col] = dist;
    }
    row = 14;
    for (col = 0; col < 15; col++) {
      dist = abs(i-row) + abs(j-col);
      if (j-col) dist += 1;
      if (dist < mindist_bottom[col]) mindist_bottom[col] = dist;
    }
  }
  dist = 0;
  for (i = 0; i < 15; i++) dist += mindist_top[i];
  for (i = 0; i < 15; i++) dist += mindist_bottom[i];
  for (i = 0; i < 15; i++) {
//    if (board[0][i] <= 0) dist += 100;
//    if (board[14][i] <= 0) dist += 100;
  }

  return dist;
}

Display *d;
Window w;
Pixmap wm;
GC gc;
GC red, green, black, blue, yellow;
int startline;
int curline = -1;
int hselect = -1;
int bselect = -1;
int redhand[7] = { -1, -1, -1, -1, -1, -1, -1 };
/* 0: normal, 1: yellow, 2: blue */
int bagcol[100];

void show_gui(world *wo)
{
  char s[256];
  char hand[7];
  int *h = wo->g->hand;
  int i, j;

  if (curline != -1 && startline + curline >= 0 &&
      startline + curline < wo->ml->size)
    play_move(wo->g, &wo->ml->v[startline+curline]);

  for (i = 0; i < 7; i++) {
    int v = h[i]; if (redhand[i] != -1) v = redhand[i];
    hand[i] = v == -1 ? '.' : v ? v + 'A' - 1 : '?';
  }

  XFillRectangle(d, wm, black, 0, 0, 900, 300);

  for (i = 0; i < 7; i++) {
    GC g;
    if (hselect == i) {
      if (redhand[i] != -1) g = black; else g = green;
    } else {
      if (redhand[i] != -1) g = red;
      else {
        g = gc;
        if (wo->g->baghand[i] != -1 && bagcol[wo->g->baghand[i]] == 1)
          g = yellow;
        if (wo->g->baghand[i] != -1 && bagcol[wo->g->baghand[i]] == 2)
          g = blue;
      }
    }
    XDrawString(d, wm, g, 10+i*8, 20, hand+i, 1);
  }

  if (wo->g->bagpos == 100) XDrawString(d, wm, gc, 10, 35, "(empty)", 7);
  else {
    for (i = wo->g->bagpos; i < 100; i++) {
      GC g = gc;
      int l = wo->g->bag[i];
      char ll = l == -1 ? '.' : l ? l + 'A' - 1 : '?';
      if (bagcol[i] == 1) g = yellow;
      if (bagcol[i] == 2) g = blue;
      XDrawString(d, wm, g, 10+(i-wo->g->bagpos)*8, 35, &ll, 1);
    }
  }

  for (i = 0; i < 100; i++) {
    GC g = gc;
    int l = wo->g->bag[i];
    char ll = l == -1 ? '.' : l ? l + 'A' - 1 : '?';
    if (bagcol[i] == 1) g = yellow;
    if (bagcol[i] == 2) g = blue;
    if (bselect == i) g = black;
    XDrawString(d, wm, g, 10+i*8, 250+14, &ll, 1);
  }

  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) {
    int l = wo->g->board[i][j];
    char ll;
    ll = l ? abs(l) > 26 ?
                abs(l) - 26 + 'a' - 1 :
                abs(l) + 'A' - 1 : '.';
    XDrawString(d, wm, gc, 10+j*14, 50+i*14, &ll, 1);
  }

  if (wo->ml->size == 0)
    XDrawString(d, wm, gc, 10+17*14, 50, "(empty)", 7);
  for (i = startline, j = 0; i < wo->ml->size && j < 15; i++, j++) {
    char word;
    int k;
    move *m = &wo->ml->v[i];
    for (k = 0; k < m->wordlen; k++) {
      int letter = abs(m->word[k]);
      word = letter > 26 ? letter-26 + 'a' -1 : letter + 'A' - 1;
      if (m->word[k] < 0)
        XDrawString(d, wm, green, 10+17*14+120+k*8, 50+j*14, &word, 1);
      else
        XDrawString(d, wm, gc, 10+17*14+120+k*8, 50+j*14, &word, 1);
    }
    sprintf(s, "%4d(%d) [%3d]", i, wo->ml->size, m->score);
    XDrawString(d, wm, gc, 10+17*14, 50+j*14, s, strlen(s));
  }

  sprintf(s, "score: %4d", wo->g->score);
  XDrawString(d, wm, gc, 10, 55+16*14, s, strlen(s));

  XCopyArea(d, wm, w, gc, 0, 0, 900, 300, 0, 0);

  if (curline != -1 && startline + curline >= 0 &&
      startline + curline < wo->ml->size)
    unplay_move(wo->g, &wo->ml->v[startline+curline]);
}

void dig(game *g)
{
  letterstack stack = { stacklen : 0 };
  movelist ml = { NULL, 0, 0 };
  world wo;
  XEvent ev;

  wo.g = g;
  wo.s = &stack;
  wo.ml = &ml;
  get_dmoves(&wo);

  d = XOpenDisplay(0); if (!d) abort();
  w = XCreateSimpleWindow(d, DefaultRootWindow(d), 0, 0, 900, 300, 0, 0, 0);
  wm = XCreatePixmap(d, w, 900, 300, DefaultDepth(d, DefaultScreen(d)));
  gc = DefaultGC(d, DefaultScreen(d));
  XGCValues val;
  Font font = XLoadFont(d,
             "-misc-fixed-medium-r-normal-*-14-130-75-75-c-70-iso10646-1");
  val.foreground = 0xaaaaaa;
  val.font = font;
  XChangeGC(d, gc, GCForeground | GCFont, &val);
  XMapWindow(d, w);
  XSelectInput(d, w,
      KeyPressMask | ExposureMask | ButtonPressMask | PointerMotionMask);
  red = XCreateGC(d, w, 0, 0);
  green = XCreateGC(d, w, 0, 0);
  black = XCreateGC(d, w, 0, 0);
  blue = XCreateGC(d, w, 0, 0);
  yellow = XCreateGC(d, w, 0, 0);
  XCopyGC(d, gc, -1L, red);
  XCopyGC(d, gc, -1L, green);
  XCopyGC(d, gc, -1L, black);
  XCopyGC(d, gc, -1L, blue);
  XCopyGC(d, gc, -1L, yellow);
  val.foreground = 0xff0000; XChangeGC(d, red, GCForeground, &val);
  val.foreground = 0x00ff00; XChangeGC(d, green, GCForeground, &val);
  val.foreground = 0x000000; XChangeGC(d, black, GCForeground, &val);
  val.foreground = 0x0000ff; XChangeGC(d, blue, GCForeground, &val);
  val.foreground = 0xffff00; XChangeGC(d, yellow, GCForeground, &val);
  while (1) {
    int todraw = 0;
    int x, y, button;
    XNextEvent(d, &ev);
    switch (ev.type) {
    case Expose: todraw = 1; break;
    case MotionNotify:
      todraw = 1;
      x = ev.xmotion.x;
      y = ev.xmotion.y;
      /* hand */
      if (x >= 10 && x <= 10 + 7 * 8 && y >= 10 && y <= 20) 
        hselect = (x-10) / 8;
      else hselect = -1;
      if (y >= 50-14 && y <= 50-14+15*14) curline = (y-50+14)/14;
      else curline = -1;
      if (x >= 10 && x <= 10 + 100 * 8 && y >= 250 && y <= 264)
        bselect = (x-10) / 8;
      else bselect = -1;
      break;
    case ButtonPress:
      todraw = 1;
      button = ev.xbutton.button;
      if (button == 4) /* scroll up */
        { startline -= 5; if (startline < 0) startline = 0; }
      if (button == 5) /* scroll down */
        { startline+=5; if (startline>=wo.ml->size) startline=wo.ml->size-1;}
      if (button == 1 && hselect != -1) {
        if (redhand[hselect] == -1) {
          redhand[hselect] = wo.g->hand[hselect];
          wo.g->hand[hselect] = -1;
        } else {
          wo.g->hand[hselect] = redhand[hselect];
          redhand[hselect] = -1;
        }
        wo.ml->size = 0; get_dmoves(&wo);
        startline = 0;
      }
      if (button == 1 && bselect != -1) {
        bagcol[bselect]++;
        if (bagcol[bselect] == 3) bagcol[bselect] = 0;
      }
      if (button == 1 && curline != -1 && startline + curline >= 0 &&
          startline + curline < wo.ml->size) {
        int i;
        for (i=0;i<7;i++) if (redhand[i]!=-1)
          { wo.g->hand[i] = redhand[i]; redhand[i] = -1; }
        wo.g->moves[wo.g->movesize] = wo.ml->v[startline + curline];
        wo.g->moves[wo.g->movesize].a = startline;
        play_move(wo.g, &wo.g->moves[wo.g->movesize]);
        print_moves(wo.g);
        wo.ml->size = 0; get_dmoves(&wo);
        startline = 0;
      }
      if (button == 3 && wo.g->movesize) {
        int i;
        for (i=0;i<7;i++) if (redhand[i]!=-1)
          { wo.g->hand[i] = redhand[i]; redhand[i] = -1; }
        startline = wo.g->moves[wo.g->movesize-1].a;
        unplay_move(wo.g, &wo.g->moves[wo.g->movesize-1]);
        wo.ml->size = 0; get_dmoves(&wo);
      }
      break;
    case KeyPress:
      todraw = 1;
      if (ev.xkey.keycode == 111) /* up */
        { startline--; if (startline < 0) startline = 0; }
      if (ev.xkey.keycode == 116) /* down */
        { startline++; if (startline>=wo.ml->size) startline = wo.ml->size-1;}
      break;
    }
    if (todraw) show_gui(&wo);
  }
  exit(1);
}
