#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include "dict.h"
#include "scrabble.h"

extern letter english_letters[];

typedef struct {
  dict_letter *l;
  int hl;         /* hand's value (to check for jokers) */
  int i, j;
} stack_elem;

typedef struct {
  /* solver state, from main.c */
  int board[15][15];
  int is_joker[15][15];
  int hand[7];
  int starti, startj, inci, incj;
  stack_elem stack[15];
  int stacklen;
  int nb_input_letters;
  /* local state, from contest.c */
  int next_letter;
  int cumul_score;
  int steps[100];
  int steps_size;
  int wordscore;
  int nb_letters_used;
  double gain_factor;
  int current_total_score;  /* unused? */
  /* local data, not saved/restored in push/pop (be careful with that) */
  int expected_score;
} game_state;

#define G_STEP_I(x)        ( (x)       & 0x0f)
#define G_STEP_J(x)        (((x) >> 4) & 0x0f)
#define G_STEP_LEN(x)      (((x) >> 8) & 0x0f)
#define G_STEP_VERTICAL(x) (((x) >> 12) & 0x01)
#define G_STEP_SCORE(x)    (((x) >> 16) & 0xffff)

#define S_STEP_I(x, v)        x = ((x) &   ~0x0f) |  (v)
#define S_STEP_J(x, v)        x = ((x) &   ~0xf0) | ((v) << 4)
#define S_STEP_LEN(x, v)      x = ((x) &  ~0xf00) | ((v) << 8)
#define S_STEP_VERTICAL(x, v) x = ((x) & ~0x1000) | ((v) << 12)
#define S_STEP_SCORE(x, v)    x = ((x) & ~0xffff0000) | ((v) << 16)

static char letters[200];
static int translated_letters[200];

/* local state */
static int next_letter;
static int cumul_score;
int steps[100];
int steps_size;
int wordscore;
int nb_letters_used;
double gain_factor;
int current_total_score;

game_state game_stack[1024];
int game_stack_size;

/* solver state */
extern int board[15][15];
extern int is_joker[15][15];
extern int hand[7];
extern int starti, startj, inci, incj;
extern stack_elem stack[15];
extern int stacklen;
extern int nb_input_letters;

int my_isalpha(int n);
int up(int n);

void add_step(int i, int j, int inci, int len, int score)
{
  if (steps_size == 100) abort();
  steps[steps_size] = 0;
  S_STEP_I(steps[steps_size], i);
  S_STEP_J(steps[steps_size], j);
  S_STEP_VERTICAL(steps[steps_size], inci);
  S_STEP_LEN(steps[steps_size], len);
  S_STEP_SCORE(steps[steps_size], score);
  steps_size++;
}

void translate_letters(void)
{
  int i;
  for (i = 0; letters[i]; i++) {
    int c = letters[i];
    if (c == '?') c = 0;                 /* joker */
    else { if (!my_isalpha(c)) abort(); c = up(c); }
    translated_letters[i] = c;
  }
}

void push_game(void)
{
  game_state *g;
  if (game_stack_size == 1024) abort();
  g = &game_stack[game_stack_size];
  game_stack_size++;
  /* solver state */
  memcpy(g->board, board, 15*15*sizeof(int));
  memcpy(g->is_joker, is_joker, 15*15*sizeof(int));
  memcpy(g->hand, hand, 7*sizeof(int));
  g->starti = starti;
  g->startj = startj;
  g->inci = inci;
  g->incj = incj;
  memcpy(g->stack, stack, 15*sizeof(stack_elem));
  g->stacklen = stacklen;
  g->nb_input_letters = nb_input_letters;
  /* local state */
  g->next_letter = next_letter;
  g->cumul_score = cumul_score;
  memcpy(g->steps, steps, 100*sizeof(int));
  g->steps_size = steps_size;
  g->wordscore = wordscore;
  g->nb_letters_used = nb_letters_used;
  g->gain_factor = gain_factor;
  g->current_total_score = current_total_score;
}

void pop_game(void)
{
  game_state *g;
  if (game_stack_size == 0) abort();
  game_stack_size--;
  g = &game_stack[game_stack_size];
  /* solver state */
  memcpy(board, g->board, 15*15*sizeof(int));
  memcpy(is_joker, g->is_joker, 15*15*sizeof(int));
  memcpy(hand, g->hand, 7*sizeof(int));
  starti = g->starti;
  startj = g->startj;
  inci = g->inci;
  incj = g->incj;
  memcpy(stack, g->stack, 15*sizeof(stack_elem));
  stacklen = g->stacklen;
  nb_input_letters = g->nb_input_letters;
  /* local state */
  next_letter = g->next_letter;
  cumul_score = g->cumul_score;
  memcpy(steps, g->steps, 100*sizeof(int));
  steps_size = g->steps_size;
  wordscore = g->wordscore;
  nb_letters_used = g->nb_letters_used;
  gain_factor = g->gain_factor;
  current_total_score = g->current_total_score;
}

extern dictionary dict;

int is_letter(int i, int j);
void read_hand(char *h);
void go(dict_letter *cur, int i, int j);

char eat_letter(void)
{
  char r = letters[next_letter];
  if (!r) return '.';
  next_letter++;
  if (r == '?') return '_';
  return r;
}

#define NB_RAND 5

/* list to store found words by solver */
#define MAXLIST 1000

static game_state *list;
static int list_size;
static int list_maxsize;

int cmp_list(const void *_a, const void *_b)
{
  const game_state *a = _a;
  const game_state *b = _b;
  double va, vb;

  /* order by cumul_score */
  if (a->cumul_score < b->cumul_score) return -1;
  if (a->cumul_score > b->cumul_score) return 1;
  /* other metrics? */
  return 0;

  va = a->current_total_score * 10000 + a->cumul_score;
  vb = b->current_total_score * 10000 + a->cumul_score;

  if (va < vb) return -1;
  if (va > vb) return 1;
  return 0;

  if (a->gain_factor < b->gain_factor) return -1;
  if (a->gain_factor > b->gain_factor) return 1;
  if (a->cumul_score < b->cumul_score) return -1;
  if (a->cumul_score > b->cumul_score) return 1;
  return 0;

  if (a->cumul_score < b->cumul_score) return -1;
  if (a->cumul_score > b->cumul_score) return 1;
  if (a->next_letter > b->next_letter) return -1;
  if (a->next_letter < b->next_letter) return 1;
return 0;

  /* order by cumul_score */
  if (a->cumul_score < b->cumul_score) return -1;
  if (a->cumul_score > b->cumul_score) return 1;
  /* other metrics? */
  return 0;
}

void sort_list(void)
{
  qsort(list, list_size, sizeof(game_state), cmp_list);
}

void put_in_list(void)
{
  if (list_size == list_maxsize) {
    list_maxsize += 1024;
    list = realloc(list, list_maxsize * sizeof(game_state)); if(!list)abort();
  }
  push_game();
  memcpy(&list[list_size], &game_stack[game_stack_size-1], sizeof(game_state));
  pop_game();
  list_size++;

#if 0
  if (list_maxsize == 0) {
    list_maxsize += MAXLIST;
    list = realloc(list, list_maxsize * sizeof(game_state)); if(!list)abort();
  }
  if (list_size == MAXLIST) {
    int i;
    int min;
    push_game();
    min = 0;
    for (i = 1; i < MAXLIST; i++)
      if (cmp_list(&list[i], &list[min]) < 0) min = i;
    if (cmp_list(&game_stack[game_stack_size-1], &list[min]) > 0)
      memcpy(&list[min], &game_stack[game_stack_size-1], sizeof(game_state));
    pop_game();
    return;
  }
  push_game();
  memcpy(&list[list_size], &game_stack[game_stack_size-1], sizeof(game_state));
  pop_game();
  list_size++;
#endif
}

void put_in_list_random_run(void)
{
  if (list_size == list_maxsize) {
    list_maxsize += 1024;
    list = realloc(list, list_maxsize * sizeof(game_state)); if(!list)abort();
  }
  push_game();
  memcpy(&list[list_size], &game_stack[game_stack_size-1], sizeof(game_state));
  pop_game();
  list_size++;
}

int timeout_level = 1;
volatile int timed_out;

void signal_timeout(int _)
{
  //printf("timeout!\n");
  //timed_out = 1;
  //alarm(1);
}

int cmp_random_runs(const void *_a, const void *_b)
{
  const game_state *a = _a;
  const game_state *b = _b;
  double va, vb;

  va = a->expected_score;
  vb = b->expected_score;

  if (va < vb) return -1;
  if (va > vb) return 1;
  return 0;
}

int random_run(game_state *g)
{
  game_state random_state;
  int i, j;
  game_state *save_list;
  int save_list_size, save_list_maxsize;
  int expect = g->current_total_score;
  int len;

  save_list = list;
  save_list_size = list_size;
  save_list_maxsize = list_maxsize;

    push_game();

  len = 0;

more:
  len++; if (len > 4) goto over;

  list = 0;
  list_size = list_maxsize = 0;
  
    /* to set game state to the one in the list, simulate push and then pop */
    if (game_stack_size == 1024) abort();
    memcpy(&game_stack[game_stack_size], g, sizeof(game_state));
    game_stack_size++;

    pop_game();

    stacklen = 0;

    /* run next hand */
    inci = 0; incj = 1;
    for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) {
      starti = i; startj = j;
      /* must be a start of a word (no letter before) */
      if (is_letter(starti-inci, startj-incj)) continue;
      go(dict.top, i, j);
    }
    inci = 1; incj = 0;
    for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) {
      starti = i; startj = j;
      /* must be a start of a word (no letter before) */
      if (is_letter(starti-inci, startj-incj)) continue;
      go(dict.top, i, j);
    }

  if (list_size) {
    /* choose a random element in the list and go to it */
    int i = (double)rand()/RAND_MAX * list_size;
    if (i == list_size) i--;
#if 0
/* take best (according to sorting criterium) */
sort_list();
i = list_size-1;
#endif
    memcpy(&random_state, &list[i], sizeof(game_state));
    g = &random_state;
    expect = g->current_total_score;
    free(list);
    goto more;
  }
  if (list_maxsize) free(list);

over:
  list = save_list;
  list_size = save_list_size;
  list_maxsize = save_list_maxsize;

    pop_game();

printf("random run done\n");
  return expect;
}

game_state best_list[MAXLIST];
int best_list_size;

void do_list(int level)
{
  game_state *_list;
  int _list_size, _list_maxsize;
  int i, j;
  int k;

more:
  sort_list();

  _list = list;
  _list_size = list_size;
  _list_maxsize = list_maxsize;

  list = 0;
  list_size = list_maxsize = 0;

  if (_list_size == 0) return;

  printf("level %d\n", level);

  int game_count = 0;
  for (k = _list_size-1; k>=0 /*&& game_count < 10*/; k--, game_count++) {
    int l;
    int max = -100000;
    for (l = 0; l < NB_RAND; l++) {
printf("level %d random %d for k %d\n", level, l, k);
      int expected = random_run(&_list[k]);
      if (expected > max) max = expected;
      _list[k].expected_score = expected;
    }
  }

  qsort(_list, _list_size, sizeof(game_state), cmp_random_runs);

  /* beware! best_list is ordered descending (best nodes first)
   * lists using qsort are ordered ascending!
   */
  best_list_size = 0;
  for (k=_list_size-1; k>=0 && best_list_size<MAXLIST;k--,best_list_size++)
    memcpy(&best_list[best_list_size], &_list[k], sizeof(game_state));

  if (_list_maxsize) free(_list);

  /* we use best_list, so ascending order to deal with best nodes first! */
  /* (well, I think we don't care in fact...) */
  for (k = 0; k < best_list_size; k++) {
    game_state *g;

    push_game();

    /* to set game state to the one in the list, simulate push and then pop */
    if (game_stack_size == 1024) abort();
    g = &best_list[k];
    memcpy(&game_stack[game_stack_size], g, sizeof(game_state));
    game_stack_size++;

    pop_game();

    stacklen = 0;

    /* run next hand */
    inci = 0; incj = 1;
    for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) {
      starti = i; startj = j;
      /* must be a start of a word (no letter before) */
      if (is_letter(starti-inci, startj-incj)) continue;
      if (timed_out && timeout_level <= level) goto over;
      go(dict.top, i, j);
    }
    inci = 1; incj = 0;
    for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) {
      starti = i; startj = j;
      /* must be a start of a word (no letter before) */
      if (is_letter(starti-inci, startj-incj)) continue;
      if (timed_out && timeout_level <= level) goto over;
      go(dict.top, i, j);
    }

over:
    pop_game();

    /* timeout not handled, useless code, left in case... */
    if (timed_out && timeout_level <= level) break;
  }

  level++;
  goto more;
}

void contest(char *bag)
{
  char cur_hand[8];
  int i, j;

  signal(SIGALRM, signal_timeout);
  alarm(5);

  if (strlen(bag) > 199) abort();
  strcpy(letters, bag);
  translate_letters();

  for (i = 0; i < 7; i++) cur_hand[i] = eat_letter();
  cur_hand[7] = 0;
  read_hand(cur_hand);

  inci = 0; incj = 1;
  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) {
    starti = i; startj = j;
    /* must be a start of a word (no letter before) */
    if (is_letter(starti-inci, startj-incj)) continue;
    go(dict.top, i, j);
  }

  sort_list();
  do_list(0);
}

int best_score = -10000000;

int total_score(int score)
{
  int i;
  for (i = 0; i < 7; i++)
    if (hand[i] < 0x1000000) score -= english_letters[hand[i]].value;
  for (i = next_letter; letters[i]; i++)
    score -= english_letters[translated_letters[i]].value;
  return score;
}

void print_board(void)
{
  int i, j;
  for (i = 0; i < 15; i++) {
    for (j = 0; j < 15; j++) printf("%c", board[i][j]!=-10000000 ? 'A' + board[i][j]-1 : '.');
    printf("\n");
  }
}

void print_steps(void)
{
  int i, j;
  for (i = 0; i < steps_size; i++) {
    int row = G_STEP_I(steps[i]);
    int col = G_STEP_J(steps[i]);
    int len = G_STEP_LEN(steps[i]);
    int vert = G_STEP_VERTICAL(steps[i]);
    int score = G_STEP_SCORE(steps[i]);
    int incrow, inccol;
    if (!vert) {
      printf("%d%c ", row+1, col+'A');
      incrow = 0;
      inccol = 1;
    } else {
      printf("%c%d ", col+'A', row+1);
      incrow = 1;
      inccol = 0;
    }
    for (j = 0; j < len; j++, row += incrow, col += inccol) {
      int c = board[row][col];
      int joker = is_joker[row][col];
      printf("%c", c+'A'-1 + joker * ('a'-'A'));
    }
    printf(" %d", score);
    if (i == steps_size-1) printf("\n"); else printf(",");
  }
}

void put_word(int score, int _starti, int _startj, int vert, int *letters)
{
  int i;
  int len;
  int letters_raw_score;

  push_game();

  wordscore = score;
  cumul_score += score;

  nb_letters_used = 0;
  letters_raw_score = 0;

  /* "play" on the board */
  /* board[i][j] = letter (A=1) or -10000000 if empty */
  /* is_joker[i][j] = 1 if letter at (i,j) is a joker */
  /* hand[k] = 1 (A) .. 26 (Z), or 0 for joker or 0x1000000 if no letter
   * + 0x1000000 if used
   */
  for (i = 0; i < stacklen; i++) {
    if (board[stack[i].i][stack[i].j] == -10000000) {
      nb_letters_used++;
      letters_raw_score += english_letters[stack[i].hl].value;
    }
    board[stack[i].i][stack[i].j] = stack[i].l->l;
    if (stack[i].hl == 0) is_joker[stack[i].i][stack[i].j] = 1;
  }
  gain_factor = (double)score/letters_raw_score;
  /* recompute hand */
  nb_input_letters = 0;
  for (i = 0; i < 7; i++) {
    int c;
    if (hand[i] < 0x1000000) { nb_input_letters++; continue; }
    c = eat_letter();
    if (c != '.') nb_input_letters++;
    if (c == '_') c = 0;                 /* joker */
    else if (c == '.') c = 0x1000000;  /* no letter, mark it as used */
    else { if (!my_isalpha(c)) abort(); c = up(c); }
    hand[i] = c;
  }

  /* compute len of played word */
  for (len = 0; letters[len]; len++) /* do nothing */;
  add_step(_starti, _startj, vert, len, score);

  current_total_score = total_score(cumul_score);
  if (current_total_score > best_score) {
    best_score = current_total_score;
    printf("current score %d (next_letter %d score %d)\n", best_score, next_letter, cumul_score);
    print_board();
    print_steps();
  }

  put_in_list();

  pop_game();
}
