#include "dict.h"
#include "defs.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

dictionary dict;
wordlist wlist[16];

void usage(void)
{
  printf("usage: [options] <dictionary> <bag>\n");
  printf("options:\n");
  printf("    -preplay <moves>\n");
  printf("        preplays the given moves\n");
  exit(1);
}

int main(int n, char **v)
{
  game g;
  char *fdict = NULL;
  char *bag = NULL;
  char *premoves = NULL;
  int i;
  int wordcount;

  for (i = 1; i < n; i++) {
    if (!strcmp(v[i], "-preplay")) {if(i>n-2)usage();premoves=v[++i];continue;}
    if (fdict == NULL) { fdict = v[i]; continue; }
    if (bag == NULL) { bag = v[i]; continue; }
    usage();
  }
  if (fdict == NULL || bag == NULL) usage();

  new_dict(&dict);
  read_dict(fdict, &dict, 0);
  for (i = 2; i <= 15; i++) {
    init_wordlist(&wlist[i], i);
    dict_to_wordlist(&dict, &wlist[i], i);
  }

  wordcount = 0;
  for (i = 2; i <= 15; i++) {
    printf("words of length %d [%d]\n", i, wlist[i].size);
    wordcount += wlist[i].size;
  }
  printf("%d words in dict\n", wordcount);

  init_game(&g, bag);

  void dig(game *);
  dig(&g);

  void oxy(game *);
  oxy(&g);
exit(1);

  add_constraint_wordat(&g, WORDAT_ABSOLUTE, char2int("oxyphenbutazone"),
                        15, 0, 0);
  /* upper tripples must be played together */
  add_constraint_play_together(&g, 0, 0, 0, 7);
  add_constraint_play_together(&g, 0, 0, 0, 15);
  /* lower tripples must be played together */
  add_constraint_play_together(&g, 15, 0, 15, 7);
  add_constraint_play_together(&g, 15, 0, 15, 15);

  if (premoves != NULL) game_preplay(&g, premoves);

  print_board(g.board);
  print_hand(g.hand);
  printf("score %d\n", g.score);

  movelist l;
  init_movelist(&l);
  get_moves(&g, &l, &dict);
  printf("%d moves\n", l.size);
  for (i = 0; i < l.size; i++) {
    int k;
    for (k = 0; k < l.v[i].wordlen; k++) {
      int val = abs(l.v[i].word[k]) - 1;
      if (val >= 26) val = val - 26 + 'a';
      else          val = val + 'A';
      printf("%c", val);
    }
    printf("\t\trow %d col %d vert %d\t", l.v[i].row, l.v[i].col, l.v[i].vert);
    for (k = 0; k < 7; k++) printf(" %2d", l.v[i].next_hand[k]);
    printf("\t%d", l.v[i].score);
    printf("\n");
  }

  solve_constraints(&g);

  return 0;
}
