#include "defs.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdint.h>

#if 0
static int letterat(int board[15][15], int row, int col)
{
  return row >= 0 && row < 15 && col >= 0 && col < 15 && board[row][col] != 0;
}
#endif
static int is_positive_letter(int board[15][15], int row, int col)
{
  return row >= 0 && row < 15 && col >= 0 && col < 15 && board[row][col] > 0;
}

typedef struct {
  int w[15];
  int wlen;
} word_t;

typedef struct {
  int row, col, vert;
  int w[15];
  int wlen;
  word_t *words;
  int wordsize;
  int wordmaxsize;
  int current_word;
} path_item;

typedef struct {
  path_item p[4];
  int size;
  int preboard[15][15];
  int prelcount[27];
} path;

void add_pathword(path_item *p, int *w, int wlen)
{
  int i;
  if (p->wordsize == p->wordmaxsize) {
    p->wordmaxsize += 128;
    p->words = realloc(p->words, p->wordmaxsize * sizeof(word_t));
    if (p->words == NULL) abort();
  }
  for (i = 0; i < wlen; i++)
    p->words[p->wordsize].w[i] = w[i];
  p->words[p->wordsize].wlen = wlen;
  p->wordsize++;
}

void fill_board(game *g, path *p)
{
  int pi;
  memcpy(g->board, p->preboard, 15*15*sizeof(int));
  memcpy(g->lcount, p->prelcount, 27*sizeof(int));
  for (pi = 0; pi < p->size; pi++) {
    path_item *it = p->p + pi;
    int row, col, incrow, inccol, l;
    if (it->vert) { incrow = 1; inccol = 0; } else { incrow = 0; inccol = 1; }
    for (row=it->row,col=it->col,l=0;l<it->wlen;l++,row+=incrow,col+=inccol) {
      if (g->board[row][col] == 0) g->lcount[it->w[l]]--;
      g->board[row][col] = -it->w[l];
    }
  }
}

static int target[15][15];
static int stack[15];
static int stacklen;

void push(int l)
{
  stack[stacklen++] = l;
}

void pop(void)
{
  stacklen--;
}

int wordcount;
void store_stackword(path_item *it)
{wordcount++;
}

static int other_word_ok(game *g, int l, int i, int j,
                         int inci, int incj)
{
  extern dictionary dict;
  int si = i, sj = j;
  int word[16];
  int w;
  int ii = 1 - inci;
  int jj = 1 - incj;
  int starti = i, startj = j;
  int endi, endj;
  g->board[si][sj] = l;
  /* find start/end pos of word */
  while (starti >= 0 && startj >= 0 && g->board[starti][startj] != 0)
    starti -= ii, startj -= jj;
  starti += ii; startj += jj;
  endi = starti; endj = startj;
  while (endi < 15 && endj < 15 && g->board[endi][endj] != 0)
    endi += ii, endj += jj;
  endi -= ii; endj -= jj;
  /* get word */
  for (w=0, i=starti, j=startj; i!=endi+ii || j!=endj+jj; w++, i+=ii, j+=jj) {
    int letter = abs(g->board[i][j]); if (letter > 26) letter -= 26;
    word[w] = letter;
  }
  word[w] = 0;
  g->board[si][sj] = 0;
  return w == 1 || find_word(&dict, word);
}

void go(game *g, path_item *it, dict_letter *l, int newlen, int on_target,
        int row, int col, int incrow, int inccol)
{
  if (newlen > 7) return;
  while (l != NULL) {
    if (g->board[row][col]) {
      /* must match */
      int letter = abs(g->board[row][col]);
      if (letter > 26) letter -= 26;
      if (letter != l->l) goto next;
      if (target[row][col]) on_target = 1;
      push(l->l);
      if (l->w && on_target &&
        (row+incrow==15||col+inccol==15||g->board[row+incrow][col+inccol]==0))
        store_stackword(it);
      go(g, it, l->child, newlen, on_target, row+incrow, col+inccol,
         incrow, inccol);
      pop();
    } else {
      /* one matching free letter must be here */
      if (!g->lcount[l->l]) goto next;
      if (!other_word_ok(g, l->l, row, col, incrow, inccol)) goto next;
      if (target[row][col]) on_target = 1;
      push(l->l);
      g->lcount[l->l]--;
      if (l->w && on_target &&
        (row+incrow==15||col+inccol==15||g->board[row+incrow][col+inccol]==0))
        store_stackword(it);
      go(g, it, l->child, newlen+1, on_target, row+incrow, col+inccol,
         incrow, inccol);
      g->lcount[l->l]++;
      pop();
    }
next:
    l = l->next;
  }
}

void get_words(game *g, path_item *it)
{
  extern dictionary dict;
  int row, col, incrow, inccol, l;
  memset(target, 0, sizeof(int) * 15 * 15);
  if (it->vert) { incrow = 1; inccol = 0; } else { incrow = 0; inccol = 1; }
  for (row=it->row,col=it->col,l=0;l<it->wlen;l++,row+=incrow,col+=inccol)
    target[row][col] = 1;
#if 0
  for (row = 0; row < 15; row++) for (col = 0; col < 15; col++) {
    if (!letterat(g->board, row-1, col))
      go(g, it, dict.top, 0, 0, row, col, 1, 0);
    if (!letterat(g->board, row, col-1))
      go(g, it, dict.top, 0, 0, row, col, 0, 1);
  }
#endif
  go(g, it, dict.top, 0, 0, 12, 4, 1, 0);
}

path *new_path(game *g, int row, int col, int vert, int *w, int wlen)
{
  path *p;
  int i;

  p = malloc(sizeof(path)); if (p == NULL) abort();
  p->size = 1;
  memcpy(p->preboard, g->board, 15*15*sizeof(int));
  memcpy(p->prelcount, g->lcount, 27*sizeof(int));
  p->p[0].row = row;
  p->p[0].col = col;
  p->p[0].vert = vert;

  for (i = 0; i < wlen; i++) p->p[0].w[i] = w[i];
  p->p[0].wlen = wlen;
  p->p[0].words = NULL;
  p->p[0].wordsize = p->p[0].wordmaxsize = 0;
  p->p[0].current_word = 0;

  fill_board(g, p);
  get_words(g, &p->p[0]);

printf("%d words\n", p->p[0].wordsize);
  return p;
}

/* return 0 if no more path */
int next_path(game *g, path *p)
{
  return 0;
}

void oxy(game *g)
{
  int i;
  path *hen /*, *uta, *on */;
  int HEN[3] = { 8, 5, 14 };
  int OXY[15] = { O, X, Y, P, H, E, N, B, U, T, A, Z, O, N, E };
  for (i = 0; i < 15; i++) {
    if (g->lcount[OXY[i]] == 0) abort();
    g->lcount[OXY[i]]--;
  }

  hen = new_path(g, 14, 4, 0, HEN, 3);
  fill_board(g, hen);
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

typedef struct {
  int row, col, vert;
  int len;
  int cur[15];
  int preboard[15];
} wordat;

typedef struct {
  wordat *w;
  int wlen;
  game *g;
  letterstack *s;
  movelist *ml;
} world;

int putit(world *w, wordat *at)
{
  int count[27];
  int row, col, incrow, inccol, l;
  if (at->vert) { incrow = 1; inccol = 0; } else { incrow = 0; inccol = 1; }
  /* check we can put it */
  for (l = 0; l < 27; l++) count[l] = w->g->lcount[l];
  for (row=at->row,col=at->col,l=0;l<at->len;l++,row+=incrow,col+=inccol) {
    if (w->g->board[row][col] && w->g->board[row][col]!=-at->cur[l]) return 0;
    if (w->g->board[row][col]) continue;
    if (count[at->cur[l]] == 0) return 0;
    count[at->cur[l]]--;
  }
  /* put it */
  for (row=at->row,col=at->col,l=0;l<at->len;l++,row+=incrow,col+=inccol) {
    at->preboard[l] = w->g->board[row][col];
    if (!at->preboard[l]) w->g->lcount[at->cur[l]]--;
    w->g->board[row][col] = -at->cur[l];
  }
  return 1;
}

void unputit(world *w, wordat *at)
{
  int row, col, incrow, inccol, l;
  if (at->vert) { incrow = 1; inccol = 0; } else { incrow = 0; inccol = 1; }
  for (row=at->row,col=at->col,l=0;l<at->len;l++,row+=incrow,col+=inccol) {
    w->g->board[row][col] = at->preboard[l];
    if (!at->preboard[l]) w->g->lcount[at->cur[l]]++;
  }
}

static int word_ok(game *g, int i, int j, int inci, int incj)
{
  extern dictionary dict;
  int word[16];
  int w;
  int ii = inci;
  int jj = incj;
  int starti = i, startj = j;
  int endi, endj;
  /* find start/end pos of word */
  while (starti >= 0 && startj >= 0 && g->board[starti][startj] != 0)
    starti -= ii, startj -= jj;
  starti += ii; startj += jj;
  endi = starti; endj = startj;
  while (endi < 15 && endj < 15 && g->board[endi][endj] != 0)
    endi += ii, endj += jj;
  endi -= ii; endj -= jj;
  /* get word */
  for (w=0, i=starti, j=startj; i!=endi+ii || j!=endj+jj; w++, i+=ii, j+=jj) {
    int letter = abs(g->board[i][j]); if (letter > 26) letter -= 26;
    word[w] = letter;
  }
  word[w] = 0;
  return w == 1 || find_word(&dict, word);
}

int word_possible(world *w, wordat *at)
{
  int row, col, incrow, inccol, l;
  row = at->row;
  col = at->col;
  if (at->vert) { incrow = 1; inccol = 0; } else { incrow = 0; inccol = 1; }
#if 0
  /* check we don't have something before/after */
  if (letterat(w->g->board, row-incrow, col-inccol)) return 0;
  if (letterat(w->g->board, row+incrow*at->len, col+inccol*at->len)) return 0;
#endif
  for (row=at->row,col=at->col,l=0;l<at->len;l++,row+=incrow,col+=inccol) {
    if (w->g->board[row][col] <= 0) {
      w->g->board[row][col] = at->cur[l];
      if (!word_ok(w->g, row,col,inccol,incrow)) return 0;
    } else {
      if (abs(w->g->board[row][col]) != at->cur[l]) return 0;
    }
  }
  if (!word_ok(w->g, at->row,at->col,incrow,inccol)) return 0;
  return 1;
}

int future_possible(world *w)
{
  int ret = 1;
  int b[15][15];
  int i;
  wordat a;

  memcpy(b, w->g->board, 15*15*sizeof(int));

  for (i = 0; i < w->wlen; i++) {
    wordat *at = w->w + i;
    if (!word_possible(w, at)) { ret = 0; goto end; }
  }

  a = (wordat){ 14, 0, 0, 15, {O,X,Y,P,H,E,N,B,U,T,A,Z,O,N,E} };
  if (!word_possible(w, &a)) { ret = 0; goto end; }

end:
  memcpy(w->g->board, b, 15*15*sizeof(int));
  return ret;
}

static void dpush(letterstack *s, game *g, dict_letter *n, int hl, int new)
{
  s->stack[s->stacklen].l = n;
  s->stack[s->stacklen].hl = hl;
  s->stack[s->stacklen].new = new;
  s->stacklen++;
}

static void dpop(letterstack *s)
{
  s->stacklen--;
}

static void check_word(game *g, letterstack *s, movelist *ml,
                int starti, int startj, int inci, int incj)
{
  move m;
  int i, j, l, ok = 0;
  /* word must contain at least 1 new letter */
  for (l=0, i=starti, j=startj; l<s->stacklen; l++, i+=inci, j+=incj)
    if (!is_positive_letter(g->board, i, j)) { ok = 1; break; }
  if (!ok) return;
  ok = 0;
  for (l=0, i=starti, j=startj; l<s->stacklen; l++, i+=inci, j+=incj)
    if (is_positive_letter(g->board, i, j) ||             /* letter on board */
        is_positive_letter(g->board, i-incj, j-inci) ||   /* touch on board */
        is_positive_letter(g->board, i+incj, j+inci) ||   /* touch on board */
        (i == 7 && j == 7))            /* word is at center pos (first word) */
    { ok = 1; break; }
  if (!ok) return;
  for (l = 0; l < 7; l++) m.next_hand[l] = g->hand[l];
  for (l = 0; l < s->stacklen; l++) {
    m.word[l] = s->stack[l].l->l;
    if (s->stack[l].hl == 0) m.word[l] += 26;
    if (s->stack[l].new) m.word[l] = -m.word[l];
  }
  m.wordlen = s->stacklen;
  m.row = starti;
  m.col = startj;
  m.vert = inci ? 1 : 0;
  m.score = -1; //score(g, m.word, m.wordlen, starti, startj, inci, incj);
  add_movelist(ml, &m);
}

void gogo(world *w, dict_letter *l,
          int starti, int startj, int i, int j, int inci, int incj)
{
  int h;
  int hletter;
  uint32_t letter_done = 0, joker_done = 0;

  if (l==NULL) return;
  if (i==15||j==15) return;
again:
  if (w->g->board[i][j] > 0) {
    int letter = w->g->board[i][j]; if (letter > 26) letter -= 26;
    if (l->l == letter) {
      dpush(w->s, w->g, l, w->g->board[i][j] > 26 ? 0 : w->g->board[i][j], 0);
      if (l->w && (i+inci==15||j+incj==15||w->g->board[i+inci][j+incj]==0))
        check_word(w->g, w->s, w->ml, starti, startj, inci, incj);
      gogo(w, l->child, starti, startj, i + inci, j + incj, inci, incj);
      dpop(w->s);
      goto next;
    }
    goto next;
  }
  if (w->g->board[i][j] < 0) {
    int letter = abs(w->g->board[i][j]); if (letter > 26) {letter-=26;abort();}
    if (l->l != letter) goto next;
    /* reserved letter, must be in hand */
    for (h = 0; h < 7; h++) if (w->g->hand[h] == -w->g->board[i][j]) break;
    if (h == 7) goto next;
    hletter = w->g->hand[h];
    if (letter_done & (1 << hletter)) goto next;
    /* other (already there) word must be ok */
    if (!word_ok(w->g,i, j, incj, inci)) goto next;
    w->g->hand[h] = -1;
    letter_done |= 1 << hletter;
    /* no lcount decrement, letter reservation has already done that */
    dpush(w->s, w->g, l, hletter, 1);
    if (l->w && (i+inci==15||j+incj==15||w->g->board[i+inci][j+incj]==0))
      check_word(w->g, w->s, w->ml, starti, startj, inci, incj);
    gogo(w, l->child, starti, startj, i+inci, j+incj, inci, incj);
    dpop(w->s);
    w->g->hand[h] = hletter;
    goto next;
  }

  /* no letter on board, no reserved letter */
  for (h = 0; h < 7; h++) if (w->g->hand[h]==l->l || w->g->hand[h]==0) break;
  if (h == 7) goto next;
  hletter = w->g->hand[h];
  /* must not have been tried before */
  if (hletter && (letter_done & (1 << hletter))) goto next;
  if (!hletter && (joker_done & (1 << l->l))) goto next;
  /* if not a joker, letter must be in lcount */
  if (/*hletter && */!w->g->lcount[hletter]) goto next;
  if (!other_word_ok(w->g, l->l, i, j, inci, incj)) goto next;
  w->g->hand[h] = -1;
  if (hletter) letter_done |= 1 << hletter;
  if (!hletter) joker_done |= 1 << l->l;
  w->g->lcount[hletter]--;
  dpush(w->s, w->g, l, hletter, 1);
  if (l->w && (i+inci==15||j+incj==15||w->g->board[i+inci][j+incj]==0))
    check_word(w->g, w->s, w->ml, starti, startj, inci, incj);
  gogo(w, l->child, starti, startj, i+inci, j+incj, inci, incj);
  dpop(w->s);
  w->g->lcount[hletter]++;
  w->g->hand[h] = hletter;

next:
  l = l->next;
  if (l != NULL) goto again;
}

void get_dmoves(world *w)
{
  extern dictionary dict;
  int row, col;
  for (row = 0; row < 15; row++) for (col = 0; col < 15; col++) {
    if (!is_positive_letter(w->g->board, row-1, col))
      gogo(w, dict.top, row, col, row, col, 1, 0);
    if (!is_positive_letter(w->g->board, row, col-1))
      gogo(w, dict.top, row, col, row, col, 0, 1);
  }
}

int move_score(const void *_a, const void *_b)
{
  const move *a = _a;
  const move *b = _b;
  if (a->score < b->score) return -1;
  if (a->score > b->score) return 1;
  return 0;
}

void compute_score(world *w, move *m)
{
  int i, j;
  int score = 0;
  play_move_noneg(w->g, m);
  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++)
    if (w->g->board[i][j] < 0) score--;
  m->score = score;
  unplay_move(w->g, m);
}

void depth(world *w, int level)
{
if (w->g->board[14][4] > 0 &&
    w->g->board[14][5] > 0 &&
    w->g->board[14][6] > 0 &&
    w->g->board[14][8] > 0 &&
    w->g->board[14][9] > 0 &&
    w->g->board[14][10] > 0 &&
    w->g->board[14][12] > 0 &&
    w->g->board[14][13] > 0) abort();
  int i;
  movelist ml = { NULL, 0, 0 };
  movelist *prev = w->ml;
  w->ml = &ml;
  get_dmoves(w);
  for (i = 0; i < ml.size; i++) compute_score(w, &ml.v[i]);
  qsort(ml.v, ml.size, sizeof(move), move_score);
  printf("%d moves [%d]\n", ml.size, level);
  for (i = ml.size-1; i >= 0; i--) {
    int k;
    printf("move %d [%d,%d,%s]", i,
           ml.v[i].row, ml.v[i].col, ml.v[i].vert ? "vert": "hori");
    for (k = 0; k < ml.v[i].wordlen; k++) {
      int letter = abs(ml.v[i].word[k]);
      int new = ml.v[i].word[k] < 0;
      int joker = abs(ml.v[i].word[k]) > 26;
      if (letter > 26) letter -= 26;
      if (new) printf("\x1b[32m");
      printf("%c", joker ? letter + 'a' - 1 : letter + 'A' -1);
      if (new) printf("\x1b[0m");
    }
    printf("\n");
fflush(stdout);
    play_move_noneg(w->g, &ml.v[i]);
    printf("hand: ");
    for (k = 0; k < 7; k++)
      printf("%c", w->g->hand[k]==-1 ? '.' :
                     w->g->hand[k] ? w->g->hand[k] + 'A' - 1 : '?');
    printf("\n");
    printf("bag(%d): ", w->g->bagpos);
    for (k = w->g->bagpos; k < 100; k++)
      printf("%c", w->g->bag[k] ? w->g->bag[k] + 'A' - 1 : '?');
    printf("\n");
    print_board(w->g->board);
    depth(w, level+1);
    unplay_move(w->g, &ml.v[i]);
  }
  if (ml.maxsize) free(ml.v);
  w->ml = prev;
}

void doit(world *w, int pos)
{
//printf("doit %d\n", pos);
  extern wordlist wlist[16];
  wordat *at;
  wordlist *wl;
  int i;
  if (pos == w->wlen) { depth(w, 0); return; }
#if 0
static int cc=0;
//w->g->board[13][0] = Z;
w->g->board[7][6] = 0;
printf("%d\n", future_possible(w)); exit(1);
printf("\r%d\n", ++cc); print_board(w->g->board); return;}
#endif
  at = w->w + pos;
  wl = wlist + at->len;
  for (i = 0; i < wl->size; i++) {
    memcpy(at->cur, wl->w[i], at->len * sizeof(int));
    if (putit(w, at)) { doit(w, pos+1); unputit(w, at); }
  }
}

void dig(game *g)
{
  letterstack stack = { stacklen : 0 };
  movelist ml = { NULL, 0, 0 };
  wordat ats[8] = {
    {  7,  7, 0, 3 },
    {  7,  9, 1, 4 },
    { 10,  7, 0, 6 },
    { 10,  7, 1, 3 },
    { 10, 12, 1, 5 },
    { 12,  4, 0, 5 },
    { 12,  4, 1, 3 },
    { 12,  8, 1, 3 },
  };
  world w;
  int oxy[15] = { O,X,Y,P,H,E,N,B,U,T,A,Z,O,N,E };
  int i;
  for (i = 0; i < 15; i++) {
    g->board[14][i] = -oxy[i];
    if (g->lcount[oxy[i]] == 0) abort();
    g->lcount[oxy[i]]--;
  }
  g->board[7][7] = -E;
  w.w = ats;
  w.wlen = 8;
  w.g = g;
  w.s = &stack;
  w.ml = &ml;
  doit(&w, 0);
  exit(1);
}
