#ifndef _DEFS_H_
#define _DEFS_H_

#include "dict.h"
#include "dictint.h"

typedef struct {
  unsigned char A:1;
  unsigned char B:1;
  unsigned char C:1;
  unsigned char x:5;
} charABC;

typedef struct {
  /* ABCxxxxx
   * xxxxx: letter
   * C: 0/1 no jpker/joker
   * B: 0 new letter, 1 was on board
   * A: unused
   */
  charABC letters[15];
  char letters_size;

  unsigned char vertical:1;
  unsigned char col:7;
  char row;

  /* hand as it is before the move (some letters may be undecided)
   * xxxxx: letter (0 for unset joker)
   * C: 0/1 no joker/joker
   * B: 1 if no decision on that letter
   * A: unused
   */
  charABC hand[7];

  /* bag position before the move */
  char bag_pos;
} move;

typedef struct {
  dict_letter *l;
  unsigned char hl;         /* hand's value (to check for jokers) */
  unsigned char i, j;
} stack_elem;

typedef struct {
  /* a char in the bag is:
   * ABCxxxxx
   * xxxxx: 1-26 for A-Z, 0 for unset joker
   * C: 0 for normal letter, 1 for joker (and xxxxx defines what it is set to)
   * B: 1 if letter used, 0 if free
   * A: not used
   */
  charABC *bag;
  char bag_size;
  char bag_count[27];

  int bag_pos;

  /* a char on the board is:
   * ABCxxxxx
   * xxxxx: 1-26 for A-Z, 0 for empty
   * C: 0/1 for normal letter/joker
   * B: 0 if available to put a letter, 1 if not usable
   *    (used by a later move, must be kept as is)
   * A: unused
   */
  charABC board[15][15];

  move moves[100];
  char moves_size;

  stack_elem stack[15];
  int stacklen;
  unsigned char starti, startj;

  int target_word_length;
} solver_state;

void play_word(solver_state *, charABC *, int len, int row, int col, int vert);
void push_state(solver_state *);
void pop_state(solver_state *);

void color_print(charABC board[15][15]);

extern dictionary dict;
extern dictionary dictX[16];
extern dictionary_int done_dict;

#endif /* _DEFS_H_ */
