#include <stdio.h>
#include <stdlib.h>
#include "defs.h"
#include "dictint.h"

int is_letter(solver_state *s, int row, int col)
{
  if (row < 0 || row > 14 || col < 0 || col > 14) return 0;
  if (s->board[row][col].x) return 1;
  return 0;
}

void put_word(solver_state *s, charABC *word, int len,
    int row, int col, int vert)
{
  int r, c, incr, incc;
  int i;

  if (vert) { incr = 1; incc = 0; } else { incr = 0; incc = 1; }
  for (r=row, c=col, i=0; i<len; i++, r+=incr, c+=incc) {
    s->board[row][col] = word[i];
    s->board[row][col].B = 1;
  }
}

#if 0
/* way too much cases... */
int get_pos(solver_state *s, int start, charABC letter)
{
  int i;
  for (i = start; i < s->bag_size; i++)
    if ((s->bag[i].x == letter.x || s->bag[i].C) && !s->bag[i].B) goto found;
  return -1;
found:
  s->bag[i].B = 1;
  s->bag[i].x = letter.x;
  return i;
}

int next_pos(solver_state *s, int *letter_pos, charABC *word, int len)
{
  int i = len-1;
  while (1) {
    if (letter_pos[i] != -1) s->bag[letter_pos[i]].B = 0;
    letter_pos[i] = get_pos(s, letter_pos[i]+1, word[i]);
    if (letter_pos[i] != -1) { i++; if (i == len) return 0; continue; }
    if (i == 0) goto no;
    i--;
  }
no:
  for (i = 0; i < len; i++)
    if (letter_pos[i] != -1) s->bag[letter_pos[i]].B = 0;
  return -1;
}

int first_pos(solver_state *s, int *letter_pos, charABC *word, int len)
{
  int i;
  for (i = 0; i < len; i++) {
    letter_pos[i] = get_pos(s, 0, word[i]); if (letter_pos[i] == -1) goto no;
  }
  return 0;
no:
  for (i = 0; i < len; i++)
    if (letter_pos[i] != -1) s->bag[letter_pos[i]].B = 0;
  return -1;
}
#endif

int get_letters(solver_state *s, charABC *word, int len,
                int row, int col, int vert)
{
  int i;
  int incrow, inccol;
  if (vert) { incrow = 1; inccol = 0; } else { incrow = 0; inccol = 1; }
  /* if letter is not already on board, we must take it from the bag */
  for (i = 0; i < len; i++, row+=incrow, col+=inccol) {
    if (s->board[row][col].x) {
      if (s->board[row][col].x != word[i].x) abort(); /* unnecessary? */
      word[i].C = s->board[row][col].C;
      continue;
    }
    /* not on board, try to take it from bag */
    word[i].C = 0;
    if (s->bag_count[word[i].x] == 0) {
      /* no more letter, try a joker */
      if (s->bag_count[0] == 0) return -1;
      s->bag_count[0]--;
      word[i].C = 1;
    } else
      s->bag_count[word[i].x]--;
  }
  return 0;
}

int holed_valid(charABC *w, int len)
{
  int x[16];
  int xlen;
  int i;
  xlen = 0;
  for (i = 0; i < len; i++) {
    x[xlen] = w[i].x;
    /* end of word */
    if (!x[xlen] || i == len-1) {
      if (!xlen) continue;
      if (xlen > 7) return 0;
      x[xlen+1] = 0;
      if (x[1] != 0 && !find_word(&dict, x)) return 0;
      xlen = -1;
    }
    xlen++;
  }
  return 1;
}

void reverse_play(solver_state *s, charABC *w, charABC *holed, int len,
    int row, int col, int vert, charABC *computed_hand)
{
  int i, inccol, incrow;
  //int letters_played = 0;
  move *m = &s->moves[(int)s->moves_size];
  s->moves_size++; if (s->moves_size == 101) abort();
  if (vert) { inccol = 0; incrow = 1; } else { inccol = 1; incrow = 0; }
  m->letters_size = len;
  m->vertical = vert;
  m->col = col;
  m->row = row;
  /* no decision on hand for now */
  for (i = 0; i < 7; i++) m->hand[i].B = 1;
  for (i = 0; i < len; i++, row+=incrow, col+=inccol) {
    m->letters[i] = w[i];
    if (holed[i].x != 0) m->letters[i].B = 1;
    s->board[row][col] = holed[i];
    if (holed[i].x == 0) s->board[row][col].B = 1;
    else                 s->board[row][col].B = 0;
#if 0
    /* played letter implies decision in hand */
    if (holed[i].x == 0) {
      m->hand[letters_played] = w[i];
      m->hand[letters_played].B = 0;
      letters_played++;
    }
#endif
  }
  for (i = 0; i < 7; i++) m->hand[i] = computed_hand[i];
  m->bag_pos = s->bag_pos;
}

void next_word(solver_state *s);

/* this function updates s->bag_pos and sets computed_hand */
int next_hand_compatible(solver_state *s, charABC *w, charABC *holed, int len,
    charABC *computed_hand)
{
  /* target[0] is counting blank, 1..26 for A-Z, 27 for undecided */
  int target[28];
  int i, h;
  move *next;
  int count;

  /* first reverse move has no future */
  if (s->moves_size == 0) return 1;

  for (i = 0; i < 28; i++) target[i] = 0;

  /* set target */
  next = &s->moves[s->moves_size-1];
  for (i = 0; i < 7; i++)
    if (next->hand[i].B) target[27]++;
    else if (next->hand[i].C) target[0]++;
    else target[next->hand[i].x]++;

  /* update bag pos */
  count = 0;
  for (i = 0; i < len; i++) if (holed[i].x == 0) count++;

  /* check that count previous letters in bag are compatible with target */
  if (count > s->bag_pos) return 0;

  for (i = 0; i < count; i++) {
    int pos;
    charABC l = s->bag[s->bag_pos-count+i];
    if (l.C) pos = 0;
    else pos = l.x;
    if (target[pos] == 0) {
      if (target[27] == 0) return 0;
      target[27]--;
    } else
      target[pos]--;
  }

  s->bag_pos -= count;

  /* set computed_hand */
  /* when calling, computed_hand[X].B == 1 for all X */
  h = 0;
  /* letters from next move and not coming from bag are already in hand */
  for (i = 0; i < 27; i++) {
    int k;
    for (k = 0; k < target[i]; k++, h++) {
      if (h==7) abort();
      computed_hand[h].x = i;
      if (i == 0) computed_hand[h].C = 1;
      computed_hand[h].B = 0;
    }
  }
  /* letters just played are also in hand*/
  for (i = 0; i < len; i++) {
    if (holed[i].x == 0) {
      if (h==7) abort();
      computed_hand[h] = w[i];
      computed_hand[h].B = 0;
      h++;
    }
  }

  return 1;
}

void holify(solver_state *s, charABC *w, int len, int row, int col, int vert)
{
  charABC holed[15];
  int nb_holes;
  int pos[7];
  int i, j, k;
  int valid;
  charABC computed_hand[7];

  for (nb_holes = 1; nb_holes <= 7 && nb_holes <= len; nb_holes++) {
    /* initial holes: at start of word */
    for (i = 0; i < nb_holes; i++) pos[i] = i;

next:
    /* deal with holes */
    for (i = 0; i < len; i++) holed[i] = w[i];
    for (i = 0; i < nb_holes; i++) holed[pos[i]].x = 0;
    valid = holed_valid(holed, len);
    /* first played word must have x3 letters holed */
    if (valid && s->moves_size == 0)
      if (holed[0].x || holed[7].x || holed[14].x) valid = 0;
    if (valid) {
      push_state(s);
#if 0
      /* put back unused letters in bag */
      for (i = 0; i < len; i++) if (holed[i].x) {
        if (holed[i].C) s->bag_count[0]++;
        else            s->bag_count[holed[i].x]++;
      }
#endif
      clear_dictint(&done_dict);
      /* clear decision on hand */
      for (i = 0; i < 7; i++) computed_hand[i].B = 1;
      /* next_hand_compatible/reverse_play must be played as is
       * because of computed_hand
       */
      if (next_hand_compatible(s, w, holed, len, computed_hand))
        reverse_play(s, w, holed, len, row, col, vert, computed_hand);
      color_print(s->board); printf("\n");
      next_word(s);
      pop_state(s);
    }

    /* compute next holes position */
    j = nb_holes-1;
retry:
    if (j == -1) goto no_next;
    pos[j]++;
    if (pos[j] == len) { j--; goto retry; }
    for (k = j+1; k < nb_holes; k++) {
      pos[k] = pos[j] + k - j;
      if (pos[k] == len) { j--; goto retry; }
    }
    goto next;
no_next:;
  }
}

void play_word(solver_state *s, charABC *word, int len,
    int row, int col, int vert)
{
  push_state(s);

  if (get_letters(s, word, len, row, col, vert) == 0) {
    holify(s, word, len, row, col, vert);
  }

  pop_state(s);
}

typedef struct {
  charABC w[15];
  int wlen;
  int starti, startj;
  int vert;
} word;

word *wordlist;
int wordlist_size;
int wordlist_maxsize;

static int printfuture;

void one_word(solver_state *s, charABC *w, int wlen, int starti, int startj,
              int vert)
{
  int i;
int dw[15+3+1]; dw[0] = starti+256; dw[1] = startj+256; dw[2] = vert+256;
for (i = 0; i < wlen; i++) dw[i+3] = w[i].x;
dw[i+3] = 0;
if (find_wordint(&done_dict, dw)) {
/*printf("already there\n");*/return;}
add_wordint(&done_dict, dw);
//if (starti==0&&startj==13&&wlen==3) {printfuture=1;check_future(s,w,wlen,starti,startj,vert);printfuture=0;}
  if (wordlist_size == wordlist_maxsize) {
    wordlist_maxsize += 2048;
    wordlist = realloc(wordlist, wordlist_maxsize * sizeof(word));
    if (!wordlist) abort();
  }
if (wlen > 15) abort();
  for (i = 0; i < wlen; i++)
    wordlist[wordlist_size].w[i] = w[i];
  wordlist[wordlist_size].wlen = wlen;
  wordlist[wordlist_size].starti = starti;
  wordlist[wordlist_size].startj = startj;
  wordlist[wordlist_size].vert = vert;
  wordlist_size++;
printf("%d         \r", wordlist_size); fflush(stdout);
}

int other_word_ok(solver_state *s, dict_letter *n, int i, int j,
                  int inci, int incj);

/* return 1 if future is possible after putting word w, 0 if not */
int check_future(solver_state *s, charABC *w, int wlen,
    int starti, int startj, int vert)
{
int si=starti,sj=startj;
  int i, j, inci, incj, k, len;
  dict_letter n;
  push_state(s);
  /* put new word (no check necessary, already done) */
  if (vert) { inci = 1; incj = 0; } else { inci = 0; incj = 1; }
  for (len=0, i=starti, j=startj; len<wlen; i+=inci, j+=incj, len++)
    s->board[i][j].x = w[len].x;
  /* put all other words, in order, up to the end or error */
  for (k = s->moves_size-1; k >= 0; k--) {
    move *m = &s->moves[k];
    starti = m->row;
    startj = m->col;
    if (m->vertical) { inci = 1; incj = 0; } else { inci = 0; incj = 1; }
    for (len=0,i=starti,j=startj; len<m->letters_size; i+=inci,j+=incj,len++) {
      n.l = m->letters[len].x;
      if (!other_word_ok(s, &n, i, j, inci, incj)) goto err;
      s->board[i][j] = m->letters[len];
    }
  }
if (printfuture)
if (si==0&&sj==13&&wlen==3) {color_print(s->board);printf("\n\nOKOK\n");}
  pop_state(s);
  return 1;
err:
//if (si==0&&sj==13) color_print(s->board);
  pop_state(s);
  return 0;
}

void check_word(solver_state *s, int inci, int incj)
{
  /* to be ok a word must include an already positioned letter or touch one
   * or be first
   * note that when check_word is called we already checked that putting
   * the letters on the board give existing word (or words)
   */
  charABC w[15];
  int wlen;
  int i, j, l, ok = 0;
  /* word must contain at least 1 new letter */
  for (l=0, i=s->starti, j=s->startj; l<s->stacklen; l++, i+=inci, j+=incj)
    if (!is_letter(s, i, j)) { ok = 1; break; }
  if (!ok) return;
  ok = 0;
  for (l=0, i=s->starti, j=s->startj; l<s->stacklen; l++, i+=inci, j+=incj)
    if (is_letter(s, i, j) ||             /* word contain letter on board */
        is_letter(s, i-incj, j-inci) ||   /* word touch letter on board */
        is_letter(s, i+incj, j+inci))   /* word touch letter on board */
//        (is_first))                    /* word is first */
//        (i == 7 && j == 7))            /* word is at center pos (first word) */
    { ok = 1; break; }
  if (!ok) return;
  for (l=0; l<s->stacklen; l++) {
    w[l].x = s->stack[l].l->l;
    w[l].C = 0;
    if (!s->stack[l].hl) w[l].C = 1;
  }
  wlen = s->stacklen;
  if (check_future(s, w, wlen, s->starti, s->startj, inci))
    one_word(s, w, wlen, s->starti, s->startj, inci);
}

void push(solver_state *s, dict_letter *n, int hl, int ni, int nj,
          int i, int j, int inci, int incj)
{
  s->stack[s->stacklen].l = n;
  s->stack[s->stacklen].hl = hl;
  s->stack[s->stacklen].i = i;
  s->stack[s->stacklen].j = j;
  s->stacklen++;
  /* if n->w we are in an end letter of a word in the dictionary
   * so if on the board there is nothing after current letter,
   * we check that the word is ok here at (i,j)
   */
  if (n->w && (ni==15 || nj==15 || s->board[ni][nj].x==0) &&
      s->stacklen == s->target_word_length)
    check_word(s, inci, incj);
}

void pop(solver_state *s) { s->stacklen--; }

int other_word_ok(solver_state *s, dict_letter *n, int i, int j,
                  int inci, int incj)
{
  int si = i, sj = j;
  int word[16];
  int w;
  int ii = 1 - inci;
  int jj = 1 - incj;
  int starti = i, startj = j;
  int endi, endj;
  s->board[si][sj].x = n->l;
  /* find start/end pos of word */
  while (starti >= 0 && startj >= 0 && s->board[starti][startj].x != 0)
    starti -= ii, startj -= jj;
  starti += ii; startj += jj;
  endi = starti; endj = startj;
  while (endi < 15 && endj < 15 && s->board[endi][endj].x != 0)
    endi += ii, endj += jj;
  endi -= ii; endj -= jj;
  /* get word */
  for (w=0, i=starti, j=startj; i!=endi+ii || j!=endj+jj; w++, i+=ii, j+=jj)
    word[w] = s->board[i][j].x;
  word[w] = 0;
  s->board[si][sj].x = 0;
  return w == 1 || find_word(&dict, word);
}

void go(solver_state *s, dict_letter *cur, int i, int j, int inci, int incj,
        int letter)
{
  /* this function works vertically or horizontally depending on inci/incj */
  /* if in the dictionary the letter at (i,j) matches one in cur,
   * the go to next pos on board and go down in the dictionary tree.
   * the word is checked in push
   */
  int h;

  if (cur == NULL) return;
  if (i == 15 || j == 15) return;
  /* if position if forbidden, don't use it */
  if (s->board[i][j].B && s->board[i][j].x == 0) return;
  /* if word too long, guess what? */
  if (letter == s->target_word_length) return;

  if (s->board[i][j].x) {
    /* must match board[i][j] */
    while (cur) {
      if (cur->l == s->board[i][j].x) {
        push(s, cur, cur->l, i + inci, j + incj, i, j, inci, incj);
        go(s, cur->child, i + inci, j + incj, inci, incj, letter+1);
        pop(s);
        return;
      }
      cur = cur->next;
    }
    return;
  }

  /* must match one of not-yet-used letters */
  for (h = 0; h <= 26; h++) {
    dict_letter *n = cur;
    charABC g;
    if (s->bag_count[h] == 0) continue;
    g.x = h; g.C = g.x ? 0 : 1;
    s->bag_count[h]--;
    while (n) {
      /* letter must match or g is joker
       * and (eventual) word in other direction must be ok */
      if ((n->l == g.x || g.x == 0) && other_word_ok(s,n,i,j,inci,incj)) {
        push(s, n, g.x, i + inci, j + incj, i, j, inci, incj);
        go(s, n->child, i + inci, j + incj, inci, incj, letter+1);
        pop(s);
      }
      n = n->next;
    }
    s->bag_count[h]++;
  }
}

int possible_play(solver_state *s, int i, int j, int inci, int incj,
                  int wordlen)
{
  int ret = 0;
  int put_letters = 0;
  int len = wordlen;
  if (i + inci * wordlen > 15) return 0;
  if (j + incj * wordlen > 15) return 0;
  for (; wordlen; wordlen--, i+=inci, j+=incj) {
if (i==7 && j==7) return 0;
    if (s->board[i][j].B) return 0;
    if (is_letter(s, i, j)) { ret = 1; put_letters++; }
    if (is_letter(s, i-incj, j-inci)) ret = 1;
    if (is_letter(s, i+incj, j+inci)) ret = 1;
  }
  if (len == put_letters) return 0;
//  if (is_letter(s, i, j)) return 1;
  return ret;
}

int reverse_wordlength(const void *_a, const void *_b)
{
  const word *a = _a;
  const word *b = _b;
  if (a->wlen > b->wlen) return -1;
  if (b->wlen > a->wlen) return 1;
  return 0;
}

void next_word(solver_state *s)
{
  int len[16];
  int i, j, inci, incj;
  int wordlen;
  wordlist_size = 0;
  inci = 0; incj = 1;
  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++)
  for (wordlen = 2; wordlen <= 15; wordlen++) {
    s->starti = i; s->startj = j;
    /* must be a start of a word (no letter before) */
    if (is_letter(s, i-inci, j-incj)) continue;
    if (!possible_play(s, i, j, inci, incj, wordlen)) continue;
printf("\t\t\t\tdoing %d %d %d    \r", i, j, wordlen); fflush(stdout);
    s->target_word_length = wordlen;
    go(s, dictX[wordlen].top, i, j, inci, incj, 0);
  }
  inci = 1; incj = 0;
  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++)
  for (wordlen = 2; wordlen <= 15; wordlen++) {
    s->starti = i; s->startj = j;
    /* must be a start of a word (no letter before) */
    if (is_letter(s, i-inci, j-incj)) continue;
    if (!possible_play(s, i, j, inci, incj, wordlen)) continue;
printf("\t\t\t\tdoing vert %d %d %d    \r", i, j, wordlen); fflush(stdout);
    s->target_word_length = wordlen;
    go(s, dictX[wordlen].top, i, j, inci, incj, 0);
  }
  for (i = 0; i <= 15; i++) len[i] = 0;
  for (i = 0; i < wordlist_size; i++) len[wordlist[i].wlen]++;
  printf("\x1b[1J\x1b[18Awordlist size %d [", wordlist_size);
  for (i = 0; i <= 15; i++) printf("%d:%d ", i, len[i]);
  printf("\n");

  word *list;
  int list_size;
  int list_maxsize;
  list = wordlist;
  list_size = wordlist_size;
  list_maxsize = wordlist_maxsize;
  wordlist = 0;
  wordlist_size = wordlist_maxsize = 0;

  qsort(list, list_size, sizeof(word), reverse_wordlength);

  for (i = 0; i < list_size; i++) {
    play_word(s, list[i].w, list[i].wlen, list[i].starti, list[i].startj,
              list[i].vert);
  }
  if (list_maxsize) free(list);
}
