#include <stdio.h>
#include <stdlib.h>

int (*words)[15];
int words_size;

typedef struct {
  int word1, word2, distance, score1, score2;
} distance;

distance *d;

int fail(void) { abort(); }
void read_words(char *fn)
{
  char l[256];
  int i;
  FILE *f = fopen(fn, "r"); if (!f) { perror(fn); exit(1); }
  while (1) {
    if (fgets(l, 256, f) == NULL) break;
    words_size++;
    words = realloc(words, words_size * 15 * sizeof(int));
    if (words == NULL) abort();
    for (i = 0; i<15; i++) words[words_size-1][i] =
      l[i] >= 'a' && l[i] <= 'z' ? l[i] - 'a' + 1 :
      l[i] >= 'A' && l[i] <= 'Z' ? l[i] - 'A' + 1 :
      fail();
  }
  fclose(f);
}

int dist(int a, int b)
{
  int common_letters = 0;
  int i;
  int l[27] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 0 };
  for (i = 0; i < 15; i++) l[words[a][i]]++;
  for (i = 0; i < 15; i++) {
    if (l[words[b][i]] == 0) continue;
    l[words[b][i]]--;
    common_letters++;
  }
  return common_letters;
}

int order(const void *_a, const void *_b)
{
  const distance *a = _a;
  const distance *b = _b;
  if (a->distance < b->distance) return -1;
  if (a->distance > b->distance) return 1;
  if (a->score1 + a->score2 > b->score1 + b->score2) return -1;
  if (a->score1 + a->score2 < b->score1 + b->score2) return 1;
  return 0;
}

typedef struct {
  int car;
  int value;
  int count;
} letter;

letter letters[] = {
  {   0, 0, 2 },           /* joker */
  { 'A', 1, 9 },
  { 'B', 3, 2 },
  { 'C', 3, 2 },
  { 'D', 2, 4 },
  { 'E', 1, 12 },
  { 'F', 4, 2 },
  { 'G', 2, 3 },
  { 'H', 4, 2 },
  { 'I', 1, 9 },
  { 'J', 8, 1 },
  { 'K', 5, 1 },
  { 'L', 1, 4 },
  { 'M', 3, 2 },
  { 'N', 1, 6 },
  { 'O', 1, 8 },
  { 'P', 3, 2 },
  { 'Q', 10, 1 },
  { 'R', 1, 6 },
  { 'S', 1, 4 },
  { 'T', 1, 6 },
  { 'U', 1, 4 },
  { 'V', 4, 2 },
  { 'W', 4, 2 },
  { 'X', 8, 1 },
  { 'Y', 4, 2 },
  { 'Z', 10, 1 }};

int score(int word)
{
  int i;
  int score = 0;
  for (i = 0; i < 15; i++) score += letters[words[word][i]].value;
  return score;
}

int main(int n, char **v)
{
  int i, j;
  if (n!=2) { printf("gimme <wordlist>\n"); return 1; }
  read_words(v[1]);
  d = malloc(words_size * words_size * sizeof(distance)); if (!d) abort();
  for (i = 0; i < words_size; i++) for (j = 0; j < words_size; j++) {
    d[i*words_size+j].word1 = i;
    d[i*words_size+j].word2 = j;
    d[i*words_size+j].distance = dist(i, j);
    d[i*words_size+j].score1 = score(i);
    d[i*words_size+j].score2 = score(j);
  }
  qsort(d, words_size * words_size, sizeof(distance), order);
  for (i = 0; i < words_size * words_size; i++) {
    printf("%d %d %d %d ", d[i].distance, d[i].score1, d[i].score2, d[i].score1+d[i].score2);
    for (j = 0; j < 15; j++) printf("%c", words[d[i].word1][j] + 'a' - 1);
    printf(" ");
    for (j = 0; j < 15; j++) printf("%c", words[d[i].word2][j] + 'a' - 1);
    printf("\n");
  }
  return 0;
}
