#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "defs.h"

int is_positive_letter(int board[15][15], int row, int col)
{
  return row >= 0 && row < 15 && col >= 0 && col < 15 && board[row][col] > 0;
}

int other_word_ok(game *g, int l, int i, int j, int inci, int incj)
{
  extern dictionary dict;
  int si = i, sj = j;
  int word[16];
  int w;
  int ii = 1 - inci;
  int jj = 1 - incj;
  int starti = i, startj = j;
  int endi, endj;
  g->board[si][sj] = l;
  /* find start/end pos of word */
  while (starti >= 0 && startj >= 0 && g->board[starti][startj] != 0)
    starti -= ii, startj -= jj;
  starti += ii; startj += jj;
  endi = starti; endj = startj;
  while (endi < 15 && endj < 15 && g->board[endi][endj] != 0)
    endi += ii, endj += jj;
  endi -= ii; endj -= jj;
  /* get word */
  for (w=0, i=starti, j=startj; i!=endi+ii || j!=endj+jj; w++, i+=ii, j+=jj) {
    int letter = abs(g->board[i][j]); if (letter > 26) letter -= 26;
    word[w] = letter;
  }
  word[w] = 0;
  g->board[si][sj] = 0;
  return w == 1 || find_word(&dict, word);
}

int check_word(game *g, int *w, int wlen,
               int starti, int startj, int inci, int incj)
{
  int i, j, l, ok = 0;
  /* word must contain at least 1 new letter */
  for (l=0, i=starti, j=startj; l<wlen; l++, i+=inci, j+=incj)
    if (!is_positive_letter(g->board, i, j)) { ok = 1; break; }
  if (!ok) return 0;
  ok = 0;
  for (l=0, i=starti, j=startj; l<wlen; l++, i+=inci, j+=incj)
    if (is_positive_letter(g->board, i, j) ||             /* letter on board */
        is_positive_letter(g->board, i-incj, j-inci) ||   /* touch on board */
        is_positive_letter(g->board, i+incj, j+inci) ||   /* touch on board */
        (i == 7 && j == 7))            /* word is at center pos (first word) */
    { ok = 1; break; }
  if (!ok) return 0;
  for (l=0, i=starti, j=startj; l<wlen; l++, i+=inci, j+=incj) {
    if (w[l] < 0) 
      if (!other_word_ok(g, w[l], i, j, inci, incj)) return 0;
  }
  return 1;
}

int can_play_move(game *g, move *m)
{
  int lcount[27];
  int row, col, incrow, inccol, k, l;
  if (m->vert) { incrow = 1; inccol = 0; } else { incrow = 0; inccol = 1; }
  for (k = 0; k < 27; k++) lcount[k] = g->lcount[k];
  for (row=m->row,col=m->col,l=0; l<m->wordlen; row+=incrow,col+=inccol,l++) {
    if (row < 0 || row > 15 || col < 0 || col > 15) return 0;
    if (m->word[l] < 0) {
      int letter = -m->word[l];
      if (letter > 26) letter = 0;
      if (lcount[letter] == 0) return 0;
      lcount[letter]--;
    }
    if (m->word[l] > 0 && g->board[row][col] != m->word[l]) return 0;
    if (m->word[l] < 0 && g->board[row][col] != 0) return 0;
  }
  return check_word(g, m->word, m->wordlen, m->row, m->col, incrow, inccol);
}

int play_possible(game *g)
{
  game sg = *g;
  int ret = 1;
  int i;

  for (i = 0; i < g->pmsize; i++) {
    if (!g->do_premove[i]) continue;
    if (!can_play_move(g, &g->premoves[i])) { ret = 0; goto end; }
    play_move(g, &g->premoves[i]);
  }

end:
  *g = sg;
  return ret;
}

void _sk(game *g, game *g2, int m, int *played)
{
  int i;
  if (m == g->pmsize) {
    if (!play_possible(g)) return;
    for (i = 0; i < g->pmsize; i++) g->do_premove[i] = played[i];
    printf("digging ");
    for (i = 0; i < g->pmsize; i++) printf(" %d", g->do_premove[i]);
    printf("\n");
    dig(g);
    return;
  }

  //if (g->must_play[m] == 0) _sk(g, g2, m+1, played);

  if (can_play_move(g2, &g->premoves[m])) {
    play_move(g2, &g->premoves[m]);
    played[m] = 1;
    _sk(g, g2, m+1, played);
    played[m] = 0;
    unplay_move(g2, &g->premoves[m]);
  }

  if (g->must_play[m] == 0) _sk(g, g2, m+1, played);
}

void skip_moves(game *g, int m)
{
  game g2 = *g;
  int *played = malloc(sizeof(int) * g->pmsize);
  if (played == NULL) abort();
  memset(played, 0, sizeof(int) * g->pmsize);
  _sk(g, &g2, 0, played);
  free(played);
}
