#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include "defs.h"

void add_movelist(movelist *l, move *m)
{
  if (l->size == l->maxsize) {
    l->maxsize += 1024;
    l->v = realloc(l->v, l->maxsize * sizeof(move)); if (l->v==NULL) abort();
  }
  l->v[l->size] = *m;
  l->size++;
}

static void dpush(letterstack *s, game *g, dict_letter *n, int hl, int new)
{
  s->stack[s->stacklen].l = n;
  s->stack[s->stacklen].hl = hl;
  s->stack[s->stacklen].new = new;
  s->stacklen++;
}

static void dpop(letterstack *s)
{
  s->stacklen--;
}

static void check_word(game *g, letterstack *s, movelist *ml,
                int starti, int startj, int inci, int incj)
{
  move m;
  int i, j, l, ok = 0;
  /* word must contain at least 1 new letter */
  for (l=0, i=starti, j=startj; l<s->stacklen; l++, i+=inci, j+=incj)
    if (!is_positive_letter(g->board, i, j)) { ok = 1; break; }
  if (!ok) return;
  ok = 0;
  for (l=0, i=starti, j=startj; l<s->stacklen; l++, i+=inci, j+=incj)
    if (is_positive_letter(g->board, i, j) ||             /* letter on board */
        is_positive_letter(g->board, i-incj, j-inci) ||   /* touch on board */
        is_positive_letter(g->board, i+incj, j+inci) ||   /* touch on board */
        (i == 7 && j == 7))            /* word is at center pos (first word) */
    { ok = 1; break; }
  if (!ok) return;
  for (l = 0; l < s->stacklen; l++) {
    m.word[l] = s->stack[l].l->l;
    if (s->stack[l].hl == 0) m.word[l] += 26;
    if (s->stack[l].new) m.word[l] = -m.word[l];
  }
  m.wordlen = s->stacklen;
  m.row = starti;
  m.col = startj;
  m.vert = inci ? 1 : 0;
  m.score = score(g, m.word, m.wordlen, starti, startj, inci, incj);
  add_movelist(ml, &m);
}

static int other_word_ok(game *g, int l, int i, int j,
                         int inci, int incj)
{
  extern dictionary dict;
  int si = i, sj = j;
  int word[16];
  int w;
  int ii = 1 - inci;
  int jj = 1 - incj;
  int starti = i, startj = j;
  int endi, endj;
  g->board[si][sj] = l;
  /* find start/end pos of word */
  while (starti >= 0 && startj >= 0 && g->board[starti][startj] != 0)
    starti -= ii, startj -= jj;
  starti += ii; startj += jj;
  endi = starti; endj = startj;
  while (endi < 15 && endj < 15 && g->board[endi][endj] != 0)
    endi += ii, endj += jj;
  endi -= ii; endj -= jj;
  /* get word */
  for (w=0, i=starti, j=startj; i!=endi+ii || j!=endj+jj; w++, i+=ii, j+=jj) {
    int letter = abs(g->board[i][j]); if (letter > 26) letter -= 26;
    word[w] = letter;
  }
  word[w] = 0;
  g->board[si][sj] = 0;
  return w == 1 || find_word(&dict, word);
}

static void gogo(world *w, dict_letter *l,
          int starti, int startj, int i, int j, int inci, int incj)
{
  int hletter;
  uint32_t letter_done = 0, joker_done = 0;

  if (l==NULL) return;
  if (i==15||j==15) return;
again:
  if (w->g->board[i][j] > 0) {
    int letter = w->g->board[i][j]; if (letter > 26) letter -= 26;
    if (l->l == letter) {
      dpush(w->s, w->g, l, w->g->board[i][j] > 26 ? 0 : w->g->board[i][j], 0);
      if (l->w && (i+inci==15||j+incj==15||w->g->board[i+inci][j+incj]==0))
        check_word(w->g, w->s, w->ml, starti, startj, inci, incj);
      gogo(w, l->child, starti, startj, i + inci, j + incj, inci, incj);
      dpop(w->s);
      goto next;
    }
    goto next;
  }
  if (w->g->board[i][j] < 0) abort();

  /* no letter on board, no reserved letter */
  hletter = l->l;
_joker:
  if (w->lcount[hletter]) {
    /* must not have been tried before */
    if (hletter && (letter_done & (1 << hletter))) goto _next;
    if (!hletter && (joker_done & (1 << l->l))) goto _next;
    /* letter must be in lcount */
    if (!w->lcount[hletter]) goto _next;
    if (!other_word_ok(w->g, l->l, i, j, inci, incj)) goto _next;
    if (hletter) letter_done |= 1 << hletter;
    if (!hletter) joker_done |= 1 << l->l;
    w->lcount[hletter]--;
    dpush(w->s, w->g, l, hletter, 1);
    if (l->w && (i+inci==15||j+incj==15||w->g->board[i+inci][j+incj]==0))
      check_word(w->g, w->s, w->ml, starti, startj, inci, incj);
    gogo(w, l->child, starti, startj, i+inci, j+incj, inci, incj);
    dpop(w->s);
    w->lcount[hletter]++;
  }
_next:
  if (hletter) { hletter = 0; goto _joker; }

next:
  l = l->next;
  if (l != NULL) goto again;
}

void get_dmoves(game *g, movelist *ml, int *lcount)
{
  extern dictionary dict;
  letterstack stack;
  world w;
  int row, col;
  stack.stacklen = 0;
  w.g = g;
  w.s = &stack;
  w.ml = ml;
  w.lcount = lcount;
  for (row = 0; row < 15; row++) for (col = 0; col < 15; col++) {
    if (!is_positive_letter(w.g->board, row-1, col))
      gogo(&w, dict.top, row, col, row, col, 1, 0);
    if (!is_positive_letter(w.g->board, row, col-1))
      gogo(&w, dict.top, row, col, row, col, 0, 1);
  }
}

int order_highscore_first(const void *_a, const void *_b)
{
  const move *a = _a;
  const move *b = _b;
  if (a->score > b->score) return -1;
  if (a->score < b->score) return 1;
  return 0;
}

void go(game *g, int m, int *lcount)
{
  int i;
  movelist ml;

  if (m == g->pmsize+1) return;

  /* skip level? */
  if (m < g->pmsize && g->do_premove[m] == 0) return go(g, m+1, lcount);

  /* get all possible moves with letters in lcount */
  ml.size = 0;
  ml.maxsize = 0;
  ml.v = NULL;
  get_dmoves(g, &ml, lcount);

  qsort(ml.v, ml.size, sizeof(move), order_highscore_first);

  /* play all moves */
  for (i = 0; i < ml.size && i < 32; i++) {
    int l;
    for (l = 0; l < ml.v[i].wordlen; l++) if (ml.v[i].word[l] < 0) {
      int letter = -ml.v[i].word[l];
      if (letter > 26) letter = 0;
      if (!lcount[letter]) abort();
      lcount[letter]--;
    }
    play_move(g, &ml.v[i]);
    /* play the move of this level */
    if (m < g->pmsize && can_play_move(g, &g->premoves[m])) {
      play_move(g, &g->premoves[m]);
      go(g, m+1, lcount);
      unplay_move(g, &g->premoves[m]);
    }
    unplay_move(g, &ml.v[i]);
    for (l = 0; l < ml.v[i].wordlen; l++) if (ml.v[i].word[l] < 0) {
      int letter = -ml.v[i].word[l];
      if (letter > 26) letter = 0;
      lcount[letter]++;
    }
  }

  if (ml.maxsize) free(ml.v);

  /* play move itself with no intermediate */
  if (m == g->pmsize) return;

  if (!can_play_move(g, &g->premoves[m])) return;
  play_move(g, &g->premoves[m]);
  go(g, m+1, lcount);
  unplay_move(g, &g->premoves[m]);
}

void dig(game *g)
{
  int lcount[27];
  game sg = *g;
  int i;
  /* compute available lcount by playing "do_premove" moves */
  for (i = 0; i < g->pmsize; i++)
    if (g->do_premove[i] == 1) play_move(g, &g->premoves[i]);
  memcpy(lcount, g->lcount, sizeof(lcount));
  for (i = 0; i < 27; i++) {
    if (!lcount[i]) continue;
    printf("%c %d\n", i?i+'A'-1:'?', lcount[i]);
  }
  *g = sg;
  go(g, 0, lcount);
  *g = sg;
}
