#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "defs.h"

void reset_lcount(game *g)
{
  int l[27] = { 2, 9, 2, 2, 4, 12, 2, 3, 2, 9, 1, 1, 4, 2, 6, 8, 2,
                1, 6, 4, 6, 4, 2, 2, 1, 2, 1 };
  memcpy(g->lcount, l, sizeof(l));
}

void init_game(game *g)
{
  reset_lcount(g);
  memset(g->board, 0, sizeof(g->board));
  g->msize = 0;
  g->score = 0;
  g->pmsize = 0;
}

void read_moves(game *g, char *moves)
{
  int truescore;
  move m;
  char *p = moves;
  while (*p) {
    int row, col, incrow, inccol, vert, l, dascore = -1;
    char colc;
    while (*p && (isspace(*p) || *p == ',')) p++;
    if (*p == '*') { g->must_play[g->pmsize] = 1; p++; }
    else           g->must_play[g->pmsize] = 0;
    if (*p >= '0' && *p <= '9') {
      vert = 0; incrow = 0; inccol = 1;
      if (sscanf(p, "%d%c", &row, &colc) != 2) abort();
    } else {
      vert = 1; incrow = 1; inccol = 0;
      if (sscanf(p, "%c%d", &colc, &row) != 2) abort();
    }
    row--;
    col = colc >= 'a' && colc <= 'z' ? colc - 'a' : colc - 'A';
    if (row < 0 || row > 15 || col < 0 || col > 15) abort();
    while (*p && (!isspace(*p))) p++;
    while (*p && isspace(*p)) p++;
    m.wordlen = 0;
    while ((*p >= 'a' && *p <= 'z') || (*p >= 'A' && *p <= 'Z')) {
      if (m.wordlen > 15) abort();
      if (*p >= 'a' && *p <= 'z') m.word[m.wordlen] = *p - 'a' + 1 + 26;
      else                        m.word[m.wordlen] = *p - 'A' + 1;
      m.wordlen++;
      p++;
    }
    sscanf(p, "%d", &dascore);
    truescore = score(g, m.word, m.wordlen, row, col, incrow, inccol);
    if (dascore > 0 && dascore != truescore) abort();
    while (*p && *p != ',') p++;
    m.row = row;
    m.col = col;
    m.vert = vert;
    m.score = truescore;
    for (l=0; l<m.wordlen; row+=incrow,col+=inccol,l++) {
      if (row < 0 || row > 15 || col < 0 || col > 15) abort();
      if (m.word[l] != g->board[row][col]) m.word[l] = -m.word[l];
    }
    play_move(g, &m);
    g->premoves[g->pmsize] = m;
    g->do_premove[g->pmsize] = 1;
    g->pmsize++;
  }
  memset(g->board, 0, sizeof(g->board));
  reset_lcount(g);
  g->msize = 0;
  g->score = 0;
}

void play_move(game *g, move *m)
{
  int row, col, incrow, inccol, k, l;
  int truescore;
  if (m->vert) { incrow = 1; inccol = 0; } else { incrow = 0; inccol = 1; }
  for (k = 0; k < 27; k++) m->prelcount[k] = g->lcount[k];
  truescore = score(g, m->word, m->wordlen, m->row, m->col, incrow, inccol);
  for (row=m->row,col=m->col,l=0; l<m->wordlen; row+=incrow,col+=inccol,l++) {
    if (row < 0 || row > 15 || col < 0 || col > 15) abort();
    m->preboard[l] = g->board[row][col];
    if (m->word[l] < 0) {
      int letter = -m->word[l];
      if (letter > 26) letter = 0;
      if (g->lcount[letter] == 0) abort();
      g->lcount[letter]--;
    }
    if (m->word[l] > 0 && g->board[row][col] != m->word[l]) abort();
    if (m->word[l] < 0 && g->board[row][col] != 0) abort();
    g->board[row][col] = abs(m->word[l]);
  }
  //g->score += m->score;
  m->score = truescore;
  g->score += truescore;
  if (g->msize == 100) abort();
  g->moves[g->msize] = *m;
  g->msize++;
{
  static int max = 0;
  if (g->score > max) {
    max = g->score;
    printf("NEW max %d\n", max);
    print_board(g->board);
    print_moves(g);
  }
}
}

void unplay_move(game *g, move *m)
{
  int row, col, incrow, inccol, l, k;
  int truescore;
  if (m->vert) { incrow = 1; inccol = 0; } else { incrow = 0; inccol = 1; }
  for (row=m->row,col=m->col,l=0;l<m->wordlen;row+=incrow,col+=inccol,l++) {
    if (m->word[l] >= 0) continue;
    if (g->board[row][col] != -m->word[l]) abort();
    g->board[row][col] = m->preboard[l];
  }
  for (k = 0; k < 27; k++) g->lcount[k] = m->prelcount[k];
  truescore = score(g, m->word, m->wordlen, m->row, m->col, incrow, inccol);
  g->score -= truescore;
  g->msize--;
}

void print_moves(game *g)
{
  int i, j;
  for (i = 0; i < g->msize; i++) {
    int row = g->moves[i].row;
    int col = g->moves[i].col;
    int len = g->moves[i].wordlen; 
    int vert = g->moves[i].vert;
    int score = g->moves[i].score;
    int incrow, inccol;
    if (!vert) {
      printf("%d%c ", row+1, col+'A');
      incrow = 0;
      inccol = 1;
    } else {
      printf("%c%d ", col+'A', row+1);
      incrow = 1;
      inccol = 0;
    }
    for (j = 0; j < len; j++, row += incrow, col += inccol) {
      int c = g->board[row][col];
      int joker = c > 26; //is_joker[row][col];
      if (c > 26) c -= 26;
      printf("%c", c+'A'-1 + joker * ('a'-'A'));
    }
    printf(" %d", score);
    if (i == g->msize-1) printf("\n"); else printf(",");
  }
}

void print_board(int b[15][15])
{
  int i, j;
  for (i = 0; i < 15; i++) {
    for (j = 0; j < 15; j++) {
      int l = b[i][j];
      if (l < 0) printf("\x1b[32m");
      printf("%c", b[i][j] ?
                    abs(b[i][j]) > 26 ?
                      abs(b[i][j]) - 26 + 'a' - 1 :
                      abs(b[i][j]) + 'A' - 1 : '.');
      if (l < 0) printf("\x1b[0m");
    }
    printf("\n");
  }
}
