#include <stdio.h>
#include <stdlib.h>

#ifndef _DICT_H_
#define _DICT_H_

typedef struct dict_letter {
  struct dict_letter *next;
  struct dict_letter *child;
  int l;
  int w;
} dict_letter;

typedef struct {
  dict_letter *top;
} dictionary;

typedef struct {
  int **w;
  int wordlen;
  int size;
  int maxsize;
} wordlist;

void new_dict(dictionary *);
void add_word(dictionary *, int *);
void dump(dictionary *);
void dump_cb(dictionary *, void (*)(int *));
int find_word(dictionary *, int *);

void read_dict(char *fn, dictionary *dict, int reverse);

void init_wordlist(wordlist *, int wordlen);
void dict_to_wordlist(dictionary *, wordlist *, int wordlen);

int *char2int(char *);

#endif /* _DICT_H_ */

/*******************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

static void go(dict_letter *l, int *w)
{
  dict_letter *t;

  if (*w == 0) return;

more:
  if (l->l == *w) {
    if (w[1] == 0) { l->w = 1; return; }
    if (l->child) { go(l->child, w+1); return; }
    w++;
    goto suffix;
  }
  if (l->next) { l = l->next; goto more; }

  t = l->next = malloc(sizeof(dict_letter)); if (!t) abort();

next:
  t->next = 0;
  t->child = 0;
  t->l = *w;
  t->w = w[1] ? 0 : 1;
  l = t;
  w++;
  if (!*w) return;
suffix:
  t = l->child = malloc(sizeof(dict_letter)); if (!t) abort();
  goto next;
}

void add_word(dictionary *d, int *w)
{
  go(d->top, w);
}

void new_dict(dictionary *d)
{
  d->top = malloc(sizeof(dict_letter)); if (!d->top) abort();
  d->top->next = 0;
  d->top->child = 0;
  d->top->l = -1;
  d->top->w = 0;
}

static int word[1024];
static int pos;

static void print(void)
{
  int i;
  for (i = 0; i < pos; i++) printf("%c", word[i] + 'A' - 1);
  printf("\n");
}

static void dodump(dict_letter *l)
{
next:
  if (!l) return;
  if (pos == 1024) abort();
  word[pos] = l->l; pos++;
  if (l->w) print();
  dodump(l->child);
  pos--;
  l = l->next;
  goto next;
}

void dump(dictionary *d)
{
  pos = 0;
  dodump(d->top->next);
}

static void dodump_cb(dict_letter *l, void (*cb)(int *))
{
next:
  if (!l) return;
  if (pos == 1024) abort();
  word[pos] = l->l; pos++;
  if (l->w) { if (pos==1024) abort(); word[pos] = 0; cb(word); }
  dodump_cb(l->child, cb);
  pos--;
  l = l->next;
  goto next;
}

void dump_cb(dictionary *d, void (*cb)(int *))
{
  pos = 0;
  dodump_cb(d->top->next, cb);
}

static int find(dict_letter *l, int *w)
{
  if (*w == 0) return 0;
  if (l == NULL) return 0;
next:
  if (l->l == *w) {
    if (l->w && w[1] == 0) return 1;
    return find(l->child, w+1);
  }
  l = l->next;
  if (l != NULL) goto next;
  return 0;
}

#if 0
char *get_word(int *w)
{
  int i;
  static char v[16];
  for (i = 0; w[i]; i++) v[i] = w[i] + 'A' - 1;
  v[i] = 0;
  return v;
}
#endif

int find_word(dictionary *d, int *w)
{
//printf("yo %s %d\n", get_word(w), find(d->top, w));
  return find(d->top, w);
}

int up(int n)
{
  if (isascii(n)) return toupper(n)-'A'+1;
  abort();
}

int utf8(unsigned char *n, int *_len)
{
  int v = 0;
  if (*n == 0xc0 || *n == 0xc1 ||
      *n >= 0xf5) abort();
  if (!(*n & 0x80)) {
    v = *n;
    *_len = 1;
  } else
  if ((*n & 0xe0) == 0xc0) { /* 110xxxxx */
    if ((*(n+1) & 0xc0) != 0x80) abort();
    v = ((*n & 0x1f) << 6) | (*(n+1) & 0x3f);
    if (v < 0x80 || v > 0x7ff) abort();
    *_len = 2;
  } else
  if ((*n & 0xf0) == 0xe0) { /* 1110xxxx */
    if ((*(n+1) & 0xc0) != 0x80 ||
        (*(n+2) & 0xc0) != 0x80) abort();
    v = ((*n & 0x0f) << (6*2)) | ((*(n+1) & 0x3f) << 6) | (*(n+2) & 0x3f);
    if (v < 0x800 || v > 0xffff) abort();
    *_len = 3;
  } else
  if ((*n & 0xf8) == 0xf0) { /* 11110xxx */
    if ((*(n+1) & 0xc0) != 0x80 ||
        (*(n+2) & 0xc0) != 0x80 ||
        (*(n+3) & 0xc0) != 0x80) abort();
    v = ((*n & 0x07) << (6*3)) | ((*(n+1) & 0x3f) << (6*2)) |
        ((*(n+2) & 0x3f) << 6) | (*(n+3) & 0x3f);
    if (v < 0x10000 || v > 0x10ffff) abort();
    *_len = 4;
  } else abort();
  return v;
}

void read_dict(char *fn, dictionary *dict, int reverse)
{
  FILE *f = fopen(fn, "r"); if (!f) { perror(fn); abort(); }
  while (1) {
    unsigned char n[128];
    int ni[128];
    int pos;
    int i;
    int len;
    int wordlen;
    if (!fgets((char*)n, 128, f)) break;
    for (i = 0; i < 127 && n[i]; i++) if (isspace(n[i])) { n[i] = 0; break; }
    n[i] = 0;
    if ((len=wordlen=strlen((char*)n)) == 0 || len > 15) continue;
    if (reverse)
      for (i = 0, pos = 0; n[pos]; i++, pos+=len)
        ni[i] = utf8(n+wordlen-1-pos, &len);
    else
      for (i = 0, pos = 0; n[pos]; i++, pos+=len) ni[i] = utf8(n+pos, &len);
    ni[i] = 0;
    for (i = 0; ni[i]; i++) ni[i] = up(ni[i]);
    ni[i] = 0;
    add_word(dict, ni);
  }
  fclose(f);
  return;
}

void init_wordlist(wordlist *w, int wordlen)
{
  w->w = NULL;
  w->size = w->maxsize = 0;
  w->wordlen = wordlen;
}

void wordlist_add(wordlist *w, int *word)
{
  int *x = malloc(sizeof(int) * w->wordlen); if (x == NULL) abort();
  if (w->size == w->maxsize) {
    w->maxsize += 32768;
    w->w = realloc(w->w, w->maxsize * sizeof(int *));
    if (w->w == NULL) abort();
  }
  memcpy(x, word, sizeof(int) * w->wordlen);
  w->w[w->size] = x;
  w->size++;
}

static int dwstack[16];
static int dwsize;

static void dw_push(int l)
{
  if (dwsize == 15) abort();
  dwstack[dwsize] = l;
  dwsize++;
}

static void dw_pop(void)
{
  if (dwsize == 0) abort();
  dwsize--;
}

static void go_dw(dict_letter *l, wordlist *w, int wordlen, int len)
{
  if (len > wordlen) return;
  while (l != NULL) {
    dw_push(l->l);
    if (l->w && dwsize == wordlen) wordlist_add(w, dwstack);
    go_dw(l->child, w, wordlen, len+1);
    dw_pop();
    l = l->next;
  }
}

void dict_to_wordlist(dictionary *d, wordlist *w, int wordlen)
{
  go_dw(d->top->next, w, wordlen, 1);
}

int *char2int(char *x)
{
  static int y[16];
  int i;
  for (i = 0; *x && i < 15; i++, x++)
    if (*x == '?')
      y[i] = 0;
    else if (*x >= 'A' && *x <= 'Z')
      y[i] = *x - 'A' + 1;
    else if (*x >= 'a' && *x <= 'z')
      y[i] = *x - 'a' + 1;
    else abort();
  if (*x) abort();
  y[i] = 0;
  return y;
}
/*******************************************************/

dictionary dict;

int fail(void) { abort(); }

int main(int n, char **v)
{
  char w[128];
  int i;
  int l,p;
  int word[16];
  int subword[16];
  if (n != 2) { printf("gimme <dict>\n"); return 1; }
  new_dict(&dict);
  read_dict(v[1], &dict, 0);
more:
  for (i = 0; i < 15; i++)
    { int c = getchar(); if (c == EOF) goto end; w[i] = c; }
  if (getchar() != '\n') abort();
  w[i] = 0;
  for (i = 0; i < 15; i++)
    word[i] = w[i] >= 'a' && w[i] <= 'z' ? w[i] - 'a' + 1 :
              w[i] >= 'A' && w[i] <= 'Z' ? w[i] - 'A' + 1 : fail();
  word[i] = 0;
  for (l = 15; l > 1; l--) for (p = 0; p < 16-l; p++) {
    for (i = 0; i < l; i++) subword[i] = word[i+p];
    subword[i] = 0;
    if (find_word(&dict, subword)) {
      printf("%.2d: ", l);
      for (i = 0; i < p; i++) printf(" ");
      for (i = 0; subword[i]; i++) printf("%c", subword[i] + 'a' - 1);
      printf("\n");
    }
  }
  goto more;
end:
  return 0;
}
